/**
	$Id: InternalSlotManager.java 6444 2014-04-09 12:21:20Z nicklas $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.jobagent.slotmanager;


import net.sf.basedb.clients.jobagent.Agent;
import net.sf.basedb.core.Job;
import net.sf.basedb.util.Values;

/**
	The internal slot manager assigns slot to jobs based on their
	estimated execution time. The configuration file gives the 
	maximum number of slots for each execution time:
	
	<ul>
	<li>agent.shortest.slots: {@link net.sf.basedb.core.Job.ExecutionTime#SHORTEST}
	<li>agent.short.slots: {@link net.sf.basedb.core.Job.ExecutionTime#SHORT}
	<li>agent.medium.slots: {@link net.sf.basedb.core.Job.ExecutionTime#MEDIUM}
	<li>agent.long.slots: {@link net.sf.basedb.core.Job.ExecutionTime#LONG}
	</ul>

	@author Nicklas
	@since 2.16
	@base.modified $Date: 2014-04-09 14:21:20 +0200 (Wed, 09 Apr 2014) $
*/
public class InternalSlotManager
	implements SlotManager
{

	private static final org.slf4j.Logger log = 
		org.slf4j.LoggerFactory.getLogger(InternalSlotManager.class);

	private Agent agent;
	private int[] maxSlots;
	private int[] usedSlots;
	
	public InternalSlotManager()
	{}
	
	/*
		From the SlotManager interface
		------------------------------
	*/
	@Override
	public void init(Agent agent)
	{
		log.info("Initalizing internal slot manager");
		this.agent = agent;
		this.maxSlots = new int[Job.ExecutionTime.values().length];
		this.usedSlots = new int[Job.ExecutionTime.values().length];
		for (Job.ExecutionTime et : Job.ExecutionTime.values())
		{
			int index = et.ordinal();
			String configName = "agent."+et.name().toLowerCase() + ".slots";
			int configuredSlots = Values.getInt(agent.getProperty(configName), et.getDefaultSlots());
			maxSlots[index] = configuredSlots;
			usedSlots[index] = 0;
			log.debug("Slot " + et.name() + "=" + configuredSlots);
		}
	}

	@Override
	public void close()
	{
		log.info("Closing internal slot manager");
		this.agent = null;
		this.maxSlots = null;
		this.usedSlots = null;
	}

	@Override
	public synchronized Slot getSlot(Job job)
	{
		return getSlot(job.getEstimatedExecutionTime());
	}

	@Override
	public synchronized void releaseSlot(Slot slot)
	{
		Job.ExecutionTime time = slot.getEstimatedExecutionTime();
		int index = time.ordinal();
		usedSlots[index]--;
		if (log.isDebugEnabled())
		{
			log.debug("Released slot " + time.name() + ": used=" + usedSlots[index] + "; max=" + maxSlots[index]);
		}
	}
	// -----------------------------------------------
	
	/**
		Get the job agent this slot manager belongs to.
		@see #init(Agent)
	*/
	protected Agent getAgent()
	{
		return agent;
	}
	
	/**
		Get a slot for a job with the given estimated execution time. If
		no slot is available at the specified execution time, longer times
		are searched. If all slots are taken, null is returned.
		
		@param estimated The estimated execution time of the slot
		@return A slot or null if no slot is available
	*/
	protected synchronized Slot getSlot(Job.ExecutionTime estimated)
	{
		Slot slotToUse = null;

		// Check all slots from the requested execution time and longer execution times
		for (int i = estimated.ordinal(); i < Job.ExecutionTime.values().length; ++i)
		{
			Job.ExecutionTime time = Job.ExecutionTime.values()[i];
			if (log.isDebugEnabled())
			{
				log.debug("Slot " + time.name() + ": used=" + usedSlots[i] + "; max=" + maxSlots[i]);
			}
			if (usedSlots[i] < maxSlots[i])
			{
				// This slot has free jobs
				usedSlots[i]++;
				slotToUse = new Slot(time);
				break;
			}
		}
		// If null we couldn't find a free slot
		return slotToUse;
	}
}
