package net.sf.basedb.clients.web.extensions.list;

import java.util.ArrayList;
import java.util.List;

import net.sf.basedb.clients.web.extensions.AbstractJspActionFactory;
import net.sf.basedb.clients.web.extensions.JspContext;
import net.sf.basedb.core.Annotatable;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemContext;
import net.sf.basedb.core.Metadata.PropertyPath;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.extensions.InvokationContext;

/**
	Action factory for adding columns to a table that pull data from a parent
	or child item from an arbitrary location up or down the chain of linked items.
	The factory will scan the "columns" setting from the current context for 
	entries matching the pattern: /ITEMTYPE/Subtype/property-or-annotation
	
	It will use the item list API to find a
	parent item that matches the ITEMTYPE and Subtype and the retrieve
	the property using a {@link PropertyPath} or an annotation.

	@author nicklas
	@since 3.16, 3.17
*/
public class RelatedItemColumnActionFactory
	extends AbstractJspActionFactory<ListColumnAction<Annotatable, Object>>
{
	
	public RelatedItemColumnActionFactory()
	{}

	/**
		We get the "columns" setting from the current ItemContext and check if
		there seems to be an entry for a related item.
	*/
	@Override
	public boolean prepareContext(InvokationContext<? super ListColumnAction<Annotatable, Object>> context) 
	{
		JspContext jspContext = (JspContext)context.getClientContext();
		SessionControl sc = jspContext.getSessionControl();
		GuiContext guiContext = jspContext.getGuiContext();
		Item item = guiContext.getItem();
		String subContext = Values.getString(guiContext.getSubContext(), "");
		
		ItemContext cc = sc.getCurrentContext(item, subContext, null);
		if (cc == null) return false;
		
		String allColumns = cc.getSetting("columns");
		if (allColumns == null || !allColumns.contains("/")) return false;

		jspContext.addScript(jspContext.getRoot() + "/include/scripts/lazy-items.js");
		return super.prepareContext(context);
	}

	
	@SuppressWarnings({"unchecked", "rawtypes"})
	@Override
	public ListColumnAction<Annotatable, Object>[] getActions(InvokationContext<? super ListColumnAction<Annotatable, Object>> context) 
	{
		JspContext jspContext = (JspContext)context.getClientContext();
		SessionControl sc = jspContext.getSessionControl();
		GuiContext guiContext = jspContext.getGuiContext();
		Item item = guiContext.getItem();
		String subContext = Values.getString(guiContext.getSubContext(), "");
		
		ItemContext cc = sc.getCurrentContext(item, subContext, null);
		if (cc == null) return null;
		
		String allColumns = cc.getSetting("columns");
		if (allColumns == null || !allColumns.contains("/")) return null;
		
		
		boolean lazy = !Boolean.FALSE.equals(jspContext.getAttribute("lazy-loading"));
		DbControl dc = jspContext.getDbControl();
		RelatedItemHelper helper = new RelatedItemHelper(sc, lazy);
		List<ListColumnAction<Annotatable, Object>> actions = new ArrayList<>(); 
		for (String col : allColumns.split(","))
		{
			if (col.startsWith("/"))
			{
				ListColumnAction<Annotatable, Object> action = RelatedItemColumn.create(dc, actions.size(), item, col, helper);
				if (action != null) actions.add(action);
			}
		}
		
		if (actions.size() == 0) return null;
		
		return actions.toArray(new ListColumnAction[actions.size()]);
	}
	
}
