/**
	$Id: ToolbarUtil.java 7605 2019-02-26 10:10:15Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.extensions.toolbar;


import net.sf.basedb.clients.web.extensions.ExtensionsControl;
import net.sf.basedb.clients.web.extensions.JspContext;
import net.sf.basedb.core.Annotatable;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.Ownable;
import net.sf.basedb.core.Removable;
import net.sf.basedb.core.Shareable;
import net.sf.basedb.core.data.AnnotatableData;
import net.sf.basedb.core.data.OwnableData;
import net.sf.basedb.core.data.RemovableData;
import net.sf.basedb.core.data.ShareableData;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.util.extensions.ExtensionPointBean;
import net.sf.basedb.util.extensions.ExtensionsInvoker;
import net.sf.basedb.util.extensions.Registry;

/**
	Utility class for toolbar extension points.
	
	@author Nicklas
	@version 2.12
	@base.modified $Date: 2019-02-26 11:10:15 +0100 (Tue, 26 Feb 2019) $
*/
public class ToolbarUtil
{

	/**
		An array containing all items that we are registering an
		extension point for. The extension point id for an item
		is constructed as: EP_PREFIX[0/1] + Item.name().toLowerCase().
	*/
	public static final Item[] TOOLBAR_ITEMS = new Item[] 
		{
			Item.EXPERIMENT, Item.BIOASSAYSET, Item.BIOASSAY,
			Item.TRANSFORMATION, Item.EXTRAVALUE, Item.SPOTDATA,
			Item.RAWBIOASSAY, Item.RAWDATA, Item.FORMULA,
			Item.REPORTER, Item.REPORTERLIST, Item.REPORTERSCORE, Item.REPORTERCLONETEMPLATE,
			Item.DERIVEDBIOASSAY, Item.PHYSICALBIOASSAY,
			Item.BIOSOURCE, Item.SAMPLE, Item.EXTRACT, 
			Item.TAG, Item.BIOMATERIALEVENT,
			Item.BIOPLATE, Item.BIOWELL, Item.BIOPLATETYPE,
			Item.BIOPLATEEVENT, Item.BIOPLATEEVENTTYPE,
			Item.PLATE, Item.WELL, Item.PLATEEVENT,
			Item.PLATETYPE, Item.PLATEEVENTTYPE,
			Item.PLATEGEOMETRY, Item.PLATEMAPPING,
			Item.ARRAYDESIGN, Item.ARRAYBATCH, Item.ARRAYSLIDE,
			Item.FEATURE,
			Item.USER, Item.GROUP, Item.ROLE, Item.QUOTA, Item.QUOTATYPE, Item.NEWS,
			Item.SOFTWARE, Item.HARDWARE, Item.KIT,
			Item.PROTOCOL, Item.CLIENT, Item.HELP,
			Item.PLATFORM, Item.PLATFORMVARIANT, Item.DATAFILETYPE, 
			Item.PLUGINDEFINITION, Item.PLUGINTYPE, Item.PLUGINCONFIGURATION, Item.JOBAGENT,
			Item.MIMETYPE, Item.REPORTERTYPE, Item.EXTRAVALUETYPE,
			Item.ANNOTATIONTYPE, Item.ANNOTATIONTYPECATEGORY, Item.QUANTITY, Item.UNIT,
			Item.SESSION, Item.JOB, Item.MESSAGE, Item.PROJECT, Item.FILE, Item.FILESERVER,
			Item.CHANGEHISTORY, Item.PERMISSIONTEMPLATE, Item.ITEMSUBTYPE, Item.ITEMLIST,
			Item.USERDEVICE, Item.ANYTOANY
		};

	
	/**
		ID prefixes for toolbar extension point.
		EP_PREFIX[0] = list pages; EP_PREFIX[1] = single-item pages
	*/
	public static String[] EP_PREFIX = {
		"net.sf.basedb.clients.web.toolbar.list.", 
		"net.sf.basedb.clients.web.toolbar.item."};

	public static final int LIST = 0;
	public static final int ITEM = 1;
	
	/**
		ID:s for the toolbar extension point that appear on all pages.
		EP_ALL[0] = list pages; EP_ALL[1] = single-item pages
	*/
	public static String[] EP_ALL = 
		{EP_PREFIX[LIST] + "all", EP_PREFIX[ITEM] + "all"};

	/**
		ID:s for the toolbar extension point that appear on all pages
		that displays {@link Ownable} items.
		EP_OWNABLE[0] = list pages; EP_OWNABLE[1] = single-item pages
	*/
	public static String[] EP_OWNABLE = 
		{EP_PREFIX[LIST] + "ownable", EP_PREFIX[ITEM] + "ownable"};
	
	/**
		ID:s for the toolbar extension point that appear on all pages
		that displays {@link Shareable} items.
		EP_SHAREABLE[0] = list pages; EP_SHAREABLE[1] = single-item pages
	*/
	public static String[] EP_SHAREABLE = 
		{EP_PREFIX[LIST] + "shareable", EP_PREFIX[ITEM] + "shareable"};
	
	/**
		ID:s for the toolbar extension point that appear on all pages
		that displays {@link Removable} items.
		EP_REMOVABLE[0] = list pages; EP_REMOVABLE[1] = single-item pages
	*/
	public static String[] EP_REMOVABLE = 
		{EP_PREFIX[LIST] + "removeable", EP_PREFIX[ITEM] + "removeable"};

	/**
		ID:s for the toolbar extension point that appear on all pages
		that displays {@link Annotatable} items.
		EP_ANNOTATABLE[0] = list pages; EP_ANNOTATABLE[1] = single-item pages
	*/
	public static String[] EP_ANNOTATABLE = 
		{EP_PREFIX[LIST] + "annotatable", EP_PREFIX[ITEM] + "annotatable"};

	/**
		Use toolbar extensions for a given gui context. This method
		will assemble all specific and generic extension points and
		then call {@link ExtensionsControl#useExtensions(JspContext, String...)}.
		
		@param jspContext The current jsp context
		@return An invoker instance
	*/
	public static ExtensionsInvoker<ButtonAction> useExtensions(JspContext jspContext)
	{
		String[] ep = new String[6]; // 6 is the maximum number of extension points
		int index = 0;
		GuiContext guiContext = jspContext.getGuiContext();
		int type = guiContext.getType() == GuiContext.Type.LIST ? LIST : ITEM;
		Item itemType = guiContext.getItem();
		Class<?> dataClass = itemType.getDataClass();
		ep[index++] = EP_ALL[type];
		if (dataClass != null)
		{
			if (OwnableData.class.isAssignableFrom(dataClass)) 
			{
				ep[index++] = EP_OWNABLE[type];
			}
			if (ShareableData.class.isAssignableFrom(dataClass)) 
			{
				ep[index++] = EP_SHAREABLE[type];
			}
			if (RemovableData.class.isAssignableFrom(dataClass)) 
			{
				ep[index++] = EP_REMOVABLE[type];
			}
			if (AnnotatableData.class.isAssignableFrom(dataClass)) 
			{
				ep[index++] = EP_ANNOTATABLE[type];
			}
		}
		ep[index++] = EP_PREFIX[type] + itemType.name().toLowerCase();
		return ExtensionsControl.useExtensions(jspContext, ep);
	}
	

	public static void registerExtensionPoints(Registry registry)
	{
		register(registry, EP_ALL[LIST], "Toolbar: all list pages");
		register(registry, EP_ALL[ITEM], "Toolbar: all single-item pages");
		register(registry, EP_OWNABLE[LIST], "Toolbar: list pages for items that are owned by a user");
		register(registry, EP_OWNABLE[ITEM], "Toolbar: single-item pages for items that are owned by a user");
		register(registry, EP_SHAREABLE[LIST], "Toolbar: list pages for items that can be shared");
		register(registry, EP_SHAREABLE[ITEM], "Toolbar: single-item pages for items that can be shared");
		register(registry, EP_REMOVABLE[LIST], "Toolbar: list pages for items that can be put in the trashcan");
		register(registry, EP_REMOVABLE[ITEM], "Toolbar: single-item pages for items that can be put in the trashcan");
		register(registry, EP_ANNOTATABLE[LIST], "Toolbar: list pages for annotatable items");
		register(registry, EP_ANNOTATABLE[ITEM], "Toolbar: single-item pages for annotatable items");
		
		// Special toolbars
		register(registry, EP_PREFIX[LIST] + "trashcan", "Toolbar: trashcan list");
		register(registry, EP_PREFIX[ITEM] + "trashcan", "Toolbar: trashcan detail");
		register(registry, EP_PREFIX[LIST] + "allitems", "Toolbar: all items");
		register(registry, EP_PREFIX[LIST] + "projectitems", "Toolbar: project items");
		register(registry, EP_PREFIX[ITEM] + "overview", "Toolbar: item overview");
		register(registry, EP_PREFIX[LIST] + "diskusage", "Toolbar: disk usage");
		
		for (Item item : TOOLBAR_ITEMS)
		{
			String suffix = item.name().toLowerCase();
			register(registry, EP_PREFIX[LIST] + suffix, 
					item + ": Toolbar on list page");
			register(registry, EP_PREFIX[ITEM] + suffix, 
					item + ": Toolbar on single-item page");
		}
	}
	
	private static void register(Registry registry, String id, String name)
	{
		ExtensionPointBean<ButtonAction> toolbar = new ExtensionPointBean<ButtonAction>();
		toolbar.setActionClass(ButtonAction.class);
		toolbar.setRendererFactory(new ToolbarButtonRendererFactory());
		toolbar.setAllowRendererOverrider(true);
		toolbar.setId(id);
		toolbar.setName(name);
		toolbar.setDescription(
			"Extension point for the toolbar on the specified page. " +
			"Extensions should provide 'ActionButton' instances with " +
			"information about the toolbar button. The default renderer " +
			"factory is 'ToolbarButtonRendererFactory', but this can be " +
			"overridden by the extensions. This extension point supports " +
			"custom stylesheets and java-scripts."
			);
		registry.registerExtensionPoint(toolbar, null);
	}
}
