/*
	$Id: Table.java 7957 2021-06-03 07:39:33Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.taglib.table;

import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemContext;
import net.sf.basedb.clients.web.extensions.list.ListColumnAction;
import net.sf.basedb.clients.web.taglib.Page;
import net.sf.basedb.clients.web.taglib.StylableTag;
import net.sf.basedb.clients.web.util.HTML;
import net.sf.basedb.util.formatter.Formatter;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Map;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.LinkedList;
import java.util.Arrays;
import java.util.Set;

import javax.servlet.jsp.JspException;
import javax.servlet.jsp.JspTagException;
import javax.servlet.http.HttpServletRequest;

import org.json.simple.JSONArray;
import org.json.simple.JSONObject;

/**
	This is the main tag for the table. It must be located inside
	a {@link Page &lt;base:page&gt;} tag.

	<p>
	<b>Syntax:</b>
	<pre>
   &lt;tbl:table
      id=...
      clazz=....
      style=...
      title=...
      sortby=...
      direction=ASC|DESC
      columns=...
      filterrows=...
      sc=...
      item=...
      subcontext=...
      action=...
      stickyheaders=...
      dragcolumns=true|false|null
   &gt;
</pre>

	<table class="as-grid">
	<caption>Tag attributes</caption>
	<tr>
		<td><b>Attribute</b></td>
		<td><b>Default value</b></td>
		<td><b>Required</b></td>
		<td><b>Description</b></td>
	</tr>
	<tr>
		<td>id</td>
		<td>-</td>
		<td>yes</td>
		<td>
			The ID of the table. The value of this attribute
			goes directly into the standard HTML <code>id</code> attribute.
			This means that if you need a reference to the table object in your own
			JavaScript code, you may for example use the following code:
			<pre class="code">
table = document.getElementById('&lt;id&gt;');
</pre>
		</td>
	</tr>
	<tr>
		<td>clazz</td>
		<td>-</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>class</code> attribute. This allows you to apply different styles
			to different tables.
		</td>
	</tr>
	<tr>
		<td>style</td>
		<td>-</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>style</code> attribute. This allows you to apply different styles
			to different tables.
		</td>
	</tr>
	<tr>
		<td>title</td>
		<td>-</td>
		<td>no</td>
		<td>
			An optional title that will be displayed in <code>&lt;h3&gt;</code>
			format directly above the table.
		</td>
	</tr>
	<tr>
		<td>sortby</td>
		<td>-</td>
		<td>no</td>
		<td>
			Matches the <code>property</code> attribute on a <code>&lt;tbl:columdef&gt;</code> tag.
			This column will then be marked with an upwards or downward pointer
			according to the <code>direction</code> attribute.
		</td>
	</tr>
	<tr>
		<td>direction</td>
		<td>asc</td>
		<td>no</td>
		<td>
			The value should be either <code>asc</code> or <code>desc</code> to
			indicate if the rows are sorted in ascending or descending order. 
			Other values are interpreted as <code>asc</code>.
		</td>
	</tr>
	<tr>
		<td>columns</td>
		<td>-</td>
		<td>no</td>
		<td>
			A comma-separatd list of column ID:s. Only columns with an ID in the list
			will be shown in the table. The special value <code>all</code> shows all
			columns. Individual columns may override this setting by
			specifying <code>always</code> or <code>never</code> instead of the default 
			<code>auto</code> value in their <code>show</code> attribute.
		</td>
	</tr>
	<tr>
		<td>filterrows</td>
		<td>1</td>
		<td>no</td>
		<td>
			Number of filter rows that should be displayed in the table.
		</td>
	</tr>
	<tr>
		<td>action</td>
		<td>name of current page</td>
		<td>no</td>
		<td>
			The action property of the &lt;form&gt; tag.
		</td>
	</tr>
	<tr>
		<td>sc</td>
		<td>-</td>
		<td>no</td>
		<td>
			The current SessionControl object. Used by some child taglibs for 
			generating context sensitive information.
		</td>
	</tr>
	<tr>
		<td>item</td>
		<td>-</td>
		<td>no</td>
		<td>
			The current Item object. Used by some child taglibs for 
			generating context sensitive information (ie. PresetSelector).
		</td>
	</tr>
	<tr>
		<td>subcontext</td>
		<td>"" (empty string)</td>
		<td>no</td>
		<td>
			The name of the current subcontext. Used by some child taglibs for 
			generating context sensitive information (ie. PresetSelector).
		</td>
	</tr>
	<tr>
		<td>stickyheaders</td>
		<td>-</td>
		<td>no</td>
		<td>
			Enable sticky headers and optionally set the name of a sticky column.
			Do not set this attribute if the table should not have sticky headers.
		</td>
	</tr>
	<tr>
		<td>dragcolumns</td>
		<td>null</td>
		<td>no</td>
		<td>
			If set, controls if columns can be re-ordered by dragging them in 
			the GUI. If null, the feature is enabled only if the "action" and "item"
			attributes are not null.
		</td>
	</tr>
	</table>
	<p>
	The <code>&lt;tbl:table&gt;</code> may contain the following subtags:
	<ul>
	<li>{@link Toolbar &lt;tbl:toolbar&gt;}: Creates a toolbar with buttons
	<li>{@link Navigator &lt;tbl:navigator&gt;}: Creates the page navigator
	<li>{@link Data &lt;tbl:data&gt;}: Begins the data section
	</ul>
	
	@author Nicklas
	@version 2.0
*/
public class Table
	extends StylableTag
{

	/**
		
	*/
	private static final long serialVersionUID = -6717117667917148241L;

	public Table()
	{
		super("itemlist");
	}
	
	/**
		The parent &lt;base:page&gt; tag.
	*/
	private Page page;

	/**
		An optional title of the table.
	*/
	private String title = null;

	/**
		The column to sort by. Should match the sort attribute on the column tag.
	*/
	private String sortby = null;
	private Map<String, ItemContext.SortDirection> sortColumns = null;
	
	private String action = null;
	
	private transient SessionControl sc = null;
	
	private Item itemType = null;
	
	private String subcontext = "";
	
	/**
		ASC or DESC
	*/
	private ItemContext.SortDirection sortDirection = ItemContext.SortDirection.ASC;

	/**
		Which columns that should be visible and their order
	*/
	private String columns = "all";
	
	/**
		If sticky headers is enabled or not.
	*/
	private boolean stickyHeaders = false;

	/**
		The id of a column that is "sticky".
	*/
	private String stickyCol = null;
	
	/**
		Number of filter rows in the table.
	*/
	private int numFilterRows = 1;
	
	private StringBuilder hiddenForm;
	
	/**
		Keep track of which class to use for the next row.
	*/
	private int evenodd;
	
	/**
		The class names for even and odd row respectively.
	*/
	private static final String[] rowclass = new String[] { "bg-evenrow", "bg-oddrow" } ;

	/**
		"all" was specified for the columns attribute.
	*/
	private boolean allVisible;

	/**
		The ID of all visible columns, sorted by the order they should appear.
	*/
	private Set<String> visibleColumns;
	
	/**
		The ID of all defined columns
	*/
	private Set<String> definedColumns;

	private Map<String, String> columnContent;
	private Map<String, String> columnFilter;
	private Map<String, String> columnSubtitle;
	private transient Map<String, Formatter<?>> columnFormatter;
	private List<String> hiddenFilteredColumns;
	
	@SuppressWarnings("rawtypes")
	private List<ListColumnAction> extensionColumns;
	
	private JSONArray jsonColumnDefs;
	
	private int numHeaderColumns;
	private int numFilteredColumns;

	/*
		Taglib initialization methods
	*/
	public void setTitle(String title)
	{
		this.title = title;
	}
	public String getTitle()
	{
		return title;
	}

	public void setColumns(String columns)
	{
		this.columns = columns;
	}
	public String getColumns()
	{
		return columns;
	}

	/**
		@since 3.18.1
	*/
	public void setStickyheaders(String stickyCol)
	{
		this.stickyCol = stickyCol;
		this.stickyHeaders = true;
	}
	
	/**
		@since 3.18.1
	*/
	public boolean hasStickyHeaders()
	{
		return stickyHeaders;
	}
	
	/**
		@since 3.18.1
	*/
	public String getStickyCol()
	{
		return stickyCol;
	}
	
	
	private Boolean dragColumns;
	
	/**
		@since 3.8
	*/
	public void setDragcolumns(Boolean dragColumns)
	{
		this.dragColumns = dragColumns;
	}
	
	public Boolean getDragcolumns()
	{
		return dragColumns;
	}
	
	/**
		@since 3.5
	 */
	public void setFilterrows(int filterRows)
	{
		this.numFilterRows = filterRows;
	}
	public int getFilterrows()
	{
		return numFilterRows;
	}

	public void setSortby(String sortby)
	{
		this.sortby = sortby;
		if (sortby != null) 
		{
			sortColumns = new HashMap<String, ItemContext.SortDirection>();
			for (String col : sortby.split(","))
			{
				if (col.startsWith("-"))
				{
					sortColumns.put(col.substring(1), ItemContext.SortDirection.DESC);
				}
				else if (col.startsWith("+"))
				{
					sortColumns.put(col.substring(1), ItemContext.SortDirection.ASC);
				}
				else
				{
					sortColumns.put(col, null); // NOTE! Default direction may not be set yet
				}
			}
		}
		else
		{
			sortColumns = null;
		}
	}
	public String getSortby()
	{
		return sortby;
	}
	
	public void setAction(String action)
	{
		this.action = action;
	}
	public String getAction()
	{
		return action;
	}
	
	public void setSc(SessionControl sc)
	{
		this.sc = sc;
	}
	public SessionControl getSc()
	{
		return sc;
	}
	
	public void setItem(Item itemType)
	{
		this.itemType = itemType;
	}
	public Item getItem()
	{
		return itemType;
	}
	
	public void setSubcontext(String subcontext)
	{
		this.subcontext = subcontext;
	}
	public String getSubcontext()
	{
		return subcontext;
	}


	public void setDirection(String direction)
	{
		direction = direction.toUpperCase();
		this.sortDirection = ItemContext.SortDirection.valueOf(direction);
	}
	public void setDirection(ItemContext.SortDirection direction)
	{
		this.sortDirection = direction;
	}
	public ItemContext.SortDirection getDirection()
	{
		return sortDirection == null ? ItemContext.SortDirection.ASC : sortDirection;
	}

	/*
		Package methods
	*/
	Page getPage()
	{
		return page;
	}
	@SuppressWarnings("rawtypes")
	void addColumnDef(ColumnDef cd, ListColumnAction extensionColumn)
	{
		if (extensionColumn != null)
		{
			if (extensionColumns == null) extensionColumns = new ArrayList<ListColumnAction>();
			extensionColumns.add(extensionColumn);
		}
		String columnId = cd.getId();
		boolean alwaysShow = "always".equals(cd.getShow());
		boolean alwaysHide = "never".equals(cd.getShow());
		if (alwaysHide)
		{
			visibleColumns.remove(columnId);
		}
		else if (alwaysShow || allVisible)
		{
			visibleColumns.add(columnId);
		}
		definedColumns.add(columnId);
		if (cd.getFormatter() != null) columnFormatter.put(columnId, cd.getFormatter());
		if (cd.getSubtitle() != null) columnSubtitle.put(columnId, cd.getSubtitle());
		
		JSONObject jsonCol = cd.asJsonObject();
		jsonCol.put("visible", isColumnVisible(columnId));
		jsonColumnDefs.add(jsonCol);
	}
	
	void endColumnDefinitions()
	{
		visibleColumns.retainAll(definedColumns);
	}
	
	boolean isColumnVisible(String columnId)
	{
		return allVisible || visibleColumns.contains(columnId);
	}
	Set<String> getVisibleColumns()
	{
		return visibleColumns;
	}
	boolean isColumnSticky(String columnId)
	{
		return stickyHeaders && columnId.equals(stickyCol);
	}
	Formatter<?> getFormatter(String columnId)
	{
		return columnFormatter.get(columnId);
	}
	void setColumnContent(String columnId, String content)
	{
		columnContent.put(columnId, content);
	}
	Map<String, String> getColumnContent()
	{
		return columnContent;
	}
	Map<String, String> getColumnSubtitles()
	{
		return columnSubtitle;
	}
	void setColumnFilter(String columnId, int row, String filterHtml)
	{
		columnFilter.put(columnId+"@"+row, filterHtml);
	}
	void addFilteredColumn(String columnId, boolean visible)
	{
		numFilteredColumns++;
		if (!visible) hiddenFilteredColumns.add(columnId);
	}
	List<String> getHiddenFilteredColumns()
	{
		return hiddenFilteredColumns;
	}
	int writeColumns()
		throws Exception
	{
		for (String column : visibleColumns)
		{
			String content = columnContent.get(column);
			if (content == null) content = "<td class=\"cell\"></td>";
			pageContext.getOut().print(content);
		}
		columnContent.clear();
		return visibleColumns.size();
	}
	int writeColumnFilters(int row)
		throws Exception
	{
		for (String column : visibleColumns)
		{
			String content = columnFilter.get(column+"@"+row);
			if (content == null) content = "<th class=\"propertyfilter\"></td>";
			pageContext.getOut().print(content);
		}
		return visibleColumns.size();
	}
	int writeColumnSubtitles(StringBuilder sb)
	{
		int numVisibleSubtitles = 0;
		for (String column : visibleColumns)
		{
			sb.append("<th class=\"subtitle"+(isColumnSticky(column)?" sticky-col bg-filled-100":"")+"\">");
			String content = columnSubtitle.get(column);
			if (content != null) 
			{
				sb.append(content);
				numVisibleSubtitles++;
			}
			sb.append("</th>");
		}
		return numVisibleSubtitles;
	}
	void endColumnFilters()
	{
		columnFilter.clear();
	}
	/**
		Get the sort direction on the given column. If it is not
		used for sorting, null is returned.
		@since 3.3
	*/
	ItemContext.SortDirection getSortDirection(String colum)
	{
		ItemContext.SortDirection direction = null;
		if (sortColumns != null && sortColumns.containsKey(colum))
		{
			direction = sortColumns.get(colum);
			if (direction == null) direction = getDirection();
		}
		return direction;
	}
	
	/**
		Get sort direction &lt;img&gt; element for the given direction.
		Return an empty string if direction = null.
		@since 3.3
	*/
	String getDirectionImage(ItemContext.SortDirection direction)
	{
		String image = "";
		if (direction != null)
		{
			image = "<img src=\""+getPage().getRoot()+"images/" + 
					(direction == ItemContext.SortDirection.ASC ? "sort_asc.png" : "sort_desc.png") +
					"\" alt=\"" + (direction == ItemContext.SortDirection.ASC ? "▼" : "▲") + "\">";
		}
		return image;
	}

	String getNextRowClass()
	{
		evenodd = 1-evenodd;
		return rowclass[evenodd];
	}
	
	void addHidden(String name, String value)
	{
		hiddenForm.append("<input type=\"hidden\" name=\"").append(name).append("\" value=\"");
		if (value != null) hiddenForm.append(value);
		hiddenForm.append("\">\n");
	}
	
	int numVisibleColumns()
	{
		return visibleColumns.size();
	}
	int getNumFilteredColumns()
	{
		return numFilteredColumns;
	}
	
	void setNumColumns(int numColumns)
	{
		if (numColumns > numHeaderColumns)
		{
			numHeaderColumns = numColumns;
		}
	}
	int numHeaderColumns()
	{
		return numHeaderColumns;
	}
	
	@SuppressWarnings("rawtypes")
	List<ListColumnAction> getExtensionColumns()
	{
		return extensionColumns;
	}
	
	/**
		Can header columns be dragged for re-ordering?
		If {@link #setDragcolumns(Boolean)} has been set
		we use that value, otherwise we only allow it
		if the "action" attribute has been set.
		@since 3.8
	*/
	boolean allowColumnDrag()
	{
		return dragColumns != null ? dragColumns : getAction() != null && getItem() != null;
	}
	
	/*
		Display methods
	*/
	@Override
	public int doStartTag()
		throws JspException
	{
		page = (Page)findAncestorWithClass(this, Page.class);
		if (page == null) throw new JspTagException("Tag <tbl:table> must be inside a <base:page> tag");

		// Always start at row 0
		evenodd = 0;
		numHeaderColumns = 0;
		numFilteredColumns = 0;
		columnContent = new HashMap<String, String>();
		columnSubtitle = new HashMap<String, String>();
		columnFilter = new HashMap<String, String>();
		columnFormatter = new HashMap<String, Formatter<?>>();
		hiddenFilteredColumns = new LinkedList<String>();
		jsonColumnDefs = new JSONArray();
		if ((columns == null) || columns.equals("all"))
		{
			allVisible = true;
			visibleColumns = new LinkedHashSet<String>();
		}
		else
		{
			allVisible = false;
			visibleColumns = new LinkedHashSet<String>( Arrays.asList(columns.split(",") ));
			visibleColumns.remove("-"); // "-" represents the required columns
		}
		definedColumns = new HashSet<String>();
		extensionColumns = null;

		hiddenForm = new StringBuilder();
		addHidden("ID", pageContext.getRequest().getParameter("ID"));
		addHidden("cmd", "UpdateContext");
		addHidden("columns", getColumns());
		addHidden("sortby", getSortby());
		addHidden("direction", getDirection().name());
		if (hasStickyHeaders()) addHidden("sticky_column", getStickyCol());
		
		if (sc != null && itemType != null)
		{
			ItemContext cc = sc.getCurrentContext(itemType, subcontext == null ? "" : subcontext);
			Object exclude = cc.getObject("exclude");
			if (exclude instanceof String)
			{
				addHidden("exclude", (String)exclude);
			}
		}
		
		// Submit to same page by default
		String action = getAction();
		if (action == null) action = ((HttpServletRequest)pageContext.getRequest()).getRequestURI();
		StringBuilder sb = new StringBuilder();
		sb.append("<div ");
		addIdAndStyles(sb, "auto-init", hasStickyHeaders() ? "sticky-headers" : null);
		sb.append(" data-auto-init=\"table\"");
		addDynamicAttributes(sb);
		if (itemType != null) sb.append(" data-item-type=\"").append(itemType.name()).append("\"");
		if (subcontext != null) sb.append(" data-subcontext=\"").append(subcontext).append("\"");
		sb.append(">\n");
		sb.append("<form name=\"").append(getId()).append("\"");
		if (action != null) sb.append(" action=\"").append(action).append("\"");
		sb.append(" method=\"post\">\n");
		sb.append(hiddenForm);
		hiddenForm = new StringBuilder();
		if (getTitle() != null) sb.append("<h3 class=\"docked\">").append(getTitle()).append("</h3>\n");

		try
		{
			pageContext.getOut().print(sb.toString());
		}
		catch (Exception ex)
		{
			throw new JspTagException(ex.getMessage());
		}
		return EVAL_BODY_INCLUDE;
	}

	@Override
	public int doEndTag()
		throws JspException
	{
		StringBuilder sb = new StringBuilder();
		sb.append(hiddenForm);
		sb.append("</form>\n");
		sb.append("<div id=\"").append(getId()).append(".table-data\" class=\"datacontainer\"");
		sb.append(" data-column-defs=\"").append(HTML.encodeTags(jsonColumnDefs.toJSONString())).append("\"");
		sb.append("></div>\n");
		sb.append("</div>\n");
		try
		{
			pageContext.getOut().print(sb.toString());
		}
		catch (Exception ex)
		{
			throw new JspTagException(ex.getMessage());
		}
		return EVAL_PAGE;
	}

}

