/*
	$Id: ArrayBatch.java 7381 2017-05-22 12:35:27Z nicklas $

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.ArrayBatchData;
import net.sf.basedb.core.data.HardwareData;
import net.sf.basedb.core.data.ProtocolData;
import net.sf.basedb.core.hibernate.TypeWrapper;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;

import java.util.Date;
import java.util.Set;
import java.util.HashSet;

/**
	This class represents the making of a new batch of microarray
	slides of a particular design.

	@author Enell
	@version 2.0
*/
public class ArrayBatch 
	extends AnnotatedItem
	implements Registered
{
	/**
	 The type of item represented by this class.
	 @see Item#ARRAYBATCH
	 @see #getType()
	 */
	public static final Item TYPE = Item.ARRAYBATCH;

	/**
		Create a new <code>ArrayBatch</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param arrayDesign The <code>ArrayDesign</code> used for the batch
		@return The new <code>ArrayBatch</code> item
		@throws PermissionDeniedException If the logged in user doesn't
			have create permission for this type
			of item, or use permission for the design
		@throws BaseException If there is another error
	*/
	public static ArrayBatch getNew(DbControl dc, ArrayDesign arrayDesign) 
		throws BaseException
	{
		ArrayBatch ab = dc.newItem(ArrayBatch.class);
		ab.setName("New array batch");
		ab.setArrayDesign(arrayDesign);
		ab.getData().setEntryDate(new Date());
		return ab;
	}

	/**
	 Get a <code>ArrayBatch</code> object when you know the ID.

	 @param dc The <code>DbControl</code> which will be used for
	 	permission checking and database access.
	 @param id The ID of the item to load
	 @return The <code>ArrayBatch</code> item
	 @throws ItemNotFoundException If an item with the specified ID is
	 	not found
	 @throws PermissionDeniedException If the logged in user doesn't
	 	have read permission to the item
	 @throws BaseException If there is another error
	 */
	public static ArrayBatch getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		ArrayBatch ab = dc.loadItem(ArrayBatch.class, id);
		if (ab == null) throw new ItemNotFoundException("ArrayBatch[id=" + id + "]");
		return ab;
	}

	/**
		Get a {@link ItemQuery} that returns array batches.
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<ArrayBatch> getQuery()
	{
		return new ItemQuery<ArrayBatch>(ArrayBatch.class);
	}
	
	ArrayBatch(ArrayBatchData annotatedData)
	{
		super(annotatedData);
	}
	
	@Override
	ArrayBatchData getData()
	{
		return (ArrayBatchData)getBasicData();
	}

	/*
	 From the Identifiable interface
	 -------------------------------------------
 	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Annotatable interface
		-------------------------------------------
	*/
	/**
		Get the array design.
	*/
	@Override
	public Set<Annotatable> getAnnotatableParents()
		throws BaseException
	{
		Set<Annotatable> annotatable = new HashSet<Annotatable>();
		ArrayBatchData data = getData();
		
		addAnnotatableParents(getDbControl(), annotatable, 
			data.getArrayDesign(),
			data.getProtocol(),
			data.getPrintRobot()
		);
		return annotatable;
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	@Override
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	@Override
	public void setEntryDate(Date entryDate)
	{
		if (isInDatabase()) 
		{
			throw new PermissionDeniedException("Not allowed to change entryDate of " + this);
		}
		getData().setEntryDate(DateUtil.setDateOrToday(entryDate));
	}
	// -------------------------------------------

	/*
	 From the BasicItem class
	 -------------------------------------------
	 */
	@Override
	public boolean isUsed() 
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.query.Query<Long> query = HibernateUtil.getPredefinedQuery(session, 
			"GET_ARRAYSLIDES_FOR_BATCH", Long.class, "count(*)");
		/*
			SELECT {1}
			FROM ArraySlideData a
			WHERE a.arrayBatch = :arrayBatch
		*/
		query.setParameter("arrayBatch", getId(), TypeWrapper.H_INTEGER);
		return HibernateUtil.loadData(query) > 0 || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link ArraySlide}:s in this batch
		</ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.query.Query<Integer> query = HibernateUtil.getPredefinedQuery(session, 
			"GET_ARRAYSLIDES_FOR_BATCH", Integer.class, "a.id");
			/*
			SELECT {1}
			FROM ArraySlideData a
			WHERE a.arrayBatch = :arrayBatch
			*/
		query.setParameter("arrayBatch", getId(), TypeWrapper.H_INTEGER);
		addUsingItems(using, Item.ARRAYSLIDE, query);
		return using;
	}
	/**
		Set print robot and protocol from project default settings.
		@since 2.9
	*/
	@Override
	void setProjectDefaults(Project activeProject)
		throws BaseException
	{
		super.setProjectDefaults(activeProject);
		if (!robotHasBeenSet)
		{
			HardwareData robot = 
				(HardwareData)activeProject.findDefaultData(getDbControl(), Hardware.PRINT_ROBOT, false);
			if (robot != null) getData().setPrintRobot(robot);
			robotHasBeenSet = true;
		}
		if (!protocolHasBeenSet)
		{
			ProtocolData protocol = 
				(ProtocolData)activeProject.findDefaultData(getDbControl(), Protocol.PRINTING, false);
			if (protocol != null) getData().setProtocol(protocol);
			protocolHasBeenSet = true;
		}
	}
	// -------------------------------------------

	/**
		Get the {@link ArrayDesign} item that was used for this batch.
		@throws PermissionDeniedException If
			the logged in user doesn't have read
			permission to the item
		@throws BaseException If there is an error
		@return The <code>ArrayDesign</code> item
	*/
	public ArrayDesign getArrayDesign() 
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(ArrayDesign.class, getData().getArrayDesign());
	}

	/**
		Set the {@link ArrayDesign} used in this ArrayBatch.
		@param arrayDesign the arrayDesign to be set. Cannot be null.
		@throws InvalidDataException if arrayDesign is null.
		@throws PermissionDeniedException If the
			logged in user doesn't have use
			permission for the array design.
	*/
	private void setArrayDesign(ArrayDesign arrayDesign) 
		throws InvalidDataException, PermissionDeniedException
	{
		if (arrayDesign == null) throw new InvalidDataException("An arraybatch must have an arraydesign.");
		arrayDesign.checkPermission(Permission.USE);
		getData().setArrayDesign(arrayDesign.getData());
	}
	
	/**
		Get the {@link Protocol} used for printing the arrays.
		@return The <code>Protocol</code> item
		@throws PermissionDeniedException If
			the logged in user doesn't have read
			permission to the item
		@throws BaseException If there is an error
	*/
	@Override
	public Protocol getProtocol() 
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Protocol.class, getData().getProtocol());
	}

	private boolean protocolHasBeenSet = false;
	/**
		Set the {@link Protocol} used for printing the arrays.
		@param protocol A <code>Protocol</code> item of the {@link
			Protocol#PRINTING} protocol type
		@throws PermissionDeniedException If the
			logged in user doesn't have write 
			permission on this array batch or use
			permission for the protocol
	*/
	public void setProtocol(Protocol protocol)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		if (protocol != null) protocol.checkPermission(Permission.USE);
		protocolHasBeenSet = true;
		getData().setProtocol(protocol == null ? null : protocol.getData());
	}
	
	/**
		Get the {@link Hardware print robot} used for printing the arrays.
		@return A <code>Hardware</code> item of type
			{@link Hardware#PRINT_ROBOT}
		@throws PermissionDeniedException If
			the logged in user doesn't have read
			permission to the item
		@throws BaseException If there is an error
	*/
	public Hardware getPrintRobot() 
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Hardware.class, getData().getPrintRobot());
	}
	private boolean robotHasBeenSet = false;
	/**
		Set the {@link Hardware print robot} used for this array batch.
		@param printRobot A <code>Hardware</code> item of type
			{@link Hardware#PRINT_ROBOT }
		@throws PermissionDeniedException If the
			logged in user doesn't have write 
			permission on this array batch or use
			permission for the hardware
	*/
	public void setPrintRobot(Hardware printRobot)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		if (printRobot != null) printRobot.checkPermission(Permission.USE);
		robotHasBeenSet = true;
		getData().setPrintRobot(printRobot == null ? null : printRobot.getData());
	}
	/**
		Create a new {@link ArraySlide} item for this batch.
		@return The new <code>ArraySlide</code> item
		@throws PermissionDeniedException If
			the logged in user doesn't have use
			permission for this batch
		@throws BaseException If there is another
			error
	*/
	public ArraySlide newArraySlide()
		throws PermissionDeniedException, BaseException
	{
		ArraySlide as = ArraySlide.getNew(getDbControl(), this);
		return as;
	}


	/**
		Get a query that returns the slides in this batch.

		@return An <code>ItemQuery</code> object
	*/
	public ItemQuery<ArraySlide> getArraySlides()
	{
		ItemQuery<ArraySlide> query = ArraySlide.getQuery();
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("arrayBatch"), 
				Hql.entity(this)
			)
		);
		return query;
	}
}
