/*
	$Id: BioAssaySet.java 7381 2017-05-22 12:35:27Z nicklas $
	
	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.AnnotationSetData;
import net.sf.basedb.core.data.BioAssaySetData;
import net.sf.basedb.core.data.RawBioAssayData;
import net.sf.basedb.core.data.BioAssayData;
import net.sf.basedb.core.data.SharedData;
import net.sf.basedb.core.data.TransformationData;
import net.sf.basedb.core.hibernate.TypeWrapper;
import net.sf.basedb.core.data.DataCubeData;
import net.sf.basedb.core.data.DataCubeLayerData;
import net.sf.basedb.core.data.DataCubeFilterData;
import net.sf.basedb.core.data.DataCubeExtraValueData;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.Aggregations;
import net.sf.basedb.core.query.SqlResult;
import net.sf.basedb.core.query.Selects;
import net.sf.basedb.core.query.JoinType;

import java.util.Collections;
import java.util.Collection;
import java.util.Set;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.HashMap;
import java.sql.SQLException;

/**
	A bioassayset represents the current intensities of the raw data
	after some {@link Transformation} has been applied to it. The first 
	transformation is called a root transformation. A bioassayset created
	by a root transformation is called a root bioassayset. 
	<p>
	A bioassayset is created by the {@link Transformation#newProduct(String, String, boolean)}
	method. Bioassaysets can only be created if the transformation hasn't been
	committed to the database.
	
	@author Nicklas
	@version 2.0
	@see Transformation#newProduct(String, String, boolean)
	@base.modified $Date: 2017-05-22 14:35:27 +0200 (Mon, 22 May 2017) $
*/
public class BioAssaySet
	extends ChildItem
	implements Nameable, Removable, Annotatable, 
		FileStoreEnabled
{
	/**
		The type of item represented by this class.
		@see Item#BIOASSAYSET
		@see #getType()
	*/
	public static final Item TYPE = Item.BIOASSAYSET;

	/**
		Create a new <code>BioAssaySet</code> item.

		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@param transformation The transformation that created the bioassayset 
		@param layer The data cube layer where the bioassayset should store
			it's data
		@return The new <code>BioAssaySet</code> item
		@throws PermissionDeniedException If the logged in user
			doesn't have write permission for the experiment the 
			transformation belongs to
		@throws InvalidDataException If the transformation is null
		@throws BaseException If there is an error
		@see Transformation#newProduct(String, String, boolean)
	*/
	static BioAssaySet getNew(DbControl dc, Transformation transformation, DataCubeLayer layer)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		BioAssaySet bas = dc.newItem(BioAssaySet.class);
		bas.setName("New bioassayset");
		bas.setTransformation(transformation);
		bas.setExperiment(transformation.getExperiment());
		bas.setDataCubeLayer(layer);
		BioAssaySet source = transformation.getSource();
		if (source != null && layer.equals(source.getDataCubeLayer()))
		{
			bas.setDataCubeFilter(source.getDataCubeFilter());
			bas.setIntensityTransform(source.getIntensityTransform());
		}
		else
		{
			bas.setIntensityTransform(IntensityTransform.NONE);
		}
		return bas;
	}

	/**
		Get a <code>BioAssaySet</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>BioAssaySet</code> item
		@throws ItemNotFoundException If an item with the specified id 
			is not found
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static BioAssaySet getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		BioAssaySet bas = dc.loadItem(BioAssaySet.class, id);
		if (bas == null) throw new ItemNotFoundException("BioAssaySet[id=" + id + "]");
		return bas;
	}
	
	/**
		Get a query configured to retrieve bioassaysets for
		a given experiment.
		@param experiment The experiment to retreive bioassaysets for, null
			is not allowed
		@return An {@link ItemQuery} object
	 	@throws InvalidDataException If the parameter is null.
	*/
	public static ItemQuery<BioAssaySet> getQuery(Experiment experiment)
		throws InvalidDataException
	{
		if (experiment == null) throw new InvalidUseOfNullException("experiment");
		ItemQuery<BioAssaySet> query = new ItemQuery<BioAssaySet>(BioAssaySet.class, null);
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("experiment"),
				Hql.entity(experiment)
			)
		);
		return query;
	}

	/**
		Get a query configured to retrieve bioassaysets 
		created by a given transformation.
		@param transformation The transformation to retreive bioassaysets for, null
			is not allowed
		@return An {@link ItemQuery} object
	 	@throws InvalidDataException If the transformation parameter is null. 
	*/
	public static ItemQuery<BioAssaySet> getQuery(Transformation transformation)
		throws InvalidDataException
	{
		if (transformation == null) throw new InvalidUseOfNullException("transformation");
		ItemQuery<BioAssaySet> query = new ItemQuery<BioAssaySet>(BioAssaySet.class, null);
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("transformation"),
				Hql.entity(transformation)
			)
		);
		return query;
	}

	/**
	 	Creates a new experiment item from the given data.
		@param data the data
	*/
	BioAssaySet(BioAssaySetData data)
	{
		super(data);
	}

	@Override
	BioAssaySetData getData()
	{
		return (BioAssaySetData)getBasicData();
	}
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	@Override
	public String getName()
	{
		return getData().getName();
	}
	@Override
	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setName(getData(), name);
	}
	@Override
	public String getDescription()
	{
		return getData().getDescription();
	}
	@Override
	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------
	/*
		From the Removable interface
		-------------------------------------------
	*/
	@Override
	public boolean isRemoved()
	{
		return getData().getRemovedBy() != null;
	}
	@Override
	public void setRemoved(boolean removed)
		throws PermissionDeniedException
	{
		checkPermission(removed ? Permission.DELETE : Permission.WRITE);
		getData().setRemovedBy(removed ? getSessionControl().getLoggedInUserId() : null);
	}
	@Override
	public User getRemovedBy()
		throws PermissionDeniedException, ItemNotFoundException 
	{
		Integer userId = getData().getRemovedBy();
		return userId == null ? null : User.getById(getDbControl(), userId);
	}
	// -------------------------------------------
	/*
		From the Annotatable interface
		-------------------------------------------
	*/
	/**
		@since 2.2
	*/
	@Override
	public AnnotationSet getAnnotationSet()
		throws PermissionDeniedException, BaseException
	{
		checkBatchAnnotatableUsage(true, false);
		DbControl dc = getDbControl();
		AnnotationSet as = dc.getItem(AnnotationSet.class, getData().getAnnotationSet(), this);
		if (as == null)
		{
			as = AnnotationSet.getNew(dc, this);
			getData().setAnnotationSet(as.getData());
			dc.saveItemIf(this, as, false);
		}
		return as;
	}
	/**
		@since 2.2
	*/
	@Override
	public boolean isAnnotated()
	{
		return getData().getAnnotationSet() != null;
	}
	/**
		@since 2.2
	*/
	@Override
	public void removeAnnotations()
		throws PermissionDeniedException, BaseException
	{
		checkPermission(Permission.WRITE);
		AnnotationSetData as = getData().getAnnotationSet();
		if (as != null)
		{
			HibernateUtil.deleteData(getDbControl().getHibernateSession(), as);
			getData().setAnnotationSet(null);
		}
	}
	/**
		Default implementation returns null. Should be overriden by subclasses
		that has protocols.
		@return Always null
		@since 2.2
	*/
	@Override
	public Protocol getProtocol()
	{
		return null;
	}
	/**
		Get the parent bioassay set (if any) and the experiment.
		@since 2.2
	*/
	@Override
	public Set<Annotatable> getAnnotatableParents()
		throws BaseException
	{
		Set<Annotatable> annotatable = new HashSet<Annotatable>();
		DbControl dc = getDbControl();
		
		BioAssaySetData data = getData();
		addAnnotatableParents(dc, annotatable, 
			data.getExperiment()
		);
		TransformationData tData = data.getTransformation();
		if (tData != null)
		{
			addAnnotatableParents(dc, annotatable, 
				tData.getSource()
			);
		}

		return annotatable;
	}
	// -------------------------------------------
	/*
		From the FileStoreEnabled interface
		-------------------------------------------
	*/
	/**
		@since 2.8
	*/
	@Override
	public FileSet getFileSet()
	{
		DbControl dc = getDbControl();
		FileSet fs = dc.getItem(FileSet.class, getData().getFileSet(), this);
		if (fs == null)
		{
			fs = FileSet.getNew(dc, this);
			getData().setFileSet(fs.getData());
			dc.saveItemIf(this, fs, false);
		}
		return fs;
	}
	/**
		@since 2.8
	*/
	@Override
	public boolean hasFileSet()
	{
		return getData().getFileSet() != null;
	}
	/**
		@return Always null
		@since 2.8
	*/
	@Override
	public Platform getPlatform()
	{
		return null;
	}
	/**
		@return Always null
		@since 2.8
	*/
	@Override
	public PlatformVariant getVariant()
	{
		return null;
	}
	/**
		Get the file set from the parent bioassay set if it exists.
		@since 2.8
	*/
	@Override
	public Collection<FileSet> getParentFileSets()
	{
		BioAssaySet source = getTransformation().getSource();
		Collection<FileSet> parents = null;
		if (source != null && source.hasFileSet())
		{
			parents = Collections.singleton(source.getFileSet());
		}
		return parents;
	}
	// -------------------------------------------
	/*
		From BasicItem
		-------------------------------------------
	*/
	/**
		On creation, count the number of spots and reporters in the bioassays
		attached to this bioassayset. On deletion, delete all analysed
		data in this bioassayset and all child bioassaysets.
	*/
	@Override
	void onBeforeCommit(Transactional.Action action)
		throws BaseException
	{
		super.onBeforeCommit(action);
		if (action == Transactional.Action.CREATE)
		{
			checkAndCreateTables();
			countSpotsAndReporters();
		}
		else if (action == Transactional.Action.DELETE)
		{
			deleteAnalysedData();
		}
	}
	// -------------------------------------------
	/*
		From the ChildItem class
		-------------------------------------------
	*/
	/**
		Get the experiment.
	*/
	@Override
	SharedData getSharedParent()
	{
		return getData().getExperiment();
	}
	/**
		USE permission is enough to be able to manage bioassay sets.
	*/
	@Override
	Permission getPermissionForWriteDeleteAndCreate()
	{
		return Permission.USE;
	}
	
	// -------------------------------------------

	/**
		Get the experiment this bioassayset belongs to.
		@return An <code>Experiment</code> object
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission to the experiment
		@throws BaseException If there is another error
	*/
	public Experiment getExperiment()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Experiment.class, getData().getExperiment());
	}
	
	/**
		Set the experiment. This cannot be changed after the bioassayset
		has been created.
	*/
	private void setExperiment(Experiment experiment)
		throws PermissionDeniedException, InvalidDataException
	{
		if (experiment == null) throw new InvalidUseOfNullException("experiment");
		experiment.checkPermission(Permission.USE);
		getData().setExperiment(experiment.getData());
	}
	
	/**
		Get the transformation that created this bioassayset.
		@return A <code>Transformation</code> object
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission to the experiment
		@throws BaseException If there is another error
	*/
	public Transformation getTransformation()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Transformation.class, getData().getTransformation());
	}
	
	/**
		Set the transformation that created this bioassayset. This cannot be 
		changed after the bioassayset has been created.
	*/
	private void setTransformation(Transformation transformation)
		throws PermissionDeniedException, InvalidDataException
	{
		if (transformation == null) throw new InvalidUseOfNullException("transformation");
		getData().setTransformation(transformation.getData());
	}
	
	/**
		Get the raw data type this experiment uses for the raw data.
		The raw data type cannot be changed once the experiment has been created.
		@return A <code>RawDataType</code> object
	*/
	public RawDataType getRawDataType()
	{
		return RawDataTypes.getSafeRawDataType(getData().getExperiment().getRawDataType());
	}
	
	/**
		Get the number of unique reporters among all bioassays in this bioassayset.
	*/
	public int getNumReporters()
	{
		return getData().getNumReporters();
	}
	
	/**
		Get the number of spots in all bioassays in this bioassayset.
	*/
	public int getNumSpots()
	{
		return getData().getNumSpots();
	}
	
	/**
		Get the number of unique reporters for data that is stored in files.
		@since 2.8
	*/
	public int getNumFileReporters()
	{
		return getData().getNumFileReporters();
	}
	
	/**
		Set the number of unique reporters for data that is stored in files.
		@param numReporters The number of reporters
		@since 2.8
	*/
	public void setNumFileReporters(int numReporters)
	{
		checkPermission(Permission.WRITE);
		getData().setNumFileReporters(numReporters);
	}
	
	/**
		Get the number of spots that are stored in files.
		@since 2.8
	*/
	public int getNumFileSpots()
	{
		return getData().getNumFileSpots();
	}
	
	/**
		Set the number of spots that are stored in files.
		@param numSpots The number of spots
		@since 2.8
	*/
	public void setNumFileSpots(int numSpots)
	{
		checkPermission(Permission.WRITE);
		getData().setNumFileSpots(numSpots);
	}
	
	/**
		Get the number of mappings to raw data spots for the spot with
		the most number of mappings. For most bioassay sets this value is 
		probably 1. After a merge, the count can be higher which may require some
		plugins to behave differently when using values from the raw data table.
		@since 2.2
	*/
	public int getMaxRawMappingsForSpot()
	{
		return getData().getDataCubeLayer().getDataCube().getMaxRawMappingsForSpot();
	}
	
	/**
		Get information about in what way the spot intensities has been
		transformed before they were stored in the database.
		@since 2.12
	*/
	public IntensityTransform getIntensityTransform()
	{
		return IntensityTransform.fromValue(getData().getIntensityTransform());
	}
	/**
		Sets the transform used on the spot intensities before they were stored
		in the database. NOTE! Calling this method doesn't change the
		data in the database!
		@param transform The transform that was used to store the data
		@since 2.12
	*/
	public void setIntensityTransform(IntensityTransform transform)
	{
		checkPermission(Permission.WRITE);
		if (transform == null) throw new InvalidUseOfNullException("transform");
		getData().setIntensityTransform(transform.getValue());
	}
	
	/**
		Get the number of the data cube this bioassayset stores it's data in.
		@see #getDataCubeLayerNo()
		@see #getDataCubeFilterNo()
	*/
	public short getDataCubeNo()
	{
		return getData().getDataCubeLayer().getDataCube().getCubeNo();
	}
	
	/**
		Get the layer coordinate in the data cube that this bioassayset stores it's data in.
		@see #getDataCubeNo()
		@see #getDataCubeFilterNo()
	*/
	public short getDataCubeLayerNo()
	{
		return getData().getDataCubeLayer().getLayerNo();
	}
	
	/**
		Get the number of the filter that this bioassayset is using, or 0 if
		no filter is used.
		@return The number of the filter or 0 if no filter is used
		@see #getDataCubeLayerNo()
		@see #getDataCubeNo()
	*/
	public short getDataCubeFilterNo()
	{
		return getData().getDataCubeFilter() == null ? 0 : getData().getDataCubeFilter().getFilterNo();
	}
	
	/**
		Get the virtual database this bioassayset stores it's data in.
	*/
	VirtualDb getVirtualDb()
	{
		return getDbControl().getItem(VirtualDb.class, getData().getDataCubeLayer().getDataCube().getVirtualDb());
	}
	
	/**
		Get the data cube this biossayset stores it's data in.
	*/
	DataCube getDataCube()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(DataCube.class, getData().getDataCubeLayer().getDataCube());
	}

	/**
		Get the data cube layer this biossayset stores it's data in.
	*/
	DataCubeLayer getDataCubeLayer()
	{
		return getDbControl().getItem(DataCubeLayer.class, getData().getDataCubeLayer());
	}
	private void setDataCubeLayer(DataCubeLayer layer)
	{
		getData().setDataCubeLayer(layer.getData());
	}
	
	/**
		Get the data cube filter this biossayset uses to filter spot data.
	*/
	DataCubeFilter getDataCubeFilter()
	{
		return getDbControl().getItem(DataCubeFilter.class, getData().getDataCubeFilter());
	}
	private void setDataCubeFilter(DataCubeFilter filter)
	{
		getData().setDataCubeFilter(filter == null ? null : filter.getData());
	}

	
	/**
		Create a new transformation using this bioassayset as it's
		source bioassayset. 
		@param job The job that executed the transformation
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission for the experiment this bioassayset belongs to
		@throws BaseException If there is another error
	*/
	public Transformation newTransformation(Job job)
		throws PermissionDeniedException, BaseException
	{
		checkPermission(Permission.WRITE);
		Transformation t = Transformation.getNew(getDbControl(), getExperiment(), job);
		t.setSource(this);
		return t;
	}

	/**
		Get a query that return all transformations using this bioassayset
		as a source.
		
		@return An <code>ItemQuery</code> object
		@see Transformation#getQuery(BioAssaySet)
	*/
	public ItemQuery<Transformation> getTransformations()
	{
		return Transformation.getQuery(this);
	}

	/**
		Create a new <code>BioAssay</code> with links to raw
		bioassays. This type of bioassay can only be created if the
		transformation that created it is a root transformation. New
		bioassays can only be created if the bioassayset hasn't been 
		committed to the database.
		<p>
		The raw bioassay collection should contain only those raw bioassays 
		whose spots have been used to calculate the intensities stored in this
		bioassay. 
		
		@param rawParents A collection containing the raw bioassays that
			are the parents of this bioassay
		@return The new <code>BioAssay</code> item
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission for the experiment of if the bioassayset has
			been committed to the database
		@throws InvalidDataException If this bioassayset is not a product
			of a root transformation or if any of the raw bioassay parents
			isn't a member of the transformation
		@throws BaseException If something else goes wrong.
		@see #newBioAssay(BioAssay)
		@see #newBioAssay(Collection)
	*/
	public BioAssay newRootBioAssay(Collection<RawBioAssay> rawParents)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		checkPermission(Permission.WRITE);
		if (isInDatabase())
		{
			throw new PermissionDeniedException("'This bioassayset has already been committed: "+this);
		}
		if (getTransformation().getSource() != null)
		{
			throw new InvalidDataException("This bioassayset is not the product of a root transformation: "+this);
		}
		BioAssay ba = BioAssay.getNew(getDbControl(), this, getDataCube().newColumn());

		// Add rawParents to the new BioAssay. check that the are part of the transformation
		Set<RawBioAssayData> rawBioAssays = ba.getData().getRawParents();
		Set<RawBioAssayData> transformationRawSources = getData().getTransformation().getRawSources();
		String baName = null;
		for (RawBioAssay rba : rawParents)
		{
			if (!transformationRawSources.contains(rba.getData()))
			{
				throw new InvalidDataException("Raw bioassay "+rba+" is not part of the transformation: "+getTransformation());
			}
			rawBioAssays.add(rba.getData());
			baName = rba.getName();
		}
		if (rawParents.size() == 1) ba.setName(baName);
		return ba;
	}
	
	/**
		Create a new <code>BioAssay</code> in this bioassayset using a 
		single parent. The data cube column coordinate of the new bioassay is set to
		the same as that of the parent. The bioassayset of the parent bioassay
		must match the bioassayset set of the source of the transformation and this
		bioassayset must use the same data cube as the source bioassayset.
		The new bioassay is automatically linked to the same raw bioassays
		as the parent bioassay.
		<p> 
		New bioassays can only be created if the bioassayset hasn't been 
		committed to the database.
		
		@param parent The parent bioassay
		@return The new bioassay
		@throws InvalidDataException If the parent is null or it's bioassayset is not
			the same as the source bioassayset of the transformation or if the data cube
			of the source is not the same as the data cube for this bioassayset
		@throws PermissionDeniedException If the bioassayset has already been committed
			to the database
		@see #newBioAssay(Collection)
		@see #newRootBioAssay(Collection)
	*/
	public BioAssay newBioAssay(BioAssay parent)
		throws InvalidDataException, PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		if (isInDatabase())
		{
			throw new PermissionDeniedException("Bioassayset has already been committed: "+this);
		}
		if (parent == null) throw new InvalidUseOfNullException("parent");
		
		// Validate that the parent bioassay is part of the source bioassayset
		BioAssaySetData source = getData().getTransformation().getSource();
		if (!parent.getData().getBioAssaySet().equals(source))
		{
			throw new InvalidDataException("Bioassay "+parent+" is not part of the source bioassayset: "+
				getTransformation().getSource());
		}
		
		// Validate that the source is using the same data cube as this bioassayset
		if (!getData().getDataCubeLayer().getDataCube().equals(source.getDataCubeLayer().getDataCube()))
		{
			throw new InvalidDataException("The source datacube "+getTransformation().getSource().getDataCube()+
				" is not the same as this bioassaysets datacube: "+getDataCube());
		}
		
		// Everything is ok, create the new bioassay
		return newBioAssay(parent.getDataCubeColumn(), Collections.singleton(parent));
	}
	
	/**
		Create a new <code>BioAssay</code> with links to multiple parent
		bioassays. This type of bioassay can only be created if the
		transformation that created it isn't a root transformation, since a root
		transformation can't have a parent bioassayset. This type of
		bioassay must be created if this bioassayset uses a different
		data cube than the bioassayset that is the source of the transformation.
		The new bioassay is automatically linked to the same raw bioassays
		as the parent bioassay.
		<p> 
		New bioassays can only be created if the bioassayset hasn't been 
		committed to the database.
		
		@param parents A collection containing the bioassays that
			are the parents of this bioassay
		@return The new <code>BioAssay</code> item
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission for the experiment of if the bioassayset has
			been committed to the database
		@throws InvalidDataException If this bioassayset is the product
			of a root transformation or if any of the bioassay parents
			isn't a member source bioassayset of the transformation
		@see #newBioAssay(BioAssay)
		@see #newRootBioAssay(Collection)
	*/
	public BioAssay newBioAssay(Collection<BioAssay> parents)
	{
		checkPermission(Permission.WRITE);
		if (isInDatabase())
		{
			throw new PermissionDeniedException("BioAssaySet has already been committed: "+this);
		}
		if (parents == null || parents.size() == 0)
		{
			throw new InvalidUseOfNullException("parents");
		}
		
		// Validate that the all parent bioassays are part of the source bioassayset
		BioAssaySetData source = getData().getTransformation().getSource();
		for (BioAssay parent : parents)
		{
			if (!parent.getData().getBioAssaySet().equals(source))
			{
				throw new InvalidDataException("Bioassay "+parent+" is not part of the source bioassayset: "+getTransformation().getSource());
			}
		}
		
		// Validate the source is using a different datacube than this bioassayset
		if (getData().getDataCubeLayer().getDataCube().equals(source.getDataCubeLayer().getDataCube()))
		{
			throw new InvalidDataException("The source datacube "+getTransformation().getSource().getDataCube()+
				" is the same as this bioassaysets datacube: "+getDataCube());
		}

		// Everything is ok, create the new bioassay
		return newBioAssay(getDataCube().newColumn(), parents);
	}

	/**
		Create a new bioassay and link to the parents. Assume that all
		neccessary checks has been made by calling code. If the collections contains
		only a single parent, the name will be copied to the child.
	*/
	private BioAssay newBioAssay(DataCubeColumn column, Collection<BioAssay> parents)
	{
		BioAssay ba = BioAssay.getNew(getDbControl(), this, column);
		int numParents = parents.size();
		
		// Add parents and raw parents to the new BioAssay
		Set<BioAssayData> baParents = ba.getData().getParents();
		Set<RawBioAssayData> rawParents = ba.getData().getRawParents();
		for (BioAssay parent : parents)
		{
			if (numParents == 1) ba.setName(parent.getName());
			baParents.add(parent.getData());
			rawParents.addAll(parent.getData().getRawParents());
		}
		return ba;
	}
	
	/**
		Get a query returning the bioassays that are part of this
		bioassayset.
		@return An <code>ItemQuery</code> object
	*/
	public ItemQuery<BioAssay> getBioAssays()
	{
		return BioAssay.getQuery(this);
	}

	/**
		Get a query returning the extra values that are part of this
		bioassayset. To add extra values use one of the methods that
		returns a batcher.
		
		@return An <code>ItemQuery</code> object
		@see #getSpotExtraValueBatcher(Class, ExtraValueType, Job)
		@see #getPositionExtraValueBatcher(Class, ExtraValueType, Job)
	*/
	public ItemQuery<ExtraValue> getExtraValues()
	{
		return ExtraValue.getQuery(this);
	}

	/**
		Get a query returning all extra values types that are used in
		this bioassay set.
		@return An <code>ItemQuery</code> object
		@since 2.15
		@see #getExtraValues()
		@see #getExtraValue(ExtraValueType)
	*/
	public ItemQuery<ExtraValueType> getExtraValueTypes()
	{
		ItemQuery<ExtraValueType> query = ExtraValueType.getQuery();
		query.joinPermanent(Hql.innerJoin("extraValues", Item.EXTRAVALUE.getAlias()));
		query.restrictPermanent(
				Restrictions.eq(
					Hql.property(Item.EXTRAVALUETYPE.getAlias(), "bioAssaySet"),
					Hql.entity(this)
				)
			);
		return query;
	}
	
	/**
		Get the extra value of the specified type in this bioassayset. To add 
		extra values use one of the methods that returns a batcher.
	 	@param extraValueType For which extra value should be returned.
		@return An <code>ExtraValue</code> item, or null if no extra value
			of the specified type has been created  for this bioassayset
		@throws BaseException If something fails when getting the extra value.
		@see #getSpotExtraValueBatcher(Class, ExtraValueType, Job)
		@see #getPositionExtraValueBatcher(Class, ExtraValueType, Job)
	*/
	public ExtraValue getExtraValue(ExtraValueType extraValueType)
		throws BaseException
	{
		return getDbControl().getItem(ExtraValue.class, getData().getExtraValues().get(extraValueType.getData()));
	}
	
	/**
		Get a batcher that links position coordinates to reporters 
		for the data cube this bioassayset uses. The links must be
		created before the data cube is committed to the database. This means
		that only the first bioassayset in a data cube can create the links
		to reporters. To insert extra values (which can be done at any time)
		based on the position use the {@link 
		#getPositionExtraValueBatcher(Class, ExtraValueType, Job)} method to get a batcher for
		that.
		
		@return A {@link PositionBatcher} object
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission or if the data cube has already been committed
			to the database
		@see #getPositionExtraValueBatcher(Class, ExtraValueType, Job)
	*/
	public PositionBatcher getPositionBatcher()
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		if (getDataCube().isInDatabase())
		{
			throw new PermissionDeniedException("Data cube has already been committed: "+getDataCube());
		}
		return getDataCube().getPositionBatcher();
	}
	
	/**
		Get a batcher used to insert mappings to raw data. The links must be
		created before the data cube is committed to the database. This means
		that only the first bioassayset in a data cube can create the links
		to raw data.
		
		@return A {@link MappingBatcher} object
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission or if the data cube has already been committed
			to the database
	*/
	public MappingBatcher getMappingBatcher()
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		if (getDataCube().isInDatabase())
		{
			throw new PermissionDeniedException("Data cube has already been committed: "+getDataCube());
		}
		return getDataCube().getMappingBatcher();
	}

	/**
		The one and only spot batcher for this bioassayset.
		@see #getSpotBatcher()
	*/
	private SpotBatcher spotBatcher;
	
	/**
		Get a batcher used to insert spot data for this bioassayset. If this bioassayset
		uses the same layer as another bioassayset or if this bioassayset has already been
		committed to the database no more intensities can be inserted. 
		To insert extra values (which can be done at any time) based on the 
		position use the {@link #getSpotExtraValueBatcher(Class, ExtraValueType, Job)} 
		method to get a batcher for that.

		@return A {@link SpotBatcher} object
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission or if the data cube layer has already been committed
			to the database
		@see #getSpotExtraValueBatcher(Class, ExtraValueType, Job)
	*/
	public SpotBatcher getSpotBatcher()
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		if (getDataCubeLayer().isInDatabase())
		{
			throw new PermissionDeniedException("Data cube layer has already been committed: "+getDataCubeLayer());
		}
		if (spotBatcher == null)
		{
			spotBatcher = new SpotBatcher(getDbControl(), this);
		}
		return spotBatcher;
	}
	
	/**
		The one and only filter batcher for this bioassayset.
		@see #getFilterBatcher()
	*/
	private FilterBatcher filterBatcher;
	
	/**
		Get a batcher used to insert the position/column coordinates that remain
		after some filter has been applied to the spot data of a bioassayset. Filter
		information cannot be inserted after the bioassayset has been committed to
		the database. Calling this method and not inserting any position/column coordinates
		will result in a bioassayset with no data.
		
		@return A {@link FilterBatcher} object
		@throws PermissionDeniedException If the bioassayset has already been committed
			to the database or if the logged in user doesn't have write permission
		@see #getSpotBatcher()
	*/
	public FilterBatcher getFilterBatcher()
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		if (isInDatabase())
		{
			throw new PermissionDeniedException("Bioassayset has already been committed: "+this);
		}
		if (filterBatcher == null)
		{
			setDataCubeFilter(getDataCube().newFilter());
			filterBatcher = new FilterBatcher(getDbControl(), this);
		}
		return filterBatcher;
	}
	
	/**
		Cache of already created spot extra value batchers.
		@see #getSpotExtraValueBatcher(Class, ExtraValueType, Job)
	*/
	private Map<ExtraValueType, SpotExtraValueBatcher<?>> spotExtraValueBatchers;
	
	/**
		Get a batcher that is used to insert extra values per spot of the specified type.
		A bioassay set can only have one set of extra values for each type.
		If this method is called more than once in the same transaction with the
		same extra value type parameter, the same batcher is returned. 

		@param clazz The class object for the values, ie. <code>Integer.class</code>,
			<code>Float.class</code> or <code>String.class</code>
		@param extraValueType The definition of the extra value type
		@param job The job that is doing the extra value calculations or null
		@return A <code>SpotExtraValueBatcher</code> object
		@throws PermissionDeniedException If the loggged in user doesn't have write
			permission for the bioassay set
		@throws InvalidDataException If the extra value type is null or has a 
			value type that doesn't match the specified class
		@see #getSpotBatcher()
	*/
	@SuppressWarnings("unchecked")
	public <I> SpotExtraValueBatcher<I> getSpotExtraValueBatcher(Class<I> clazz, 
		ExtraValueType extraValueType, Job job)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (extraValueType == null) throw new InvalidUseOfNullException("extraValueType");
		if (extraValueType.getValueType().getValueClass() != clazz)
		{
			throw new InvalidDataException("The specified class '" + clazz + "' can't be used for extra value type: " + 
				extraValueType.getValueType());
		}

		// Check if a batcher already exists in the cache or not
		if (spotExtraValueBatchers == null) spotExtraValueBatchers = new HashMap<ExtraValueType, SpotExtraValueBatcher<?>>();
		SpotExtraValueBatcher<I> batcher = (SpotExtraValueBatcher<I>)spotExtraValueBatchers.get(extraValueType);
		if (batcher == null)
		{
			if (getData().getExtraValues().containsKey(extraValueType.getData()))
			{
				throw new InvalidDataException("The bioassay set already has extra values of the type: " + extraValueType);
			}

			// Not found, create a new extra value and batcher
			ExtraValue extraValue = ExtraValue.getNew(getDbControl(), this, extraValueType, 
				getDataCube().newExtraValue(extraValueType.getValueType(), ExtraValue.CoordinateType.SPOT), job);
		
			// Saved by Hibernate cascade
			getData().getExtraValues().put(extraValueType.getData(), extraValue.getData());
			batcher = new SpotExtraValueBatcher<I>(getDbControl(), extraValue);
			spotExtraValueBatchers.put(extraValueType, batcher);
		}
		return batcher;
	}
	
	/**
		Cache of already created position extra value batchers.
		@see #getPositionExtraValueBatcher(Class, ExtraValueType, Job)
	*/
	private Map<ExtraValueType, PositionExtraValueBatcher<?>> positionExtraValueBatchers;
	
	/**
		Get a batcher that is used to insert extra values per position of the specified 
		type. A bioassay set can only have one set of extra values for each type.
		If this method is called more than once in the same transaction with the
		same extra value type parameter, the same batcher is returned. 
		
		@param clazz The class object for the values, ie. <code>Integer.class</code>,
			<code>Float.class</code> or <code>String.class</code>
		@param extraValueType The definition of the extra value type
		@param job The job that is doing the extra value calculations or null
		@return A <code>PositionExtraValueBatcher</code> object
		@throws PermissionDeniedException If the loggged in user doesn't have write
			permission for the bioassay set
		@throws InvalidDataException If the extra value type is null or has a 
			value type that doesn't match the specified class
	*/
	@SuppressWarnings("unchecked")
	public <I> PositionExtraValueBatcher<I> getPositionExtraValueBatcher(Class<I> clazz, 
		ExtraValueType extraValueType, Job job)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (extraValueType == null) throw new InvalidUseOfNullException("extraValueType");
		if (extraValueType.getValueType().getValueClass() != clazz)
		{
			throw new InvalidDataException("The specified class '" + clazz + "' can't be used for extra value type: " + 
				extraValueType.getValueType());
		}

		// Check if a batcher already exists in the cache or not
		if (positionExtraValueBatchers == null) positionExtraValueBatchers = new HashMap<ExtraValueType, PositionExtraValueBatcher<?>>();
		PositionExtraValueBatcher<I> batcher = (PositionExtraValueBatcher<I>)positionExtraValueBatchers.get(extraValueType);
		if (batcher == null)
		{
			if (getData().getExtraValues().containsKey(extraValueType.getData()))
			{
				throw new InvalidDataException("The bioassay set already has extra values of the type: " + extraValueType);
			}
			// Not found, create a new extra value and batcher
			ExtraValue extraValue = ExtraValue.getNew(getDbControl(), this, extraValueType, 
				getDataCube().newExtraValue(extraValueType.getValueType(), ExtraValue.CoordinateType.POSITION), job);
			
			// Saved by Hibernate cascade
			getData().getExtraValues().put(extraValueType.getData(), extraValue.getData());
			batcher = new PositionExtraValueBatcher<I>(getDbControl(), extraValue);
			positionExtraValueBatchers.put(extraValueType, batcher);
		}
		return batcher;
	}
	
	/**
		Get a query that returns spot data for this bioassayset. The query will
		select column, position and spot intensities by default. Ie.

<pre class="code">
// Default query -- filter is only joined if the bioassayset is filtered
SELECT spt.column, spt.position, spt.ch1, spt.ch2, ...
FROM Dynamic#PerSpot spt
[ JOIN Dynamic#Filter flt 
   ON flt.cube = spt.cube AND flt.column = spt.column 
   AND flt.position = spt.position ]
WHERE spt.cube = &lt;cube_no&gt; AND spt.layer = &lt;layer_no&gt; 
   [ AND flt.filter = &lt;filter_no&gt; ]
</pre>
		<p>
		Note! If the bioassay set contains transformed intensities, the
		default selection will un-transform the values. 
		
		@return A <code>DynamicSpotQuery</code> object
	*/
	public DynamicSpotQuery getSpotData()
	{
		/*
			SELECT spt.column, spt.position, spt.ch1, spt.ch2, ...
			FROM Dynamic#PerSpot spt
			[ JOIN Dynamic#Filter flt 
				ON flt.cube = spt.cube AND flt.column = spt.column AND flt.position = spt.position ]
			WHERE spt.cube = <cube_no> AND spt.layer = <layer_no> [ AND flt.filter = <filter_no> ]
		*/
		DynamicSpotQuery query = new DynamicSpotQuery(this);
		query.restrictPermanent(
			new DynamicRestriction(VirtualColumn.LAYER, getDataCubeLayerNo())
		);
		
		// If a filter exists we must join the filter table and specify filter number.
		if (getDataCubeFilterNo() != 0)
		{
			query.joinPermanent(
				new DynamicJoin(
					getVirtualDb(), JoinType.INNER, VirtualTable.FILTER,
					VirtualColumn.CUBE, VirtualColumn.COLUMN, VirtualColumn.POSITION
				)
			);
			query.restrictPermanent(
				new DynamicRestriction(VirtualTable.FILTER, VirtualColumn.FILTER, getDataCubeFilterNo())
			);
		}
		return query;
	}

	/**
		Get a query that returns position data for the data cube where this bioassayset 
		is located. The query will only select the position by default. The returned result
		doesn't depend on what spots are located in this bioassayset. Ie. the query will
		return the same result for all bioassaysets in the same data cube. If you need
		position data that depends on the spots in the bioassayset, use the {@link
		#getSpotData()} method and then join reporters using the {@link 
		DynamicSpotQuery#joinReporters(JoinType)} method.
		
<pre class="code">
// Default query
SELECT pos.position
FROM Dynamic#PerPosition AS pos
WHERE pos.cube = &lt;cube_no&gt;
</pre>

		@return A <code>DynamicPositionQuery</code> object
	*/
	public DynamicPositionQuery getPositionData()
	{
		/*
			SELECT pos.position
			FROM Dynamic#PerPosition AS pos
			WHERE pos.cube = <cube_no>
		*/
		return new DynamicPositionQuery(this);
	}

	/**
		Count the number of unique reporters and spots in all bioassays in this bioassayset.
		This method in called by the <code>onBeforeCommit</code> method to calculate 
		the numbers (which cannot be changed later) before the bioassayset is 
		committed to the database.
	*/
	private void countSpotsAndReporters()
	{
		DynamicQuery query = getSpotData();
		query.join(
			new DynamicJoin(
				getVirtualDb(), JoinType.LEFT, VirtualTable.POSITION,
				VirtualColumn.CUBE, VirtualColumn.POSITION
			)
		);
		query.select(
			Selects.expression(
				Aggregations.count(null, false),
				"spotCount"
			)
		);
		query.select(
			Selects.expression(
				Aggregations.count(Dynamic.column(VirtualTable.POSITION, VirtualColumn.REPORTER_ID), true),
				"reporterCount"
			)
		);
		/*
			SELECT COUNT(*) AS spotCount, COUNT(DISTINCT pos.reporter_id) AS reporterCount
			FROM Dynamic#PerSpot spt
			LEFT JOIN Dynamic#PerPosition pos spt.cube = pos.cube AND spt.position = pos.position
			WHERE spt.cube = ? AND spt.layer = ?
		*/
		DynamicResultIterator result = query.iterate(getDbControl());
		try
		{
			int spotCount = result.getIndex("spotCount");
			int reporterCount = result.getIndex("reporterCount");
			SqlResult r = result.next();
			getData().setNumSpots(r.getInt(spotCount));
			getData().setNumReporters(r.getInt(reporterCount));
		}
		catch (SQLException ex)
		{
			throw new BaseException(ex);
		}
		finally
		{
			result.close();
		}
	}

	/**
		Stores the result of {@link #getSpotCount(int)}.
	*/
	private Map<Integer, Integer> bioAssaySpots;
	
	/**
		Get the number of spots in a single bioassay. 
		This method in called by the <code>BioAssay.onBeforeCommit</code> method to 
		calculate the numbers (which cannot be changed later) before the bioassay is 
		saved to the database. The reason it is located here is that a single query
		can calculate the spots for all bioassays in a one go.
	*/
	int getSpotCount(int column)
	{
		if (bioAssaySpots == null) countBioAssaySpots();
		return IntegerUtil.getInt(bioAssaySpots.get(column));
	}
	
	/**
		@see #getSpotCount(int)
	*/
	private void countBioAssaySpots()
	{
		DynamicQuery query = getSpotData();
		query.select(
			Dynamic.select(VirtualColumn.COLUMN)
		);
		query.select(
			Selects.expression(
				Aggregations.count(null, false),
				"spotCount"
			)
		);
		query.group(
			Dynamic.column(VirtualTable.SPOT, VirtualColumn.COLUMN)
		);
		/*
			SELECT spt.column, COUNT(*) AS spotCount
			FROM Dynamic#PerSpot spt
			WHERE spt.cube = ? AND spt.layer = ?
			GROUP BY spt.column
		*/
		bioAssaySpots = new HashMap<Integer, Integer>();
		DynamicResultIterator result = query.iterate(getDbControl());
		try
		{
			int column = result.getIndex(VirtualColumn.COLUMN.getName());
			int spotCount = result.getIndex("spotCount");
			while (result.hasNext())
			{
				SqlResult r = result.next();
				bioAssaySpots.put(r.getInt(column), r.getInt(spotCount));
			}
		}
		catch (SQLException ex)
		{
			throw new BaseException(ex);
		}
		finally
		{
			result.close();
		}
	}
	
	private void checkAndCreateTables()
	{
		VirtualDb vdb = getVirtualDb();
		vdb.createTables(VirtualTable.SPOT, VirtualTable.POSITION);
		if (getDataCubeFilterNo() != 0)
		{
			vdb.createTables(VirtualTable.FILTER);
		}
	}
	
	/**
		Called when a bioassayset is about to be deleted. Finds which data cubes, layers, 
		filters and extra values that are exclusively used by this bioassayset or one of the 
		child bioassaysets (which are deleted by Hibernate cascade). 
		Then, it deletes all those items. A bug in this method may cause stale data 
		to be left in the database (bad) or data used by other bioassaysets to be deleted
		(even worse). For each type of item it is a three-step process:
		<ol>
		<li>Find the items used by at least one of the bioassaysets that are
			going to be deleted
		<li>Among the items found, find out if any of them are also used
			by other bioassaysets that are not going to be deleted
		<li>Delete only those items that are left
		</ol>
	*/
	private void deleteAnalysedData()
		throws BaseException
	{
		DbControl dc = getDbControl();
		SessionControl sc = dc.getSessionControl();
		
		// 1. Find all child bioassaysets recursively
		Set<BioAssaySetData> all = new HashSet<BioAssaySetData>();
		all.add(this.getData());
		addChildBioAssaySets(this.getData(), all);
		
		// 2. Find all data cubes that are only used by those bioassaysets
		org.hibernate.Session session = dc.getHibernateSession();
		
		// 2a. Find all data cubes used by at least one of the bioassaysets
		org.hibernate.query.Query<DataCubeData> cubeQuery = HibernateUtil.getPredefinedQuery(session,
			"FIND_DATACUBES_FOR_BIOASSAYSETS", DataCubeData.class);
		/*
			SELECT DISTINCT cbe
			FROM BioAssaySetData bas
			JOIN bas.dataCubeLayer.dataCube cbe
			WHERE bas IN (:basList)
		*/
		cubeQuery.setParameterList("basList", all);
		List<DataCubeData> possibleCubes = HibernateUtil.loadList(cubeQuery, sc);
		
		// 2b. Among the found data cubes, find those that are also used by other bioassaysets
		if (possibleCubes.size() > 0)
		{
			cubeQuery = HibernateUtil.getPredefinedQuery(session, "FIND_DATACUBES_USED_BY_OTHERS", DataCubeData.class);
			/*
				SELECT cbe
				FROM BioAssaySetData bas
				JOIN bas.dataCubeLayer.dataCube cbe
				WHERE bas NOT IN (:basList)
				AND cbe IN (:possibleCubes)
			*/
			cubeQuery.setParameterList("basList", all);
			cubeQuery.setParameterList("possibleCubes", possibleCubes);
			possibleCubes.removeAll(HibernateUtil.loadList(cubeQuery, sc));
			
	
			// 2c. Delete data from the remaining data cubes
			for (DataCubeData cube : possibleCubes)
			{
				dc.deleteItem(dc.getItem(DataCube.class, cube));
			}
		}

		// 3. Find remaining layers that are only used by those bioassaysets
		// 3a. Find all layers that are used by at least one of the bioassaysets that are not
		//     part of the already deleted data cubes
		org.hibernate.query.Query<DataCubeLayerData> layerQuery = null;
		if (possibleCubes.size() == 0)
		{
			layerQuery = HibernateUtil.getPredefinedQuery(session, 
				"FIND_DATACUBELAYERS_FOR_BIOASSAYSETS", DataCubeLayerData.class);
			/*
				SELECT lay
				FROM BioAssaySetData bas
				JOIN bas.dataCubeLayer lay
				WHERE bas IN (:basList)
			*/
			layerQuery.setParameterList("basList", all);
		}
		else
		{
			layerQuery = HibernateUtil.getPredefinedQuery(session, 
				"FIND_DATACUBELAYERS_FOR_BIOASSAYSETS_IGNORE_DELETED_DATACUBES", DataCubeLayerData.class);
			/*
				SELECT lay
				FROM BioAssaySetData bas
				JOIN bas.dataCubeLayer lay
				WHERE bas IN (:basList)
				AND lay.dataCube NOT IN (:deletedCubes)
			*/
			layerQuery.setParameterList("basList", all);
			layerQuery.setParameterList("deletedCubes", possibleCubes);
		}
		List<DataCubeLayerData> possibleLayers = HibernateUtil.loadList(layerQuery, sc);
		
		if (possibleLayers.size() > 0)
		{
			// 3b. Among the found layers, find thos the are also used by other bioassaysets
			layerQuery = HibernateUtil.getPredefinedQuery(session, 
				"FIND_DATACUBELAYERS_USED_BY_OTHERS", DataCubeLayerData.class);
			/*
				SELECT lay
				FROM BioAssaySetData bas
				JOIN bas.dataCubeLayer lay
				WHERE bas NOT IN (:basList)
				AND lay IN (:possibleLayers)
			*/
			layerQuery.setParameterList("basList", all);
			layerQuery.setParameterList("possibleLayers", possibleLayers);
			possibleLayers.removeAll(HibernateUtil.loadList(layerQuery, sc));
			
			// 3c. Delete data from the remaining layers
			for (DataCubeLayerData layer : possibleLayers)
			{
				dc.deleteItem(dc.getItem(DataCubeLayer.class, layer));
			}
		}
		
		// 4. Find remaining filters that are only used by those bioassaysets
		// 4a. Find all filters that are used by at least one of the bioassaysets that are not
		//     part of the already deleted data cubes
		org.hibernate.query.Query<DataCubeFilterData> filterQuery = null;
		if (possibleCubes.size() == 0)
		{
			filterQuery = HibernateUtil.getPredefinedQuery(session, 
				"FIND_DATACUBEFILTERS_FOR_BIOASSAYSETS", DataCubeFilterData.class);
			/*
				SELECT flt
				FROM BioAssaySetData bas
				JOIN bas.dataCubeFilter flt
				WHERE bas IN (:basList)
			*/
			filterQuery.setParameterList("basList", all);
		}
		else
		{
			filterQuery = HibernateUtil.getPredefinedQuery(session, 
				"FIND_DATACUBEFILTERS_FOR_BIOASSAYSETS_IGNORE_DELETED_DATACUBES", DataCubeFilterData.class);
			/*
				SELECT flt
				FROM BioAssaySetData bas
				JOIN bas.dataCubeFilter flt
				WHERE bas IN (:basList)
				AND flt.dataCube NOT IN (:deletedCubes)
			*/
			filterQuery.setParameterList("basList", all);
			filterQuery.setParameterList("deletedCubes", possibleCubes);
		}
		List<DataCubeFilterData> possibleFilters = HibernateUtil.loadList(filterQuery, sc);
		
		if (possibleFilters.size() > 0)
		{
			// 4b. Among the found filters, find those the are also used by other bioassaysets
			filterQuery = HibernateUtil.getPredefinedQuery(session, "FIND_DATACUBEFILTERS_USED_BY_OTHERS", DataCubeFilterData.class);
			/*
				SELECT flt
				FROM BioAssaySetData bas
				JOIN bas.dataCubeFilter flt
				WHERE bas NOT IN (:basList)
				AND flt IN (:possibleFilters)
			*/
			filterQuery.setParameterList("basList", all);
			filterQuery.setParameterList("possibleFilters", possibleFilters);
			possibleFilters.removeAll(HibernateUtil.loadList(filterQuery, sc));

			// 4c. Delete data from the remaining layers
			for (DataCubeFilterData filter : possibleFilters)
			{
				dc.deleteItem(dc.getItem(DataCubeFilter.class, filter));
			}
		}

		// 5. Find remaining extra values that are only used by those bioassaysets
		// 5a. Find all extra values that are used by at least one of the bioassaysets that are not
		//     part of the already deleted data cubes
		org.hibernate.query.Query<DataCubeExtraValueData> extraQuery = null;
		if (possibleCubes.size() == 0)
		{
			extraQuery = HibernateUtil.getPredefinedQuery(session, 
				"FIND_DATACUBEEXTRAVALUES_FOR_BIOASSAYSETS", DataCubeExtraValueData.class);
			/*
				SELECT ev
				FROM BioAssaySetData bas
				JOIN elements(bas.extraValues) ev
				WHERE bas IN (:basList)
			*/
			extraQuery.setParameterList("basList", all);
		}
		else
		{
			extraQuery = HibernateUtil.getPredefinedQuery(session, 
				"FIND_DATACUBEEXTRAVALUES_FOR_BIOASSAYSETS_IGNORE_DELETED_DATACUBES", DataCubeExtraValueData.class);
			/*
				SELECT ev
				FROM BioAssaySetData bas
				JOIN elements(bas.extraValues) ev
				WHERE bas IN (:basList)
				AND ev.dataCube NOT IN (:deletedCubes)
			*/
			extraQuery.setParameterList("basList", all);
			extraQuery.setParameterList("deletedCubes", possibleCubes);
		}
		List<DataCubeExtraValueData> possibleExtraValues = HibernateUtil.loadList(extraQuery, sc);
		
		if (possibleExtraValues.size() > 0)
		{
			// 5b. Among the found extra values, find those the are also used by other bioassaysets
			extraQuery = HibernateUtil.getPredefinedQuery(session, "FIND_DATACUBEEXTRAVALUES_USED_BY_OTHERS", DataCubeExtraValueData.class);
			/*
				SELECT ev
				FROM BioAssaySetData bas
				JOIN elements(bas.extraValues) ev
				WHERE bas NOT IN (:basList)
				AND ev IN (:possibleExtraValues)
			*/
			extraQuery.setParameterList("basList", all);
			extraQuery.setParameterList("possibleExtraValues", possibleExtraValues);
			possibleExtraValues.removeAll(HibernateUtil.loadList(extraQuery, sc));
			
			// 5c. Delete data from the remaining extra values
			for (DataCubeExtraValueData ev : possibleExtraValues)
			{
				dc.deleteItem(dc.getItem(DataCubeExtraValue.class, ev));
			}
		}
	}

	/**
		Query to retrieve children for a bioassayset.
		@see #addChildBioAssaySets(BioAssaySetData, Set)
	*/
	private org.hibernate.query.Query<BioAssaySetData> findChildren;

	/**
		Find all bioassaysets that are children to the specified parent (ie.
		the product of a transformation using the parent as a source) and
		add those the <code>all</code> set. The method is then called recursively
		to add the grandchildren and so on.
	*/
	private void addChildBioAssaySets(BioAssaySetData parent, Set<BioAssaySetData> all)
	{
		if (findChildren == null)
		{
			findChildren = HibernateUtil.getPredefinedQuery(getDbControl().getHibernateSession(), 
				"FIND_CHILD_BIOASSAYSETS", BioAssaySetData.class);
			/*
				SELECT bas 
				FROM BioAssaySetData bas
				WHERE bas.transformation.source = :parent
			*/
		}
		findChildren.setParameter("parent", parent.getId(), TypeWrapper.H_INTEGER);
		List<BioAssaySetData> children = HibernateUtil.loadList(findChildren, null);
		for (BioAssaySetData child : children)
		{
			all.add(child);
			addChildBioAssaySets(child, all);
		}
	}
}
