/*
	$Id: DataCubeColumn.java 6874 2015-04-17 12:02:22Z nicklas $
	
	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.DataCubeColumnData;
import net.sf.basedb.core.data.SharedData;

/**
	Represents the column coordinate in a data cube. A {@link
	BioAssay} must have a link to an item of this type. It is
	possible for more than one bioassay to link to the same column.
	Actually, if two {@link BioAssaySet}:s, one which is the child of the other,
	are in the same data cube, a bioassay must link to the same column as
	the parent bioassay.
	
	@author Nicklas
	@version 2.0
	@base.modified $Date: 2015-04-17 14:02:22 +0200 (Fri, 17 Apr 2015) $
*/
public class DataCubeColumn
	extends ChildItem
{
	/**
		The type of item represented by this class.
		@see Item#DATACUBECOLUMN
		@see #getType()
	*/
	public static final Item TYPE = Item.DATACUBECOLUMN;

	/**
		Create a new <code>DataCubeColumn</code> item.

		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@param cube The cube the new column is located in
		@return The new <code>DataCubeColumn</code> item
		@throws BaseException If there is an error
	*/
	static DataCubeColumn getNew(DbControl dc, DataCube cube)
		throws BaseException
	{
		DataCubeColumn column = dc.newItem(DataCubeColumn.class);
		column.setDataCube(cube);
		column.setColumnNo(cube.getNextColumnNo());
		return column;
	}

	/**
		Get a <code>DataCubeColumn</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>DataCubeColumn</code> item
		@throws ItemNotFoundException If an item with the specified id 
			is not found
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static DataCubeColumn getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		DataCubeColumn column = dc.loadItem(DataCubeColumn.class, id);
		if (column == null) throw new ItemNotFoundException("DataCubeColumn[id=" + id + "]");
		return column;
	}


	/**
	 	Creates a new data cube item from the given data.
		@param data the data
	*/
	DataCubeColumn(DataCubeColumnData data)
	{
		super(data);
	}
	
	@Override
	DataCubeColumnData getData()
	{
		return (DataCubeColumnData)getBasicData();
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------

	/*
		From BasicItem
		-------------------------------------------
	*/
	/**
		Deny DELETE permission. Columns are automatically deleted when
		the data cube is deleted (by Hibernate cascade).
	*/
	@Override
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		super.initPermissions(granted, denied | Permission.deny(Permission.DELETE));
	}
	// -------------------------------------------
	/*
		From the ChildItem class
		-------------------------------------------
	*/
	/**
		Get the experiment.
	*/
	@Override
	SharedData getSharedParent()
	{
		return getData().getDataCube().getVirtualDb().getExperiment();
	}
	/**
		USE permission is enough to be able to manage data cubes.
	*/
	@Override
	Permission getPermissionForWriteDeleteAndCreate()
	{
		return Permission.USE;
	}
	// -------------------------------------------

	/**
		Get the data cube this layer belongs to.
		@return A <code>DataCube</code> object
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission to the item
		@throws BaseException If there is another error
	*/
	public DataCube getDataCube()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(DataCube.class, getData().getDataCube());
	}
	
	/**
		Set the experiment. This cannot be changed after the dynamic tables 
		has been created.
	*/
	private void setDataCube(DataCube cube)
		throws PermissionDeniedException, InvalidDataException
	{
		if (cube == null) throw new InvalidUseOfNullException("dataCube");
		cube.checkPermission(Permission.WRITE);
		getData().setDataCube(cube.getData());
	}

	/**
		Get the number of the data cube this layer belongs to.
	*/
	public short getDataCubeNo()
	{
		return getData().getDataCube().getCubeNo();
	}

	/**
		Get the number of this column.
	*/
	public short getColumnNo()
	{
		return getData().getColumnNo();
	}
	private void setColumnNo(short columnNo)
	{
		getData().setColumnNo(columnNo);
	}
	
	/**
		Get the virtual database this column belongs to.
	*/
	VirtualDb getVirtualDb()
	{
		return getDbControl().getItem(VirtualDb.class, getData().getDataCube().getVirtualDb());
	}


}
