/*
	$Id: PropertyFilter.java 8005 2021-08-12 08:44:12Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import net.sf.basedb.core.SyncFilter.SourceItemTransform;
import net.sf.basedb.core.data.PropertyFilterData;
import net.sf.basedb.core.data.ReporterData;
import net.sf.basedb.core.data.UnitData;
import net.sf.basedb.core.query.Annotations;
import net.sf.basedb.core.query.AnyToAnyRestriction;
import net.sf.basedb.core.query.EntityQuery;
import net.sf.basedb.core.query.Query;
import net.sf.basedb.core.query.Restriction;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.SqlQuery;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.IdListRestriction;
import net.sf.basedb.util.listable.ListableUtil;
import net.sf.basedb.util.listable.SourceItemTransformerFactory;
import net.sf.basedb.util.listable.SourceItemTransformerWithCache;
import net.sf.basedb.util.listable.TransformContext;
import net.sf.basedb.util.units.UnitUtil;

/**
	Used to store filter settings in the current context. A filter is
	a simple expression of the form: <code>property operator value</code>,
	ie. <code>name = 'Nicklas'</code> or <code>spots &gt; 10000</code>.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2021-08-12 10:44:12 +0200 (Thu, 12 Aug 2021) $
	@see SessionControl#getCurrentContext(Item)
	@see ItemContext
*/
public class PropertyFilter
{

	/**
		Create a property filter instance from what has been saved in the database.
		Basically, we need to strip the [filterrow] if it is set.
		@since 3.5
	*/
	static PropertyFilter create(String propertyFromDb, PropertyFilterData data)
	{
		int rowIndex = 0;
		// The property can string can actually be <property>[filterrow]
		// So we need to split and parse in that case
		// properties without [filterrow] belong to row 0.
		// The largest value for filterrow becomes the 'numFilterRows' value
		if (propertyFromDb.endsWith("]"))
		{
			int split = propertyFromDb.lastIndexOf('[');
			rowIndex = Integer.parseInt(propertyFromDb.substring(split+1, propertyFromDb.length()-1));
			propertyFromDb = propertyFromDb.substring(0, split);
		}
		return new PropertyFilter(propertyFromDb, rowIndex, data);
	}
	
	/**
		The maximum allowed length of a filter value.
		@since 2.6.2
	*/
	public static final int MAX_VALUE_LENGTH = PropertyFilterData.MAX_VALUE_LENGTH;
	
	/**
		Flag to use for temporary filters.
		@since 2.16
	*/
	public static final boolean TEMPORARY = true;
	
	/**
	 	The string/characters used as a separator in a string with more then one filter values.
	 */
	private static final String value_separator = "|";
	
	private final Pattern ENTITY_COLLECTION = Pattern.compile("\\&([\\w\\.]+)\\((.+)\\)");

	private final Pattern DATA_FILES = Pattern.compile("\\¤(\\d+)\\((.+)\\)");
	
	private final String property;
	private int rowIndex;
	private Type valueType;
	private Operator operator;
	private String value;
	private Unit unit;
	private boolean temporary;
	
	/**
		Create a new filter for a property on filter row 0.
		@param property The name of the property
	*/
	public PropertyFilter(String property)
	{
		this.property = property;
	}

	/**
		Create a new filter for a property.
		@param property The name of the property
		@param rowIndex The filter row to create the filter on
	*/
	public PropertyFilter(String property, int rowIndex) 
	{
		this.property = property;
		this.rowIndex = rowIndex;
	}
	
	/**
		Create a new filter for a property.
		@param property The name of the property
		@param operator The operator to use
		@param value The value to check against
		@param valueType The type of the value
	*/
	public PropertyFilter(String property, Operator operator, String value, Type valueType)
	{
		this.property = property;
		this.operator = operator;
		this.value = value;
		this.valueType = valueType;
	}
	
	/**
		Create a new filter for a property.
		@param property The name of the property
		@param operator The operator to use
		@param value The value to check against
		@param valueType The type of the value
		@param temporary A flag indicating if the filter should be temporary or not
		@since 2.16
	*/
	public PropertyFilter(String property, Operator operator, String value, Type valueType, boolean temporary)
	{
		this.property = property;
		this.operator = operator;
		this.value = value;
		this.valueType = valueType;
		this.temporary = temporary;
	}

	
	/**
		Create a new filter from a loaded database object.
	*/
	PropertyFilter(String property, int rowIndex, PropertyFilterData filter)
	{
		this.property = property;
		this.rowIndex = rowIndex;
		this.valueType = Type.fromValue(filter.getValueType());
		this.operator = Operator.fromValue(filter.getOperator());
		this.value = filter.getValue();
		if (filter.getUnit() != null)
		{
			this.unit = new Unit(filter.getUnit());
			this.unit.getName(); // Forces initialisation if the unit is a proxy
		}
	}
	
	/**
		Create a new database filter object from this filter. 
	*/
	PropertyFilterData getData()
	{
		PropertyFilterData filter = new PropertyFilterData();
		filter.setValueType(valueType.getValue());
		filter.setOperator(operator.getValue());
		filter.setValue(value);
		filter.setUnit(unit == null ? null : unit.getData());
		return filter;
	}
	
	/*
		From the Object class
		-------------------------------------------
	*/
	@Override
	public String toString()
	{
		return property + (rowIndex > 0 ? "["+rowIndex+"]" : "") + " " + operator + " " + value;
	}
	// -------------------------------------------

	/**
		Set or clear a flag indicating that this filter is temporary filter.
		Temporary filters are not saved between sessions.
		@since 2.16
	*/
	public void setTemporary(boolean temporary)
	{
		this.temporary = temporary;
	}
	
	/**
	 	Is this filter a temporary filter or not.
		@since 2.16
	*/
	public boolean isTemporary()
	{
		return temporary;
	}
	
	/**
		Get the name of the property this filter is used on.
	*/
	public String getProperty()
	{
		return property;
	}
	
	/**
		Checks if this filter is an extension filter. An extension
		filter has a {@link #getProperty()} that starts with
		"!x.prefix".
		@param prefix The extension prefix to check, or null to check any extension
		@since 3.18
	*/
	public boolean isExtensionFilter(String prefix)
	{
		return prefix == null ? property.startsWith("!x.") : property.startsWith("!x."+prefix);
	}
	
	/**
		Get the filter row index this filter is part of.
		@since 3.5
	*/
	public int getRowIndex()
	{
		return rowIndex;
	}
	
	/**
		Only used internally when removing filter rows
		@see ItemContext#removeFilterRow(int)
		@since 3.5
	*/
	void setRowIndex(int rowIndex)
	{
		this.rowIndex = rowIndex;
	}
	
	/**
		Get the type of the property.
	*/
	public Type getValueType()
	{
		return valueType;
	}
	/**
		Set the type of the property.
		@param valueType The type
	*/
	public void setValueType(Type valueType)
	{
		this.valueType = valueType;
	}
	
	/**
		Get the operator to use in the filter.
	*/
	public Operator getOperator()
	{
		return operator;
	}
	/**
		Set the operator to use in the filter.
	*/
	public void setOperator(Operator operator)
	{
		this.operator = operator;
	}
	
	/**
		Sets the unit the values for the property are stored in the
		database. When a unit is set, it is possible to have filter values
		with units. The filter values are automatically converted to this
		unit before the filter is applied. For example, if we set the unit to
		'g', someone may filter on '0.5kg', which is converted to 500 in
		the query.
		@param unit The unit, or null to remove the unit from the filter
		@since 2.9
	*/
	public void setUnit(Unit unit)
	{
		this.unit = unit;
	}
	
	/**
		Get the value to use in the filter.
		@return The value as a string
	*/
	public String getValue()
	{
		return value;
	}
	/**
		Set the value to use in the filter.
		@param value The value as a string
	*/
	public void setValue(String value)
	{
		this.value = value;
	}
	
	/**
		Get the value as an object of the correct type. If the 
		filter property is numeric and has a unit, the value is converted
		to the correct	unit automatically. For example, if the value is '5kg'
		and the unit is 'g' this method returns 5000. 
		<p>
		Note! The returned value is cast to the class supported by
		the {@link #getValueType()}. Beware of this when the 
		filter has an integer type and a unit with low precision.
		For example, if the value is '125min' and the unit is
		'hour' this method returns the integer 2. To avoid this
		use {@link #getValueAsObject(Unit, boolean)} with 
		asDouble=true instead.
		
		@return The value as an object
		@throws InvalidDataException If the string value can't be parsed
	*/
	public <T> T getValueAsObject()
	{
		return getValueAsObject(value, unit, false);
	}
	
	/**
		Get the value as an object of the proper type converted to the
		specified unit. For example, if the unit is 'g' and the
		value is '5kg' this method returns 5000.
		<p>
		If the asDouble parameter is true, the value is returned as a 
		double, otherwise it is cast to the
		proper class as specified by the {@link #getValueType()}. 
		<p>
		Note! The above only applies for numerical properties that
		has a unit. If not, the value is simply parsed with 
		{@link Type#parseString(String)}.
		
		@param unit A unit, or null to use the default unit
		@param asDouble TRUE to return the value as a double, FALSE
			to let the {@link #getValueType()} decide
		@return The value as an object
		@since 2.9
	*/
	public <T> T getValueAsObject(Unit unit, boolean asDouble)
	{
		return getValueAsObject(value, unit, asDouble);
	}
	
	@SuppressWarnings("unchecked")
	private <T> T getValueAsObject(String value, Unit unit, boolean asDouble)
	{
		Object oValue = null;
		if (unit == null) unit = this.unit;
		if (unit != null && value != null)
		{
			String[] valueAndUnit = UnitUtil.splitValueWithUnit(value);
			if (valueAndUnit != null)
			{
				boolean newSession = unit.isDetached();
				org.hibernate.Session session = null; 
				try
				{
					session = newSession ? HibernateUtil.newSession() : unit.getDbControl().getHibernateSession();
					Double dbl = convertValueToUnit(session, valueAndUnit[0], valueAndUnit[1], unit);
					oValue = asDouble ? dbl : valueType.convertNumber(dbl);
				}
				finally
				{
					if (newSession && session != null) HibernateUtil.close(session);
				}
			}
		}
		if (oValue == null)
		{
			oValue = valueType.parseString(value);
		}
		return (T)oValue;
	}

	/**
	 	Splits the value string at each "|" and creates
	 	a list with objects from the substrings that were found.
	 	@return A list with objects, or null if splitting string of values returns null.
	 	@throws InvalidDataException If any of the string values can't be parsed.
	 	@since 2.5 
	 */
	public <T> List<T> getValuesAsObjects()
	{
		return getValuesAsObjects(value, unit, false);
	}
	
	/**
		Same as {@link #getValueAsObject(Unit, boolean)} but the filter string
		is first split at each "|". Each individual string may or may not specify
		a unit. If a unit is specified for some parts but not for all, the nearest
		following unit is used. For example, the string '300g|5|10kg' is the same as
		'300g|5kg|10kg'.'

		@param unit A unit, or null to use the default unit
		@param asDouble TRUE to return the value as a double, FALSE
			to let the {@link #getValueType()} decide
		@return A list with the values
		@since 2.9
	*/
	public <T> List<T> getValuesAsObjects(Unit unit, boolean asDouble)
	{
		return getValuesAsObjects(value, unit, asDouble);
	}
	
	@SuppressWarnings("unchecked")
	private <T> List<T> getValuesAsObjects(String value, Unit unit, boolean asDouble)
	{
		if (value == null) return Collections.emptyList();
		String[] values = value.split("\\"+getValueSeparator(), -1);
		Object[] objects = new Object[values.length];
		
		if (unit == null) unit = this.unit;

		boolean newSession = unit != null && unit.isDetached();
		org.hibernate.Session session = null; 
		try
		{
			if (unit != null)
			{
				session = newSession ? HibernateUtil.newSession() : unit.getDbControl().getHibernateSession();
			}
			
			String lastUnit = null;
			for (int i = values.length - 1; i >= 0; --i)
			{
				String v = values[i];
				if (v.length() == 0) v = null;
				
				Object oValue = null;
				if (unit != null)
				{
					String[] valueAndUnit = UnitUtil.splitValueWithUnit(v);
					if (valueAndUnit != null)
					{
						if (valueAndUnit[1] != null) lastUnit = valueAndUnit[1];
						Double dbl = convertValueToUnit(session, valueAndUnit[0], lastUnit, unit);
						oValue = asDouble ? dbl : valueType.convertNumber(dbl);
					}
				}
				if (oValue == null)
				{
					oValue = valueType.parseString(v);
				}
				objects[i] = oValue;
			}
		}
		finally
		{
			if (newSession && session != null) HibernateUtil.close(session);
		}
		return (List<T>)Arrays.asList(objects);
	}

	
	/**
		Converts a string to a numeric value with the correct unit.
		@param session Hibernate session used to lookup units 
		@param value The string value which should be parsable with
			{@link Double#parseDouble(String)}
		@param unitSymbol The unit of the string value
		@param convertTo The unit to convert the value to
		@return The converted value
	*/
	private Double convertValueToUnit(org.hibernate.Session session, String value, String unitSymbol, Unit convertTo)
	{
		Double result = null;
		if (unitSymbol != null)
		{
			UnitData other = convertTo.findOtherUnit(session, unitSymbol);
			if (other != null)
			{
				result = convertTo.fromOther(Double.parseDouble(value), new Unit(other));
			}
		}
		else
		{
			result = Double.parseDouble(value);
		}
		return result;
	}
	
	/**
		Create a restriction from the filter. There are a few special cases:
		<p>
		If the property starts with a $ it is interpreted as a <code>alias.property</code> 
		value, ignoring the root alias of the query. The alias is removed from the
		property before proceeding.
		<p>
		If the property starts with a # or ## it is a filter on annotations. The rest of
		the property is the ID of the annotation type, and we use a
		{@link AnnotationSimpleRestriction} object for the restriction. If there
		is only one # in the prefix only direct annotations are searched, if there
		are two ##, inherited annotations are also searched.
		<p>
		If the property starts with a £, the rest of the property is the path that
		joins the {@link ReporterData#getReporterListScores()} collection. The filter
		value is the ID of the reporter list.
		<p>
		If the property starts with a @ the property is a collection of values and
		we always use an IN restriction, ignoring the specified operator.
		<p>
		If the property starts with &amp; it is a collection of entities. The property
		must end with a property name from the entity in the collection within paranthesis.
		Eg. &amp;experiments(name). The filter is applied as a subquery.
		<p>
		If the property starts with ! it is another special filter:
		<ul>
		<li>Since BASE 2.15: !sharedTo.name, which is used to filter 
			shareable items by the name of users/groups/projects it has been shared to. 
			The 'name' part can be replaced with any other property that is common for 
			users, groups and projects (eg. description or id). 
		<li>Since BASE 3.18: !x., which is intended to be handled by extensions. It
			is ignored by the BASE core code. The rest of the property should be a unique
			identifier for the extension that is handling the filter.
		</ul>
		<p>
		The property may also start with $@.
		@return The restriction
		@since 2.8
	*/
	public Restriction getRestriction(DbControl dc, EntityQuery query)
	{
		
		Restriction restriction = null;
		String property = getProperty();
		String alias = null;
		Expression indexExpression = null;
		
		// Ignore -- expected to be handled by extensions
		if (property != null && property.startsWith("!x.")) return null;
		
		if (property != null && property.startsWith("$"))
		{
			// Property is alias.property or alias only
			int dotIndex = property.indexOf(".");
			if (dotIndex == -1)
			{
				// Alias only
				alias = property.substring(1);
				property = null;
			}
			else if (property != null)
			{
				// Alias.property
				alias = property.substring(1, dotIndex);
				property = property.substring(dotIndex+1);
			}
		}
		
		// Indexed property
		if (property != null && property.startsWith("@") && property.endsWith(")"))
		{
			int leftIndex = property.indexOf("(");
			String indexKey = property.substring(leftIndex+1, property.length()-1);
			property = property.substring(1, leftIndex);
			indexExpression = Hql.expression(indexKey, null);
		}
		
		if (property != null && property.startsWith("#"))
		{
			// Property is id of annotation type: #id
			// If prefixed with two ## inherited annotations should be searched
			// We always search for one or the other; never both
			boolean searchInherited = property.startsWith("##");
			boolean searchPrimary = !searchInherited;
			
			if (property.contains("("))
			{
				// The annotation is on a joined property: #(path.to.join)id
				// We extract the join path and use as the alias
				int joinStart = property.indexOf('(');
				int joinEnd = property.indexOf(')');
				alias = query.getRootAlias() + "." + property.substring(joinStart+1, joinEnd);
				property = property.substring(0, joinStart) + property.substring(joinEnd+1); // Should give us: #id
			}
			
			int annotationTypeId = (Integer)Type.INT.parseString(property.substring(searchInherited ? 2 : 1));
			
			AnnotationType at = AnnotationType.getById(dc, annotationTypeId);
			if (searchInherited && at.getDisableInheritance()) return null;
			
			String valueToUse = value;
			boolean excludeDefaultValues = false;
			if (at.getProjectAnnotations() && valueToUse != null && valueToUse.endsWith("¤"))
			{
				valueToUse = valueToUse.substring(0, valueToUse.length()-1);
				if (valueToUse.length() == 0) valueToUse = null;
				excludeDefaultValues = dc.getSessionControl().getActiveProjectId() != 0;
			}
			
			Unit unit = at.getDefaultUnit();
			
			if (at.getMultiplicity() == 1 && !searchInherited)
			{
				// This is similar to a regular property expression
				Expression propertyExpression = Annotations.expression(alias, at);
				if (getValueType() == Type.TIMESTAMP)
				{
					propertyExpression = Expressions.toDate(propertyExpression);
				}
				if (operator.isListOperator())
				{
					List<Expression> parameters = new ArrayList<Expression>();
					List<Object> objects = getValuesAsObjects(valueToUse, unit, true);
					for (Object obj : objects)
					{
						Expression e = null;
						if (obj != null)
						{
							String parameterName = "p" + dc.uniqueRandom();
							Type theValueType = obj instanceof Double ? Type.DOUBLE : getValueType();
							if (theValueType == Type.TIMESTAMP) theValueType = Type.DATE;
							e = Expressions.parameter(parameterName, obj, theValueType);
						}
						parameters.add(e);
					}
					restriction = operator.getRestriction(propertyExpression, parameters.toArray(new Expression[parameters.size()]));
					if (operator.isNegationOperator() && objects.indexOf(null) == -1)
					{
						// <>'foo' should also match null unless null is in the list
						restriction = Restrictions.or(restriction, Restrictions.eq(propertyExpression, null));
					}
				}
				else
				{
					String parameterName = "p" + dc.uniqueRandom();
					Expression parameter = null;
					if (valueToUse != null)
					{
						Object value = getValueAsObject(valueToUse, unit, true);
						Type theValueType = value instanceof Double ? Type.DOUBLE : getValueType();
						// We only search on the date part in timestamp values
						if (theValueType == Type.TIMESTAMP) theValueType = Type.DATE;
						parameter = Expressions.parameter(parameterName, value, theValueType);
					}
					restriction = operator.getRestriction(propertyExpression, parameter);
					if (operator.isNegationOperator() && valueToUse != null)
					{
						// <>'foo' should also match null
						restriction = Restrictions.or(restriction, Restrictions.eq(propertyExpression, null));
					}
				}
				if (excludeDefaultValues)
				{
					// ... AND atmpNN.overrideId<>0
					// This works since AnnotationJoin joins the "Annotations" table with atmpNN alias
					restriction = Restrictions.and(
						restriction,
						Restrictions.neq(
							Hql.property("atmp"+annotationTypeId, "overrideId"),
							Expressions.integer(0)
						));
				}
			}
			else
			{
				AnnotationRestriction.Options aOptions = new AnnotationRestriction.Options();
				aOptions.setIncludePrimary(searchPrimary)
					.setIncludeInherited(searchInherited)
					.setExcludeDefaultValues(excludeDefaultValues);
				if (operator.isListOperator())
				{
					List<Object> objects = getValuesAsObjects(valueToUse, unit, true);
					if (operator == Operator.BETWEEN)
					{
						if (objects.size() != 2)
						{
							throw new InvalidDataException("Two values are needed for >< operator");
						}
						restriction = new AnnotationBetweenRestriction(alias, at,
								objects.get(0), objects.get(1), aOptions);
					}
					else
					{
						restriction = new AnnotationSimpleRestriction(alias, at, operator, objects, aOptions);
						boolean hasNullFilter = objects.indexOf(null) != -1;
						if (hasNullFilter ^ operator.isNegationOperator() && !excludeDefaultValues)
						{
							// Also find items without annotation if searching for null (equality operators) or only non-null (inequality operators) 
							restriction = Restrictions.or(restriction, new HasAnnotationRestriction(alias, at, false, aOptions));
						}
					}
					
				}
				else
				{
					Object value = getValueAsObject(valueToUse, unit, true);
					Operator operator = getOperator();
					if (value == null && (operator == Operator.EQ || operator == Operator.NEQ))
					{
						restriction = new HasAnnotationRestriction(alias, at, 
							operator != Operator.EQ, aOptions);
					}
					else
					{
						restriction = new AnnotationSimpleRestriction(alias, at, operator, value, aOptions);
						if (operator.isNegationOperator() && !excludeDefaultValues)
						{
							// Also find items without annotation
							restriction = Restrictions.or(restriction, new HasAnnotationRestriction(alias, at, false, aOptions));
						}
					}
				}
			}
		}
		
		else if (property != null && property.startsWith("£"))
		{
			// Property is reporterListScores collection
			// The value is the ID of the reporter list to
			// filter on
			Integer reporterListId = (Integer)getValueAsObject();
			if (reporterListId != null)
			{
				try
				{
					ReporterList list = ReporterList.getById(dc, reporterListId);
					if (operator == Operator.NEQ)
					{
						restriction = Hql.isNotPartOf(alias, property.substring(1), list);
					}
					else
					{
						restriction = Hql.isPartOf(alias, property.substring(1), list);
					}
				}
				catch (ItemNotFoundException ex)
				{}
			}
		}
		
		else if (property != null && property.startsWith("§"))
		{
			// Property is item list
			// The value is the ID of the item list to filter on
			// We support a single selection and either EQ or NEQ operator
			// or multi-selection with either IN or NOTIN operator
			String prefix = alias == null ? "$id" : alias + ".id";
			String listId = net.sf.basedb.util.Values.getString(getValuesAsObjects(), ",", true);
			String listOp = operator.isNegationOperator() ? "NOT IN" : "IN";
			String where = operator.isListOperator() ? "lst.id IN (" + listId + ")" : "lst.id="+listId;
			restriction = Hql.restriction(prefix+" "+listOp+" (SELECT mmb FROM ItemListData lst INNER JOIN lst.members mmb WHERE "+where+")", "$");
		}
		
		else if (property != null && property.startsWith("|"))
		{
			// Property is Any-to-any filter: |link-name|link-type|property
			// link-name is required
			// link-type is optional, if not specified filter is always of the type exists/not-exists
			// property is optional, if not specified filter is always of the type exists/not-exists
			String[] parts = property.split("\\|", 4);
			String linkName = parts[1];
			Item linkType = parts.length > 2 ? Item.valueOf(parts[2]) : null;
			String linkProperty = parts.length > 3 ? parts[3] : null;
			Object value = getValueAsObject();
			
			if (linkType == null || linkProperty == null || value == null)
			{
				if (Boolean.TRUE.equals(value) || (value == null && operator.isNegationOperator()))
				{
					restriction = AnyToAnyRestriction.exists(null, linkName, linkType, null);
				}
				else
				{
					restriction = AnyToAnyRestriction.missing(null, linkName, linkType, null);
				}
			}
			else
			{
				EntityQuery subquery = new AbstractEntityQuery(linkType, linkType, null, "id", false, null)
				{};

				PropertyFilter pp = new PropertyFilter(linkProperty, operator, getValue(), getValueType());
				subquery.restrict(pp.getRestriction(dc, subquery));
				restriction = AnyToAnyRestriction.exists(null, linkName, linkType, subquery);
			}
		}
		else if (property != null && property.startsWith("/"))
		{
			// Property is Parent-item filter: /DIRECTION/TYPE/subtype-id/(property|#annotation-id)
			String[] parts = property.split("/", 5);
			SourceItemTransform direction = SourceItemTransform.PARENT_TO_CHILD;
			int baseIndex = 1;
			if (parts.length == 5) // The /DIRECTION is optional
			{
				baseIndex = 2;
				direction = "CHILD".equals(parts[1]) ? SourceItemTransform.CHILD_TO_PARENT : SourceItemTransform.PARENT_TO_CHILD;
			}
			Item targetType = Item.valueOf(parts[baseIndex]);
			String targetSubtype = parts[baseIndex+1];
			String targetProperty = parts[baseIndex+2];
			
			// We execute a query for the parent items...
			EntityQuery subquery = new AbstractEntityQuery(targetType, targetType, null, "id", false, null)
			{};
			if (targetType == Item.RAWBIOASSAY)
			{
				subquery.restrict(
					Restrictions.eq(
						Hql.property("rawDataType"), 
						Expressions.string(targetSubtype)
				));
			}
			else
			{
				subquery.restrict(
					Restrictions.eq(
						Hql.property("itemSubtype"), 
						Expressions.integer(Integer.parseInt(targetSubtype))
				));
			}
			PropertyFilter pp = new PropertyFilter(targetProperty, operator, getValue(), getValueType());
			subquery.restrict(pp.getRestriction(dc, subquery));

			// We then transform the list if ID:s to a list of child item ID:s using ItemList functionality
			SortedSet<Integer> parentIds = new TreeSet<>(subquery.idList(dc));
			TransformContext tCtx = new TransformContext(dc);

			SourceItemTransformerFactory factory = ListableUtil.getTransformerFactory(query.getItemType());
			factory = new SourceItemTransformerWithCache(factory);
			Set<Integer> myIds = factory.create(targetType, direction).transform(tCtx, parentIds);
			
			// The final ID list is returned as a restriction
			restriction = new IdListRestriction(myIds);
		}
		else
		{
			String parameterName = "p" + dc.uniqueRandom();
			Expression parameter = null;
			if (getValue() != null && !operator.isListOperator())
			{
				Object value = getValueAsObject(unit, true);
				Type theValueType = value instanceof Double ? Type.DOUBLE : getValueType();
				// We only search on the date part in timestamp values
				if (theValueType == Type.TIMESTAMP) theValueType = Type.DATE;
				parameter = Expressions.parameter(parameterName, value, theValueType);
			}

			if (property != null && property.startsWith("@"))
			{
				// Property is a collection - we only support equal and not-equal tests
				property = property.substring(1);
				Expression elements = Hql.elements(alias, property);
				Restriction isEmpty = Restrictions.eq(Hql.size(alias, property), Expressions.integer(0));
				if (operator.isListOperator())
				{
					// filter to match one of several elements in a collection
					List<Restriction> alternatives = new ArrayList<Restriction>();
					List<Object> objects = getValuesAsObjects(unit, true);
					boolean hasEmptyCheck = false;
					for (Object obj : objects)
					{
						if (obj != null)
						{
							parameterName = "p" + dc.uniqueRandom();
							Type theValueType = obj instanceof Double ? Type.DOUBLE : getValueType();
							Expression e = Expressions.parameter(parameterName, obj, theValueType);
							alternatives.add(Restrictions.in(e, elements));
						}
						else if (!hasEmptyCheck)
						{
							hasEmptyCheck = true;
							alternatives.add(isEmpty);
						}
					}
					restriction = Restrictions.or(alternatives.toArray(new Restriction[alternatives.size()]));
				}
				else if (parameter != null)
				{
					// filter for a single element in a collection
					restriction = Restrictions.in(parameter, elements);
					if (operator == Operator.NEQ)
					{
						restriction = Restrictions.not(restriction);
					}
				}
				else
				{
					// filter that returns items with empty collections
					restriction = isEmpty;
					if (operator == Operator.NEQ)
					{
						restriction = Restrictions.not(restriction);
					}
				}
			}
			
			else if (property != null && property.startsWith("¤"))
			{
				// Property is id of data file type: ¤id(filter-property)
				Matcher m = DATA_FILES.matcher(property);
				if (!m.matches()) 
				{
					throw new InvalidDataException("Invalid collection filter: " + property);
				}
				int fileTypeId = (Integer)Type.INT.parseString(m.group(1));
				String collectionFilter = m.group(2);

				DataFileType ft = DataFileType.getById(dc, fileTypeId);
				Query subquery = new AbstractEntityQuery(Item.FILESET, null, false, null)
				{};
				Restriction joinRestriction = Restrictions.eq(Hql.property("mbr", "dataFileType.id"), Hql.entity(ft));
				subquery.join(Hql.innerJoin(null, "members", "mbr", joinRestriction, false));

				String thisAlias = "mbr";
				// LEFT JOIN the parts in the filter-property path (except the last one)
				String[] cPath = collectionFilter.split("\\.");
				for (int i = 0; i < cPath.length - 1; ++i)
				{
					String nextAlias = "fp" + i;
					subquery.join(Hql.leftJoin(thisAlias, cPath[i], nextAlias, null, false));
					thisAlias = nextAlias;
				}
				collectionFilter = cPath[cPath.length - 1];
			
				// Special case when searching for '='
				boolean notContains = false;
				Operator actualOperator = operator;
				String actualValue = getValue();
				if (actualValue == null && operator == Operator.EQ)
				{
					// IN (... WHERE <property> IS NULL) is rewritten as
					// NOT IN (... WHERE <property> LIKE '%') 
					actualOperator = Operator.LIKE;
					actualValue = "%";
					notContains = !notContains;
				}
				
				// Restrict the subquery
				PropertyFilter pp = new PropertyFilter("$" + thisAlias + "." + collectionFilter, 
						actualOperator, actualValue, getValueType());
				subquery.restrict(pp.getRestriction(dc, query));
				
				// Create the main restriction
				if (notContains)
				{
					// "not contains": NOT root-entity.fileSet IN (subquery) OR root-entity.fileSet IS NULL
					restriction = Restrictions.not(Restrictions.in(Hql.property("fileSet"), subquery));
					restriction = Restrictions.or(restriction, Restrictions.eq(Hql.property("fileSet"), null));
				}
				else
				{
					// "contains": root-entity.fileSet IN (subquery)
					restriction = Restrictions.in(Hql.property("fileSet"), subquery);
				}
			}

			else if (property != null && property.startsWith("&"))
			{
				// Property is an entity collection: collection(filter-property)
				Matcher m = ENTITY_COLLECTION.matcher(property);
				if (!m.matches()) 
				{
					throw new InvalidDataException("Invalid collection filter: " + property);
				}
				String entityCollection = m.group(1);
				String collectionFilter = m.group(2);
				boolean rewriteNullFilter = false;
				if (collectionFilter.startsWith("%"))
				{
					collectionFilter = collectionFilter.substring(1);
					rewriteNullFilter = true;
				}
				
				/*
					We are going to create a subquery like:
					root-entity = ANY(
						SELECT root-entity FROM root-entity
						LEFT JOIN collection-path cp 
						WHERE cp.filter-property ........
					)
					
					- or -
					
					root-entity <> ALL(
						SELECT root-entity FROM root-entity
						LEFT JOIN collection-path cp 
						WHERE cp.filter-property ........
					)
					
				*/
				Query subquery = new AbstractEntityQuery(query.getItemType(), null, false, null)
				{};
				// LEFT JOIN the parts in the collection path
				String[] cPath = entityCollection.split("\\.");
				String thisAlias = null;
				for (int i = 0; i < cPath.length; ++i)
				{
					String nextAlias = "cp" + i;
					subquery.join(Hql.leftJoin(thisAlias, cPath[i], nextAlias, null, false));
					thisAlias = nextAlias;
				}
				// LEFT JOIN the parts in the filter-property path (except the last one)
				cPath = collectionFilter.split("\\.");
				for (int i = 0; i < cPath.length - 1; ++i)
				{
					String nextAlias = "fp" + i;
					subquery.join(Hql.leftJoin(thisAlias, cPath[i], nextAlias, null, false));
					thisAlias = nextAlias;
				}
				collectionFilter = cPath[cPath.length - 1];
				
				// The following operators should result in "not contains" logic
				Operator subfilterOperator = operator;
				if (operator == Operator.NEQ)
				{
					subfilterOperator = Operator.EQ;
				}
				else if (operator == Operator.NOTIN)
				{
					subfilterOperator = Operator.IN;
				}
				else if (operator == Operator.NOTLIKE)
				{
					subfilterOperator = Operator.LIKE;
				}
				else if (operator == Operator.NOTLIKE_IN)
				{
					subfilterOperator = Operator.LIKE_IN;
				}
				
				boolean notContains = subfilterOperator != operator;
				Operator actualOperator = subfilterOperator;
				String actualValue = getValue();
				if (rewriteNullFilter && actualValue == null && subfilterOperator == Operator.EQ)
				{
					// IN (... WHERE <property> IS NULL) is rewritten as
					// NOT IN (... WHERE <property> LIKE '%') 
					actualOperator = Operator.LIKE;
					actualValue = "%";
					notContains = !notContains;
				}
				
				// Restrict the subquery
				PropertyFilter pp = new PropertyFilter("$" + thisAlias + "." + collectionFilter, 
						actualOperator, actualValue, getValueType());
				subquery.restrict(pp.getRestriction(dc, query));
				
				// Create the main restriction
				if (notContains)
				{
					// "not contains": NOT root-entity IN (subquery)
					restriction = Restrictions.not(Restrictions.in(Hql.alias(query.getRootAlias()), subquery));
				}
				else
				{
					// "contains": root-entity IN (subquery)
					restriction = Restrictions.in(Hql.alias(query.getRootAlias()), subquery);
				}
			}
			else if (property != null && property.startsWith("!sharedTo."))
			{
				// A special filter on on shareable items
				String filterProperty = property.substring("!sharedTo.".length());
				String filterValue = getValue();
				Operator filterOp = null;
				if (operator == Operator.EQ || operator == Operator.NEQ)
				{
					filterOp = Operator.EQ;
				}
				else if (operator == Operator.LIKE || operator == Operator.NOTLIKE)
				{
					filterOp = Operator.LIKE;
				}
				else if (operator == Operator.IN ||  operator == Operator.NOTIN)
				{
					filterOp = Operator.IN;
				}
				else if (operator == Operator.LIKE_IN ||  operator == Operator.NOTLIKE_IN)
				{
					filterOp = Operator.LIKE_IN;
				}
				else
				{
					filterOp = operator;
				}
				boolean sharedTo = operator == filterOp;
				
				Restriction users = null;
				Restriction groups = null;
				Restriction projects = null;
				if (filterValue != null)
				{
					users = new PropertyFilter("$usr." + filterProperty, 
							filterOp, filterValue, valueType).getRestriction(dc, query);
					groups = new PropertyFilter("$grp." + filterProperty, 
							filterOp, filterValue, valueType).getRestriction(dc, query);
					projects = new PropertyFilter("$prj." + filterProperty, 
							filterOp, filterValue, valueType).getRestriction(dc, query);
				}
				restriction = Hql.sharedTo(sharedTo, users, groups, projects);
			}
			else
			{
				Expression propertyExpression = Hql.property(alias, property, indexExpression);
				if (getValueType() == Type.TIMESTAMP)
				{
					propertyExpression = Expressions.toDate(propertyExpression);
				}
				
				if (operator.isListOperator())
				{
					List<Expression> parameters = new ArrayList<Expression>();
					List<Object> objects = getValuesAsObjects(unit, true);
					for (Object obj : objects)
					{
						Expression e = null;
						if (obj != null)
						{
							parameterName = "p" + dc.uniqueRandom();
							Type theValueType = obj instanceof Double ? Type.DOUBLE : getValueType();
							if (theValueType == Type.TIMESTAMP) theValueType = Type.DATE;
							e = Expressions.parameter(parameterName, obj, theValueType);
						}
						parameters.add(e);
					}
					restriction = operator.getRestriction(propertyExpression, parameters.toArray(new Expression[parameters.size()]));
					if (operator.isNegationOperator() && objects.indexOf(null) == -1)
					{
						// <>'foo' should also match null unless null is in the list
						restriction = Restrictions.or(restriction, Restrictions.eq(propertyExpression, null));
					}
				}
				else
				{
					restriction = operator.getRestriction(propertyExpression, parameter);
					if (operator.isNegationOperator() && value != null)
					{
						// <>'foo' should also match null
						restriction = Restrictions.or(restriction, Restrictions.eq(propertyExpression, null));
					}
				}
			}
		}
		return restriction;
	}

	/**
		@since 3.9.2
	*/
	public Restriction getDynamicRestriction(DbControl dc, SqlQuery query)
	{
		Restriction restriction = null;
		String property = getProperty();
		Expression propertyExpression = null;
		if (property.startsWith("£"))
		{
			// Property is reporterListScores collection
			// The value is the ID of the reporter list to
			// filter on
			Integer reporterListId = (Integer)getValueAsObject();
			if (reporterListId != null)
			{
				try
				{
					ReporterList list = ReporterList.getById(dc, reporterListId);
					if (operator == Operator.NEQ)
					{
						restriction = Dynamic.isNotPartOf(list);
					}
					else
					{
						restriction = Dynamic.isPartOf(list);
					}
				}
				catch (ItemNotFoundException ex)
				{}
			}
			return restriction;
		}
		else
		{
			propertyExpression = ItemContext.getDynamicExpression(dc, query, property);
		}
		String parameterName = "p" + dc.uniqueRandom();
		
		String value = getValue();
		Expression parameter = null;
		List<Expression> parameters = null;
		// Experimental - allows expressions like: >ch1
		if (value != null && (valueType == Type.FLOAT || valueType == Type.DOUBLE || valueType == Type.INT || valueType == Type.LONG))
		{
			if (!getOperator().isListOperator() && value.matches("^ch\\d+$"))
			{
				int channel = Integer.parseInt(value.substring(2));
				parameter = Dynamic.column(VirtualColumn.channelRaw(channel));
			}
			else if (getOperator().isListOperator())
			{
				String[] values = value.split("\\"+getValueSeparator());
				for (String s : values)
				{
					if (s.matches("^ch\\d+$"))
					{
						int channel = Integer.parseInt(s.substring(2));
						if (parameters == null)
						{
							parameters = new ArrayList<Expression>();
						}
						parameters.add(Dynamic.column(VirtualColumn.channelRaw(channel)));
					}
				}
			}			
		}
		if (parameter == null && parameters == null)
		{
			if (getOperator().isListOperator())
			{
				List<Object> objects = getValuesAsObjects(unit, true);				
				parameters = objects == null ? null : new ArrayList<Expression>(objects.size());
				for (Object obj : objects)
				{
					Expression e = null;
					if (obj != null)
					{
						parameterName = "p" + dc.uniqueRandom();
						e = Expressions.parameter(parameterName, obj);
					}
					parameters.add(e);
				}				
			}
			else
			{
				parameter = value == null ? null : Expressions.parameter(parameterName, getValueAsObject(unit, true), getValueType());
			}
		}
		if (propertyExpression != null)
		{
			if (parameters != null)
			{
				restriction = operator.getRestriction(propertyExpression, parameters.toArray(new Expression[parameters.size()]));
			}
			else
			{
				restriction = operator.getRestriction(propertyExpression, parameter);
			}
		}
		return restriction;
	}
	
	public String getValueSeparator()
	{
		return value_separator;
	}

}
