/*
	$Id: RealJoin.java 6127 2012-09-14 07:28:35Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Collection;

import net.sf.basedb.core.query.Join;
import net.sf.basedb.core.query.Query;
import net.sf.basedb.core.query.QueryElement;
import net.sf.basedb.core.query.QueryType;
import net.sf.basedb.core.query.JoinType;
import net.sf.basedb.util.EqualsHelper;

/**
	Represents a join from one virtual table in the dynamic database to
	a table in the regular database or between two tables in the
	regular database.
	<code>virtualTable vt JOIN regularTable rt ON rt.id = vt.leftColumn</code>.
	<p>
	Note that the join is always done to the 'id' of the items in the regular table.

	@author Nicklas
	@version 2.0
	@see RealTable
	@base.modified $Date: 2012-09-14 09:28:35 +0200 (Fri, 14 Sep 2012) $
*/
class RealJoin
	implements Join
{
	
	private final JoinType joinType;
	private final RealTable right;
		
	private VirtualTable vLeft;
	private VirtualColumn vLeftColumn;
	
	private RealTable rLeft;
	private String rLeftColumn;
	
	/**
		Create a new join between a virtual and a regular table.
		<code><i>SELECT ... FROM left</i> JOIN right ON right.id = left.column</code>.

		@param joinType The type of join
		@param left The left table to join from, if not specified we join
			from the root table
		@param right The table to join to
		@param leftColumn The column in the left table to join from, we always
			join to the id of the real table
	*/
	RealJoin(JoinType joinType, VirtualTable left, RealTable right, VirtualColumn leftColumn)
	{
		this.joinType = joinType;
		this.right = right;
		this.vLeft = left;
		this.vLeftColumn = leftColumn;
	}

	/**
		Create a new join between two regular tables.
		<code><i>SELECT ... FROM left</i> JOIN right ON right.id = left.column</code>.
	
		@param joinType The type of join
		@param left The left table to join from, if not specified we join
			from the root table
		@param right The table to join to
		@param leftColumn The column in the left table to join from, we always
			join to the id of the real table
	*/
	RealJoin(JoinType joinType, RealTable left, RealTable right, String leftColumn)
	{
		this.joinType = joinType;
		this.right = right;
		this.rLeft = left;
		this.rLeftColumn = leftColumn;
	}

	/*
		From the QueryElement interface
		-------------------------------------------
	*/
	@Override
	public String toQl(Query query, DbControl dc)
		throws BaseException
	{
		if (query.getQueryType() == QueryType.SQL)
		{
			String leftAlias = getLeftTableAlias(query);
			String rightAlias = right.getAlias();
			
			StringBuilder sb = new StringBuilder();
			sb.append(joinType.getJoinExpression()).append(" ").append(right.getQualifiedTableName());
			sb.append(" AS ").append(rightAlias);
			sb.append(" ON ");
			
			String qualifiedLeftColumn = HibernateUtil.quote(getLeftTableColumn());
			sb.append(rightAlias).append(".id = ").append(leftAlias).append(".").append(qualifiedLeftColumn);
			return sb.toString();
		}
		else
		{
			throw new UnsupportedOperationException("RealJoin operation not supported by "+query.getQueryType()+" query.");
		}
	}
	/**
		@return Always null
	*/
	@Override
	public Collection<? extends QueryElement> getChildren()
	{
		return null;
	}
	// -------------------------------------------
	/*
	 	From the Join interface
	 	-------------------------------------------
	*/
	@Override
	public String getThetaJoin(Query query, DbControl dc) 
		throws BaseException 
	{
		if (query.getQueryType() == QueryType.SQL)
		{
			StringBuilder sb = new StringBuilder();
			sb.append(", ").append(right.getQualifiedTableName());
			sb.append(" ").append(HibernateUtil.quote(right.getAlias()));
			return sb.toString();
		}
		else
		{
			throw new UnsupportedOperationException("ThetaJoin operation not supported by "+query.getQueryType()+" query.");
		}
	}
		
	@Override
	public String getThetaJoinCondition(Query query, DbControl dc) 
		throws BaseException 
	{
		if (query.getQueryType() == QueryType.SQL)
		{
			String leftAlias = getLeftTableAlias(query);
			String rightAlias = HibernateUtil.quote(right.getAlias());
			StringBuilder sb = new StringBuilder();

			String qualifiedColumn = HibernateUtil.quote(getLeftTableColumn());
			sb.append(leftAlias).append(".").append(qualifiedColumn);
			sb.append(" = ").append(rightAlias).append(".").append(HibernateUtil.quote("id"));
			if (joinType == JoinType.LEFT) sb.append("(+)");
			return sb.toString();
		}
		else
		{
			throw new UnsupportedOperationException("ThetaJoin operation not supported by "+query.getQueryType()+" query.");
		}
	}
	// -------------------------------------------
	/*
		From the Object class
		-------------------------------------------
	*/
	@Override
	public String toString()
	{
		return joinType.getJoinExpression()+" "+right.getTableName()+" AS "+right.getAlias()+
			" ON "+right.getAlias() + ".id = " + getLeftTableAlias(null) + "." + getLeftTableColumn();
	}
	/**
		This join is equal to another RealJoin if they 
		join between the same tables using the same columns and join types.
	*/
	@Override
	public boolean equals(Object other)
	{
		if (this == other) return true;
		if (other == null || this.getClass() != other.getClass()) return false;
		RealJoin o = (RealJoin)other;
		return vLeft == o.vLeft && rLeft == o.rLeft && right == o.right && joinType == o.joinType && 
			EqualsHelper.equals(vLeftColumn, o.vLeftColumn) && EqualsHelper.equals(rLeftColumn, o.rLeftColumn);
	}
	@Override
	public int hashCode()
	{
		return EqualsHelper.hashCode(vLeft, rLeft, right, joinType, vLeftColumn, rLeftColumn);
	}
	// -------------------------------------------

	private String getLeftTableAlias(Query query)
	{
		String alias = "";
		if (vLeft != null)
		{
			alias = vLeft.getAlias();
		}
		else if (rLeft != null)
		{
			alias = rLeft.getAlias();
		}
		else if (query != null)
		{
			alias = query.getRootAlias();
		}
		return alias;
	}
	
	private String getLeftTableColumn()
	{
		String column = null;
		if (vLeftColumn != null)
		{
			column = vLeftColumn.getColumn();
		}
		else
		{
			column = rLeftColumn;
		}
		return column;
	}
}
