/**
	$Id: QueryParameter.java 6875 2015-04-20 09:08:18Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.query;

import java.util.Collection;

import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.Type;

/**
	Holds information (name, type and value) about a single parameter 
	to a query.
	
	@author Nicklas
	@version 2.9
	@base.modified $Date: 2015-04-20 11:08:18 +0200 (Mon, 20 Apr 2015) $
*/
public class QueryParameter
{

	private final String name;
	private final boolean permanent;
	private Type type;
	private Item itemType;
	private Object value;
	
	/**
		Create a new empty non-permanent query parameter.
		@param name The name of the parameter
	*/
	public QueryParameter(String name)
	{
		this.name = name;
		this.permanent = false;
	}
	
	/**
		Create a new query parameter.
		@param name The name of the parameter
		@param permanent If the parameter should be permanent or not. A
			permanent parameter can't change it's value
		@param type The type of value
		@param value The value
	*/
	public QueryParameter(String name, boolean permanent, Type type, Object value)
	{
		this.name = name;
		this.permanent = permanent;
		this.type = type;
		this.value = value;
	}
	
	/**
		Get the name of the parameter.
	*/
	public String getName()
	{
		return name;
	}
	
	/**
		Check if the parameter is a permanent parameter. A permanent
		paramer can't be changed.
	*/
	public boolean isPermanent()
	{
		return permanent;
	}

	/**
		Get the type of the parameter.
	*/
	public Type getType()
	{
		return type;
	}

	/**
		Changes the type of the parameter. Ignored if
		the parameter is permanent. Also sets the {@link #getItemType()}
		value to null.
		@param type The new type
	*/
	public void setType(Type type)
	{
		if (permanent) return;
		this.type = type;
		this.itemType = null;
	}

	/**
		Get the value of the parameter.
	*/
	public Object getValue()
	{
		return value;
	}
	
	/**
		Change the value of the parameter. Ignored
		if the parameter is permanent.
		@param value The new value
	 */
	public void setValue(Object value)
	{
		if (permanent) return;
		this.value = value;
	}
	
	/**
		The parameter should be an entity of the given item type.
		@return An Item object or null
		@since 2.17
	*/
	public Item getItemType()
	{
		return itemType;
	}
	
	/**
		Changes the type of the parameter. Ignored if
		the parameter is permanent. Also sets the {@link #getType()}
		value to null.
		@param itemType The new type
		@since 2.17
	*/
	public void setItemType(Item itemType)
	{
		if (permanent && this.itemType != null) return;
		this.itemType = itemType;
		this.type = null;
	}
	
	/**
		Validate the parameter value using the type information if
		it has been provided. Null values are always accepted. If no
		type or item type has been set the value is also always 
		accepted. The value can be either a single value or a 
		collection of values. If a type has been set the
		value is validated with {@link Type#isCorrectType(Object)}. If
		an item type has been set it is checked that the value is an
		instance of the given item type.
		
		@since 2.17
	*/
	public void validateValue()
	{
		if (value == null || type == null && itemType == null) return;
		Class<?> entityClass = null;
		if (itemType != null)
		{
			entityClass = itemType.getItemClass();
			if (entityClass == null) entityClass = itemType.getDataClass();
		}
		if (value instanceof Collection)
		{
			Collection<?> c = (Collection<?>)value;
			int i = 0;
			for (Object o : c)
			{
				if (entityClass != null)
				{
					if (!entityClass.isInstance(o))
					{
						throw new InvalidDataException("Incorrect parameter type for '" + 
							name + "[" + i + "]'. Expected " + itemType.name() + "; found " + o);
					}
				}
				else if (!type.isCorrectType(o))
				{
					throw new InvalidDataException("Incorrect parameter type for '" + 
						name + "[" + i + "]'. Expected " + type + "; found " + o.getClass().getName());
				}
				i++;
			}
		}
		else
		{
			if (entityClass != null)
			{
				if (!entityClass.isInstance(value))
				{
					throw new InvalidDataException("Incorrect parameter type for '" + 
						name + "'. Expected " + itemType.name() + "; found " + value);
				}
			}
			else if (!type.isCorrectType(value))
			{
				throw new InvalidDataException("Incorrect parameter type for '" + name +
					"'. Expected " + type + "; found " + value.getClass().getName());
			}
		}
	}

	
	/*
		From the Object class
	 	-----------------------
	*/
	@Override
	public int hashCode()
	{
		return name.hashCode();
	}
	
	@Override
	public boolean equals(Object o)
	{
		if (this == o) return true;
		if (o == null) return false;
		if (o.getClass() != getClass()) return false;
		QueryParameter qp = (QueryParameter)o;
		return name.equals(qp.name);
	}
	
	@Override
	public String toString()
	{
		return "QueryParameter(name="+name +"; type=" + type + "; value=" + value + ")";
	}
	// ----------------------------
}
