/*
	$Id: Selects.java 6881 2015-04-21 07:59:42Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.query;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.InvalidUseOfNullException;

import java.util.regex.Pattern;

/**
	A factory class to create select expressions.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2015-04-21 09:59:42 +0200 (Tue, 21 Apr 2015) $
*/
public class Selects
{

	/**
		An alias can only contain the characters a-z, A-Z or 0-9.
	*/
	public static final Pattern ALIAS_REGEXP = Pattern.compile("[a-zA-z0-9]+");

	/**
		Same as <code>expression(e1, alias, false)</code>.
	 	@param e1 The expression to select. Null is not allowed
	 	@param alias The alias to give the expression. Only characters a-z, A-Z and 0-9 are allowed. 
	 		Null is allowed.
	 	@return A selection query element
	 	@throws InvalidDataException If the required parameter is null or 
	 		if illegal characters is used in <code>alias</code> 
		@see #expression(Expression, String, boolean)
	*/
	public static final Select expression(Expression e1, String alias)
		throws InvalidDataException
	{
		if (e1 == null) throw new InvalidUseOfNullException("e1");
		if (alias != null && !ALIAS_REGEXP.matcher(alias).matches()) 
		{
			throw new InvalidDataException("Alias '"+alias+"' has one or more invalid characters. Only a-z, A-Z and 0-9 is allowed.");
		}
		return new ExpressionSelect(e1, alias, false);
	}
	
	/**
		Create a selection from an expression, optionally giving it an alias. 
		The alias can be used when getting the result to find the column
		index. See {@link SqlResultIterator#getIndex(String)}.

	<pre class="code">
SELECT <i>&lt;expression&gt;</i> AS <i>&lt;alias&gt;</i>
FROM ...
	</pre>
		<p>
		The useInDistinctCount parameter affects the {@link Query#count(DbControl)}
		method, if the {@link Query#setDistinct(boolean)} has been set to true. In this
		case only expressions which has been created for use with the distinct function
		are used in the selection list:
		<pre class="code">
SELECT COUNT(DISTINCT <i>&lt;expression&gt;</i>)
FROM ...
	</pre>
		<p>
				
		@param e1 The expression to select
		@param alias The alias to give the expression
		@param useInDistinctCount If the expression should be used when counting distinct values
		@return A selection query element
		@throws InvalidDataException If the expression is null or if the
			alias contains invalid characters 
	*/	
	public static final Select expression(Expression e1, String alias, boolean useInDistinctCount)
		throws InvalidDataException
	{
		if (e1 == null) throw new InvalidUseOfNullException("e1");
		if (alias != null && !ALIAS_REGEXP.matcher(alias).matches()) 
		{
			throw new InvalidDataException("Alias '"+alias+"' has one or more invalid characters. Only a-z, A-Z and 0-9 is allowed.");
		}
		return new ExpressionSelect(e1, alias, useInDistinctCount);
	}
	
}
