/**
	$Id: AnnotationSnapshot.java 7703 2019-04-11 07:31:26Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.snapshot;

import java.io.IOException;
import java.io.Serializable;
import java.sql.Date;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;

import net.sf.basedb.core.Annotatable;
import net.sf.basedb.core.Annotation;
import net.sf.basedb.core.AnnotationSet;
import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.Unit;
import net.sf.basedb.core.data.AnnotationData;
import net.sf.basedb.util.filter.Filter;
import net.sf.basedb.util.units.UnitConverter;

/**
	Stores information about a single primary or inherited 
	annotation or an inherited annotation set. Actual 
	annotation values are only stored for primary annotations.
	<p>
	<b>IMPORTANT NOTE TO DEVELOPERS!!</b>
	<p>
	Do not forget to synchronize all serialization code and to increment
	the {@link AnnotationSetSnapshot#FILE_VERSION} in case the number of variables 
	that needs serialization in this class or in the {@link AnnotationSetSnapshot} 
	changes. 

	@author Nicklas
	@version 2.14
	@base.modified $Date: 2019-04-11 09:31:26 +0200 (Thu, 11 Apr 2019) $
*/
public class AnnotationSnapshot
	implements Serializable
{

	/**
		Create a comparator for sorting annotation snapshots by the name 
		of the annotation type.
		@since 3.6
	*/
	public static Comparator<AnnotationSnapshot> sortByAnnotationType(DbControl dc) 
	{
		return new AnnotationTypeComparator(dc);
	}
	
	private static final long serialVersionUID = 1167459472860706199L;

	private transient Annotation.Source source;
	private transient int annotationId;
	private transient int annotationSetId;
	private transient int annotationTypeId;
	private transient int inheritedFromId;
	private transient int inheritedFromSetId;
	private transient int unitId;
	private transient int projectId;
	private transient int overrideId;
	private transient Type valueType;
	private transient long lastUpdate;
	private transient List<? extends Serializable> values;
	private transient int itemId;
	private transient Item itemType;
	private transient AnnotationSnapshot inheritedFrom;
	
	/**
		Create a new, empty annotation snapshot.
	*/
	public AnnotationSnapshot()
	{}
	
	/**
		Initialize the snapshot with information from the given annotation.
		@param a An annotation
	*/
	@SuppressWarnings("unchecked")
	void init(AnnotationData a)
	{
		this.source = Annotation.Source.values()[a.getSource()];
		this.annotationId = a.getId();
		this.annotationSetId = a.getAnnotationSet().getId();
		this.annotationTypeId = a.getAnnotationType().getId();
		AnnotationData inheritedFrom = a.getInheritedFrom();
		if (inheritedFrom != null)
		{
			this.inheritedFromId = inheritedFrom.getId();
			this.inheritedFromSetId = inheritedFrom.getAnnotationSet().getId();
		}
		this.unitId = a.getUnit() == null ? 0 : a.getUnit().getId();
		this.projectId = a.getProjectId();
		this.overrideId = a.getOverrideId();
		this.valueType = Type.fromValue(a.getAnnotationType().getValueType());
		this.lastUpdate = a.getLastUpdate() == null ? 0 : a.getLastUpdate().getTime();
		if (a.getValues() != null)
		{
			this.values = new ArrayList<Serializable>((List<Serializable>)a.getValues().getValues());
		}
	}

	
	/**
		Get the source of the annotation.
		@since 3.6
	*/
	public Annotation.Source getSource()
	{
		return source;
	}
	
	/**
		Get the id of the annotation this is a snapshot of.
		@return The id of this annotation
		@since 3.6
	*/
	public int getThisAnnotationId()
	{
		return annotationId;
	}
	
	/**
		Get the ID of the annotation this annotation is inherited from. 
		NOTE! This may return 0 for cloned annotations that has lost their
		parent annotation.
		
		@return The id of the inherited annotation or 0 if this is primary
			annotation (or a cloned annotation that has lost it's parent)
		@since 3.6
	*/
	public int getInheritedAnnotationId()
	{
		return inheritedFromId;
	}
	
	/**
		Get the annotation this is a snapshot of.
		@return An Annotation object
		@since 3.6
	*/
	public Annotation getThisAnnotation(DbControl dc)
	{
		return Annotation.getById(dc, annotationId);
	}
	
	/**
		Get the parent annotation this annotation is inherited from. 
		NOTE! This may return null for cloned annotations that has lost their
		parent annotation.
	
		@return The inherited annotation or null if this is primary
			annotation (or a cloned annotation that has lost it's parent)
		@since 3.6
	*/
	public Annotation getInheritedAnnotation(DbControl dc)
	{
		return inheritedFromId == 0 ? null : Annotation.getById(dc, inheritedFromId);
	}

	
	/**
		Check if the logged in user has the request permission on this annotation
		or annotation set.
		@param dc The DbControl to use for database access
		@param permission The requested permission
		@return TRUE if the logged in user has the requested permission, FALSE otherwise
		@since 3.1.1
	*/
	public boolean hasPermission(DbControl dc, Permission permission)
	{
		if (annotationTypeId != 0 && permission == Permission.READ)
		{
			// Special case to avoid loading each Annotation and it's associated item from
			// the database just for the permission check which is only based on the permission
			// from the annotation type in any case
			// See tickets #1693 and #1685
			try
			{
				AnnotationType at = getAnnotationType(dc);
				return true;
			}
			catch (RuntimeException ex)
			{}
		}
		else
		{
			try
			{
				Annotation a = getThisAnnotation(dc);
				return a.hasPermission(permission);
			}
			catch (RuntimeException ex)
			{}
		}
		return false;
	}

	/**
		Get the id of the annotation set this is a snapshot of.
		@return The id of this annotation set
		@since 3.6
	*/
	public int getThisAnnotationSetId()
	{
		return annotationSetId;
	}
	
	/**
		Get the ID of the annotation set this annotation is inherited from. 
		NOTE! This may return 0 for cloned annotations that has lost their
		parent annotation.
		
		@return The id of the inherited annotation set or 0 if this is primary
			annotation (or a cloned annotation that has lost it's parent)
		@since 3.6
	*/
	public int getInheritedAnnotationSetId()
	{
		return inheritedFromSetId;
	}
	
	/**
		Get the annotation set this annotation belongs to.
		@since 3.6
	*/
	public AnnotationSet getThisAnnotationSet(DbControl dc)
	{
		return AnnotationSet.getById(dc, annotationSetId);
	}
	
	/**
		Get the annotation set this annotation is inherited from. 
		NOTE! This may return null for cloned annotations that has lost their
		parent annotation.
		@since 3.6
	*/
	public AnnotationSet getInheritedAnnotationSet(DbControl dc)
	{
		return inheritedFromSetId == 0 ? null : AnnotationSet.getById(dc, inheritedFromSetId);
	}

	/**
		Get the id of the annotation type.
		@return The annotation type id, or 0 if this snapshot is from
			an inherited annotation set
	*/
	public int getAnnotationTypeId()
	{
		return annotationTypeId;
	}

	/**
		Get the annotation type.
		@return An AnnotationType object or null if this snapshot is
			from an inherited annotation set
	*/
	public AnnotationType getAnnotationType(DbControl dc)
	{
		return annotationTypeId == 0 ? null :AnnotationType.getById(dc, getAnnotationTypeId());
	}
	
	/**
		Get a snapshot for the annotation this is inherited from. NOTE! This is
		only available for inherited annotations returned by
		{@link SnapshotManager#findAnnotations(DbControl, AnnotationSetSnapshot, Filter, boolean)}
		but may be missing for cloned annotations if the parent has been deleted.
		@since 3.6
	*/
	public AnnotationSnapshot getInheritedFrom()
	{
		return inheritedFrom;
	}

	/**
		Get the id of the project that this project-specific annotation
		belongs to.
		@return The project id, or 0 if this a default annotation
		@since 3.10
	*/
	public int getProjectId()
	{
		return projectId;
	}
	
	/**
		Get the id of the default annotation that this project-specific
		annotation is overriding.
		@since 3.10
	*/
	public int getOverrideId()
	{
		return overrideId;
	}
	
	/**
		Get the id of the unit that the annotation values should
		use when displayed.
		@since 3.6
	*/
	public int getThisUnitId()
	{
		return unitId;
	}

	/**
		Get the unit the annotatation values should use for display.
		NOTE! The unit is only available for primary and cloned annotations. Use 
		{@link #getActualUnit(DbControl)} or {@link #getInheritedFrom()}.getThisUnit() to get 
		the unit also for inherited annotations.
		
		@return A Unit object or null if this annotation doesn't use units
		@since 3.6
	*/
	public Unit getThisUnit(DbControl dc)
	{
		return unitId == 0 ? null : Unit.getById(dc, unitId);
	}
	
	/**
		Get the unit that is most relevant for this annotation.
		Eg. a unit that is defined by this annotation (primary and cloned) are
		used before inherited unit.
		@since 3.6
	 */
	public Unit getActualUnit(DbControl dc)
	{
		int id = inheritedFrom != null && unitId == 0 ? inheritedFrom.unitId : unitId;
		return id == 0 ? null : Unit.getById(dc, id);
	}

	/**
		Get the annotation values for this annotation. NOTE! The values are only
		available for primary and cloned annotations. Use {@link #getActualValues()}
		or {@link #getInheritedFrom()}.getThisValues() to get values also for inherited
		annotations.
		@return A list with the values or null
	*/
	@SuppressWarnings("unchecked")
	public <T extends Serializable> List<T> getThisValues()
	{
		return (List<T>)values;
	}
	
	/**
		Get the annotation values that are most relevant for this annotation.
		Eg. values that are defined by this annotation (primary and cloned) are
		used before inherited values.
		@since 3.6
	*/
	@SuppressWarnings("unchecked")
	public <T extends Serializable> List<T> getActualValues()
	{
		return (List<T>)(inheritedFrom != null && values == null ? inheritedFrom.values : values);
	}
		
	/**
		Get the annotations values, optionally converted to some specific unit
		and type of values.
		
		NOTE! To create a unit converter that converts that values to the unit
		specified by {@link #getActualUnit(DbControl)}:
		
		this.getActualUnit(dc).getUnitConverter(this.getAnnotationType(dc).getDefaultUnit())
		
		@param converter A converter for converting (numerical) values,
			or null to not convert the values
		@param valueType The return type of the converted values (ignored if
			no converter is specified)
		@return A list with the values
		@since 3.6
	*/
	@SuppressWarnings("unchecked")
	public <T extends Serializable> List<T> getActualValues(UnitConverter converter, Type valueType)
	{
		List<T> vals = getActualValues();
		if (converter == null) return vals;
		List<T> convertedValues = new ArrayList<>(vals.size());
		for (T value : vals)
		{
			if (value instanceof Number)
			{
				value = (T)valueType.convertNumber(converter.convertToSpecificUnit(((Number)value).doubleValue()));
			}
			convertedValues.add(value);
		}
		return convertedValues;
	}
	
	/**
		Get the date+time this annotation was last updated.
		@return The date or null if this annotation doesn't contain any values
		@since 3.6
	*/
	public Date getThisLastUpdate()
	{
		return lastUpdate == 0 ? null : new Date(lastUpdate);
	}
	
	/**
		A primary or inherited annotation is always up-to-date. A cloned annotation
		is up-to-date if the link to the source annotation exists and the source
		annotation has a date equal to or before the cloned annotation.
		@since 3.6
	*/
	public boolean isUpToDate()
	{
		boolean upToDate = true;
		if (source == Annotation.Source.CLONED)
		{
			AnnotationSnapshot inherited = getInheritedFrom();
			Date cloned = getThisLastUpdate();
			Date source = inherited != null ? inherited.getThisLastUpdate() : null;
			upToDate = cloned != null && source != null && !source.after(cloned);
		}
		return upToDate;
	}
	
	/**
	 	Get the id of the item this annotation belongs to. NOTE! this value
	 	is only available on annotations returned from the call to:
		{@link SnapshotManager#findAnnotations(DbControl, AnnotationSetSnapshot, Filter, boolean)}
	*/
	public int getThisItemId()
	{
		return itemId;
	}

	/**
		Get the item that this annotation belongs to.
		@return The item, or null if the item is not know
		@since 3.6
	*/
	public Annotatable getThisItem(DbControl dc)
	{
		return itemId == 0 ? null : itemType.getById(dc, itemId);
	}
	
	/**
		Get the item type of the item this annotation belongs to. 
		NOTE! this value is only available on annotations returned from the call to:
		{@link SnapshotManager#findAnnotations(DbControl, AnnotationSetSnapshot, Filter, boolean)}
	
		@return The item type, or null if not known
		@since 3.6
	*/
	public Item getThisItemType()
	{
		return itemType;
	}
	
	void setItem(int itemId, Item itemType)
	{
		this.itemId = itemId;
		this.itemType = itemType;
	}
	
	/**
		Set the primary annotation for an inherited annotation.
	*/
	void setInheritedFrom(AnnotationSnapshot inheritedFrom)
	{
		this.inheritedFrom = inheritedFrom;
		if (inheritedFrom == null)
		{
			this.inheritedFromId = 0;
			this.inheritedFromSetId = 0;
		}
	}

	void serializeObject(java.io.ObjectOutputStream out)
		throws IOException
	{
		writeObject(out);
	}
	
	private void writeObject(java.io.ObjectOutputStream out)
		throws IOException
	{
		out.writeInt(source.ordinal());
		out.writeInt(annotationId);
		out.writeInt(annotationSetId);
		out.writeInt(annotationTypeId);
		out.writeInt(inheritedFromId);
		out.writeInt(inheritedFromSetId);
		out.writeInt(projectId);
		out.writeInt(overrideId);
		out.writeInt(unitId);
		out.writeInt(valueType == null ? 0 : valueType.getValue());
		out.writeLong(lastUpdate);
		out.writeInt(values == null ? 0 : values.size());
		if (values != null)
		{
			for (Serializable tmp : values)
			{
				valueType.writeValue(tmp, out);
			}
		}
	}
	
	void deserializeObject(java.io.ObjectInputStream in)
		throws IOException, ClassNotFoundException
	{
		readObject(in);
	}
	private void readObject(java.io.ObjectInputStream in)
		throws IOException, ClassNotFoundException
	{
		source = Annotation.Source.values()[in.readInt()];
		annotationId = in.readInt();
		annotationSetId = in.readInt();
		annotationTypeId = in.readInt();
		inheritedFromId = in.readInt();
		inheritedFromSetId = in.readInt();
		projectId = in.readInt();
		overrideId = in.readInt();
		unitId = in.readInt();
		valueType = Type.fromValue(in.readInt());
		lastUpdate = in.readLong();
		int numValues = in.readInt();
		if (numValues > 0)
		{
			List<Serializable> values = new ArrayList<Serializable>(numValues);
			for (int i = 0; i < numValues; ++i)
			{
				values.add(valueType.readValue(in));
			}
			this.values = values;
		}
	}
	
	/**
		Comparator implementation for sorting snapshots
		by name of annotation type. If the names are
		sorted by source type of the annotation and then
		on the id of the annotation.
	*/
	static class AnnotationTypeComparator
		implements Comparator<AnnotationSnapshot>
	{
		private final DbControl dc;
		
		AnnotationTypeComparator(DbControl dc) 
		{
			this.dc = dc;
		}

		@Override
		public int compare(AnnotationSnapshot o1, AnnotationSnapshot o2) 
		{
			String n1 = o1.getAnnotationType(dc).getName();
			String n2 = o2.getAnnotationType(dc).getName();
			
			int result = n1.compareTo(n2);
			if (result == 0)
			{
				result = o1.getSource().ordinal() - o2.getSource().ordinal();
				if (result == 0)
				{
					result = o1.getThisAnnotationId() - o2.getThisAnnotationId();
				}
			}
			return result;
		}
	}
}
