/**
	$Id: InputStreamTracker.java 7876 2020-11-09 13:40:05Z nicklas $

	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util;

import java.io.IOException;
import java.io.InputStream;

/**
	This class is used to keep track of the number of bytes read from the
	underlying input stream. This is very useful for code that needs to 
	reporter progress but are having difficulties to know the number of
	processed bytes.

	@author nicklas
	@version 2.4
	@base.modified $Date: 2020-11-09 14:40:05 +0100 (Mon, 09 Nov 2020) $
*/
public class InputStreamTracker
	extends InputStream
{

	/**
		The source input stream to read from.
	*/
	private final InputStream in;
	
	/**
		A scale factor to apply when counting bytes.
	*/
	private final ScaleFactor scaleFactor;
	
	/**
		The number of bytes that has been read from the underlying 
		input stream so far.
	*/
	private long numRead;
	
	/**
		The number of bytes that had been read when {@link #mark(int)}
		was called.
	*/
	private long numReadAtMark;
	
	/**
		Create a new input stream tracker.
		@param in The input stream to read from
	*/
	public InputStreamTracker(InputStream in)
	{
		this(in, null);
	}
	/**
		Create a new input stream tracker.
		@param in The input stream to read from
		@param scaleFactor A scale factor to apply when counting bytes
		@since 3.17.1
	*/
	public InputStreamTracker(InputStream in, ScaleFactor scaleFactor)
	{
		if (in == null) throw new NullPointerException("in");
		this.in = in;
		this.scaleFactor = scaleFactor == null ? new IdentityScaleFactor() : scaleFactor;
	}
	/*
		From the InputStream class
		-------------------------------------------
	*/
	@Override
	public int available() 
		throws IOException
	{
		return in.available();
	}
	@Override
	public void close() 
		throws IOException
	{
		in.close();
	}
	@Override
	public synchronized void mark(int readlimit)
	{
		in.mark(readlimit);
		this.numReadAtMark = numRead;
	}
	@Override
	public boolean markSupported()
	{
		return in.markSupported();
	}
	@Override
	public int read() 
		throws IOException
	{
		int result = in.read();
		if (result != -1) ++numRead;
		return result;
	}
	@Override
	public int read(byte[] b, int off, int len) 
		throws IOException
	{
		int result = in.read(b, off, len);
		if (result > 0) numRead += result;
		return result;
	}
	@Override
	public int read(byte[] b) 
		throws IOException
	{
		return read(b, 0, b.length);
	}
	@Override
	public synchronized void reset() 
		throws IOException
	{
		in.reset();
		numRead = numReadAtMark;
	}
	@Override
	public long skip(long n)
		throws IOException
	{
		long result = in.skip(n);
		if (result > 0) numRead += result;
		return result;
	}
	// -------------------------------------------

	/**
		Get the number of bytes read or skipped from the input stream.
		If the underlying stream supports {@link #reset()} this will also
		reset the number of read bytes to the latest {@link #mark(int)}
		position.
		@return The number of bytes that has been read so far
	*/
	public long getNumRead()
	{
		return scaleFactor.scale(numRead);
	}

	/**
		A scale factor can be used to match number of read bytes to something
		that better resembles actual progress.
		@since 3.17.1
	*/
	public static interface ScaleFactor
	{
		public long scale(long numRead);
	}
	
	/**
		Do not scale number of bytes.
	*/
	public static class IdentityScaleFactor
		implements ScaleFactor
	{
		@Override
		public long scale(long numRead)
		{
			return numRead;
		}
	}
	
	/**
		Apply a linear scale factor so that when we have read N actual bytes
		the getNumRead() method return N*scale bytes.
	*/
	public static class LinearScaleFactor
		implements ScaleFactor
	{
		private final float scale;
		public LinearScaleFactor(float scale)
		{
			this.scale = scale;
		}
		
		@Override
		public long scale(long numRead)
		{
			return (long)(scale * numRead);
		}
	}
	
	/**
		Apply a curved scale factor that has a max upper limit. To begin
		with each byte is counted as 1, but the the more bytes that are read
	 	the scale factor is changed along curved path so that we get closer
		and closer to the max limit but never over.
	*/
	public static class CurvedScaleFactor
		implements ScaleFactor
	{
		private final double max;
		public CurvedScaleFactor(long max)
		{
			this.max = max;
		}
		
		@Override
		public long scale(long numRead)
		{
			return (long)(max * -Math.expm1(-numRead/max));
		}
	}
}
