package net.sf.basedb.util;

import net.sf.basedb.core.AbsoluteProgressReporter;

/**
	A progress reporter implementation that has a "side-effect" 
	that limits the transfer rate. The parent progress reporter
	is optional.
	
	@author nicklas
	@since 3.2
*/
public class TransferRateProgressReporter
	implements AbsoluteProgressReporter 
{

	private final AbsoluteProgressReporter parent;
	private final int maxTicksPerSecond;
	private final int maxDelay;
	private long startTime;
	
	/**
		Create a new progress reporter. Max delay per progress update is 8 seconds.
		
		@param parent An optional parent progress reporter
		@param maxTicksPerSecond Max allowed ticks-per-second, if the rate goes above this,
			the {@link #displayAbsolute(long, String)} method will wait a short time
			before returning
	*/
	public TransferRateProgressReporter(AbsoluteProgressReporter parent, int maxTicksPerSecond)
	{
		this(parent, maxTicksPerSecond, 8000);
	}
	
	
	/**
		Create a new progress reporter. Max delay per progress update is 8 seconds.
		
		@param parent An optional parent progress reporter
		@param maxTicksPerSecond Max allowed ticks-per-second, if the rate goes above this,
			the {@link #displayAbsolute(long, String)} method will wait a short time
			before returning
		@param maxDelay The max delay in milliseconds for each call to {@link #displayAbsolute(long, String)}
	*/
	public TransferRateProgressReporter(AbsoluteProgressReporter parent, int maxTicksPerSecond, int maxDelay)
	{
		this.parent = parent;
		this.maxTicksPerSecond = maxTicksPerSecond;
		this.maxDelay = maxDelay;
	}
	
	
	/*
		From the AbsoluteProgressReporter interface
		-------------------------------------------
	*/
	/**
		Forward the call to the parent if one exists. There is no
		rate control in this method.
	*/
	@Override
	public void display(int percent, String message) 
	{
		if (parent != null) parent.display(percent, message);
	}

	/**
		Forward the call to the parent if one exists.
	*/
	@Override
	public void append(String message)
	{
		if (parent != null) parent.append(message);
	}

	/**
		Forward the call to the parent if one exists. Then
		check the number of completed ticks and time since start
		against the max transfer rate. If the max rate has been exceeded
		pause the thread for a while.
	*/
	@Override
	public void displayAbsolute(long completed, String message) 
	{
		if (parent != null)	parent.displayAbsolute(completed, message);
		
		if (startTime == 0) 
		{
			startTime = System.currentTimeMillis();
		}
		else 
		{
			// Calculate actual time and expected time in milliseconds
			long actualTimeSinceStart = System.currentTimeMillis() - startTime;
			long expectedTimeSinceStart = 1000 * completed / maxTicksPerSecond;
			// If actual time is less than expected (with some slack) pause the thread 
			long delay = expectedTimeSinceStart - actualTimeSinceStart;
			if (delay > 1000)
			{
				// Do not pause more than 8 seconds
				if (delay > 8000) delay = 8000; 
				try
				{
					java.lang.Thread.sleep(delay);
				}
				catch (InterruptedException ex)
				{}
			}
		}
	}

}
