/**
	$Id: AnnotationAssayField.java 5373 2010-08-03 09:11:54Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.export.spotdata;

import java.util.Collection;

import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.snapshot.SnapshotManager;
import net.sf.basedb.util.BioAssaySetUtil;
import net.sf.basedb.util.EqualsHelper;
import net.sf.basedb.util.formatter.Formatter;

/**
	Assay field implementation that exports the annotations of a
	bioassay. Before it can be used an {@link AnnotationType}
	must be configured. One instance of this class is considered to
	be equal to another if they are configured with the same annotation
	type and title.

	@author Nicklas
	@version 2.12
	@base.modified $Date: 2010-08-03 11:11:54 +0200 (Tue, 03 Aug 2010) $
*/
public class AnnotationAssayField
	implements AssayField
{
	private AnnotationType at;
	private String title;
	private Formatter<?> formatter;
	private SnapshotManager snapshotManager;
	
	/**
		Create a new annotation field. Call {@link #setAnnotationType(AnnotationType)}
		before use.
	*/
	public AnnotationAssayField()
	{}
	
	/*
		From the AssayField interface
		-------------------------------
	*/
	/**
		Returns the name of the configured annotation type, or the empty
		string if not configured.
	 */
	@Override
	public String getTitle()
	{
		if (title != null) 
		{
			return title;
		}
		else if (at != null)
		{
			return at.getName();
		}
		return "";
	}
	/**
		Return all annotation values that are related to the specified
		bioassay, or null if not configured.
	*/
	@Override
	public Collection<?> getAssayValue(DbControl dc, BioAssay ba)
	{
		return at == null ? null : BioAssaySetUtil.getAnnotationValues(dc, snapshotManager, ba, at);
	}
	/**
		Return the value type of the annotation type, or null if not
		configured.
		@since 2.15
	*/
	@Override
	public Type getType()
	{
		return at == null ? null : at.getValueType();
	}
	@Override
	public Formatter<?> getFormatter()
	{
		return formatter;
	}
	/**
		@return Always TRUE
	*/
	@Override
	public boolean isAnnotation()
	{
		return true;
	}
	// -----------------------------------
	
	/**
		An assay field is equal to another assay field if it uses the same
		annotation type and has the same title.
		@since 2.15
	*/
	@Override
	public boolean equals(Object obj)
	{
		if (this == obj) return true;
		if (!(obj instanceof AnnotationAssayField)) return false;
		AnnotationAssayField other = (AnnotationAssayField)obj;
		return EqualsHelper.equals(this.getAnnotationType(), other.getAnnotationType()) 
			&& EqualsHelper.equals(this.getTitle(), other.getTitle());
	}

	/**
		@since 2.15
	*/
	@Override
	public int hashCode()
	{
		return 13 * EqualsHelper.hashCode(getAnnotationType()) + 
			EqualsHelper.hashCode(getTitle());
	}
	/**
	 	@since 2.15
	*/
	@Override
	public String toString()
	{
		return "AnnotationAssayField(title=" + getTitle() + "; annotation type=" + getAnnotationType() + ")";
	}
	
	/**
		Set the annotation type of the annotations that should be
		exported in this field.
	*/
	public void setAnnotationType(AnnotationType at)
	{
		this.at = at;
	}

	/**
		Get the configured annotation type.
	*/
	public AnnotationType getAnnotationType()
	{
		return at;
	}
	
	/**
		Set a formatter to use for formatting the exported values.
	*/
	public void setFormatter(Formatter<?> formatter)
	{
		this.formatter = formatter;
	}
	
	/**
		Set the snapshot manager that should be used to load annotation values.
		@since 2.14
	*/
	public void setSnapshotManager(SnapshotManager snapshotManager)
	{
		this.snapshotManager = snapshotManager;
	}
	
	/**
		Set a custom title for the assay field.
		@param title The custom title or null to use the name
		of the annotation type
	*/
	public void setTitle(String title)
	{
		this.title = title;
	}

}
