/**
	$Id: ValidatingFileSetMember.java 5623 2011-05-06 12:09:13Z nicklas $

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.fileset;

import net.sf.basedb.core.FileSetMember;

/**
	This class is a proxy used for holding information about 
	a file set member that is currently undergoing validation.

	@author Nicklas
	@since 3.0
	@base.modified $Date: 2011-05-06 14:09:13 +0200 (Fri, 06 May 2011) $
*/
public class ValidatingFileSetMember
{

	private final FileSetMember member;
	
	private boolean hasBeenAccepted;
	private boolean isValidSoFar;
	private String message;
	
	/**
	 	Create a new instance for the given real member file.
	*/
	public ValidatingFileSetMember(FileSetMember member)
	{
		this.member = member;
		this.hasBeenAccepted = false;
		this.isValidSoFar = true;
	}
	
	/**
		Get the file set member that this proxy represents.
	*/
	public FileSetMember getMember()
	{
		return member;
	}
	
	/**
		Has this member been accepted for validation by at
		least one validator? When the validation has ended,
		the core will update the validation status only for
		members that was accepted at least once.
	*/
	public boolean hasBeenAccepted()
	{
		return hasBeenAccepted;
	}

	/**
		Is the member considered valid so far in the validation process?
		Members that has been marked as invalid by a validator will not 
		be processed by other validators.
	*/
	public boolean isValidSoFar()
	{
		return isValidSoFar;
	}
	
	/**
		Get the error message if the file failed validation.
	*/
	public String getMessage()
	{
		return message;
	}
	
	/**
		Check if the member is accepted by the given validator.
		Members that have already been validated and marked as
		invalid will never be accepted by other actions.
		
		@param action The next validation action
		@return An {@link Accept} value or null if the file
			wasn't accepted
	*/
	public Accept isAcceptedByValidator(ValidationAction action)
	{
		Accept accept = null;
		if (!hasBeenAccepted || isValidSoFar)
		{
			accept = action.acceptFile(member);
			hasBeenAccepted |= (accept != null);
		}
		return accept;
	}
	
	/**
		Set the result of the validation for this file.
		
		@param isValid TRUE if the file passed validation
		@param message A message about the validation
	*/
	public void setResult(boolean isValid, String message)
	{
		this.isValidSoFar = isValid;
		this.message = message;
	}

	
}
