/**
	$Id$

	Copyright (C) 2015 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.listable;

import java.util.HashSet;
import java.util.Set;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Extract;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;

/**
	Collect extracts seen on derived bioassays and raw bioassays that
	has been loaded by other transformers. Use the {@link #collectFromDerivedBioAssays(boolean)}
	and {@link #collectFromRawBioAssays()} methods to create transformers that can be inserted
	into the chain of {@link SourceItemTransformerChain}.
	
	@author Nicklas
	@since 3.5
*/
public class CollectExtracts 
{
	
	private final Set<Integer> collected;
	private final Set<Integer> parents;
	private final Set<Integer> children;
	private boolean needWrapper = true;
	
	/**
		Create a new derived bioassay to physical bioassay transformer.
	*/
	public CollectExtracts() 
	{
		collected = new HashSet<Integer>();
		parents = new HashSet<Integer>();
		children = new HashSet<Integer>();
	}
	
	/**
		Clear all collected extracts in order to prepare for a new
		collection.
		@since 3.17
	*/
	void resetCollected()
	{
		collected.clear();
		parents.clear();
		children.clear();
	}
	
	/**
		Wraps the parent transformer with a ResetCollected transformer the
		first time this method is called. If called more than once, the
		parent is return unwrapped.
		@since 3.17
	*/
	private SourceItemTransformer wrapForReset(SourceItemTransformer transformer)
	{
		if (needWrapper) 
		{
			needWrapper = false;
			transformer = new ResetCollected(transformer);
		}
		return transformer;
	}
	
	/**
		The created transformer collect the ID:s of all extracts linked from
		the source derived bioassays and then return the same set of derived
		bioassays again. If loadParents is set, extract ID:s will also
		be collected from all parent derived bioassays (as loaded by
		{@link DerivedBioAssayToParentDerivedBioAssayTransformer}).
	*/
	public SourceItemTransformer collectFromDerivedBioAssays(boolean loadParents)
	{
		return wrapForReset(new CollectFromDerivedBioAssays(loadParents));
	}

	/**
		The created transformer collect the ID:s of all extracts linked from
		the source raw bioassays and then return the same set of raw bioassays
		again.
	*/
	public SourceItemTransformer collectFromRawBioAssays()
	{
		return wrapForReset(new CollectFromRawBioAssays());
	}
	
	/**
		The created transformer collect the ID:s of all extracts in the 
		source collection and return same set of extracts again.
	*/
	public SourceItemTransformer collectFromExtracts()
	{
		return wrapForReset(new CollectFromExtracts());
	}
	
	/**
		The created transformer collect the ID:s of all parent
		extracts to the already collected extracts and store them in the
		{@link #getCollectedParents()}. If the 'includeCollected' flag
		is set, the collected extracts are included among the parents.
	*/
	public SourceItemTransformer collectParentsOfCollected(boolean includeCollected)
	{
		return wrapForReset(new CollectParents(true, includeCollected));
	}
	
	/**
		The created transformer collect the ID:s of all parent
		extracts to the source extracts and store them in the
		{@link #getCollectedParents()}. If the 'includeSource' flag
		is set, the source extracts are included among the parents.
	*/
	public SourceItemTransformer collectParentsOfSource(boolean includeSource)
	{
		return wrapForReset(new CollectParents(false, includeSource));
	}
	
	
	/**
		The created transformer collect the ID:s of all child
		extracts to the already collected extracts and store them in the
		{@link #getCollectedChildren()}. If the 'includeCollected' flag
		is set, the collected extracts are included among the children.
	*/
	public SourceItemTransformer collectChildrenOfCollected(boolean includeCollected)
	{
		return wrapForReset(new CollectChildren(true, includeCollected));
	}
	
	/**
		The created transformer collect the ID:s of all child
		extracts to the source extracts and store them in the
		{@link #getCollectedChildren()}. If the 'includeSource' flag
		is set, the source extracts are included among the parents.
	*/
	public SourceItemTransformer collectChildrenOfSource(boolean includeSource)
	{
		return wrapForReset(new CollectChildren(false, includeSource));
	}
	
	/**
		Get all extracts that has been collected.
	*/
	public Set<Integer> getCollected()
	{
		return collected;
	}
	
	/**
		Get parent extracts that was collected by the
		{@link #collectParentsOfSource(boolean)} or 
		{@link #collectParentsOfCollected(boolean)} collector.
	*/
	public Set<Integer> getCollectedParents()
	{
		return parents;
	}
	
	/**
		Get child extracts that was collected by the
		{@link #collectChildrenOfCollected(boolean)} and/or 
		{@link #collectChildrenOfSource(boolean)} collectors.
	*/
	public Set<Integer> getCollectedChildren()
	{
		return children;
	}
	

	/**
		Transformer implementation that given a set of derived bioassays
		load their linked extracts and store that in the parent class. The
		original source of derived bioassays is returned unmodifed.
	*/
	class CollectFromDerivedBioAssays
		extends AbstractSourceItemTransformer
	{
		
		private final boolean loadParents;
		
		CollectFromDerivedBioAssays(boolean loadParents) 
		{
			super(Item.DERIVEDBIOASSAY, Item.DERIVEDBIOASSAY);
			this.loadParents = loadParents;
		}


		@Override
		public Set<Integer> transform(TransformContext context, final Set<Integer> source) 
		{
			DbControl dc = context.getDbControl();
			
			Set<Integer> derviedBioAssays = source;
			if (loadParents)
			{
				derviedBioAssays = new DerivedBioAssayToParentDerivedBioAssayTransformer(true).transform(context, source);
			}
			
			ItemQuery<Extract> query = Extract.getQuery();
			query.setIncludes(context.getInclude());
			query.joinPermanent(Hql.innerJoin("derivedBioAssays", "dba"));
			query.restrictPermanent(
				Restrictions.in(
					Hql.alias("dba"),
					Expressions.parameter("derivedBioAssays")
				)
			);

			collected.addAll(safeIdList(dc, query, "derivedBioAssays", derviedBioAssays));
			//System.out.println("Collected extracts: " + collected);
			return source;
		}
		
	}
	
	/**
		Transformer implementation that given a set of raw bioassays
		load their linked extracts and store that in the parent class. The
		original source of raw bioassays is returned unmodifed.
	*/
	class CollectFromRawBioAssays
		extends AbstractSourceItemTransformer
	{
		
		CollectFromRawBioAssays() 
		{
			super(Item.RAWBIOASSAY, Item.RAWBIOASSAY);
		}
	
	
		@Override
		public Set<Integer> transform(TransformContext context, final Set<Integer> source) 
		{
			DbControl dc = context.getDbControl();
			
			ItemQuery<Extract> query = Extract.getQuery();
			query.setIncludes(context.getInclude());
			query.joinPermanent(Hql.innerJoin("rawBioAssays", "rba"));
			query.restrictPermanent(
				Restrictions.in(
					Hql.alias("rba"),
					Expressions.parameter("rawBioAssays")
				)
			);
			
			collected.addAll(safeIdList(dc, query, "rawBioAssays", source));
			//System.out.println("Collected extracts: " + collected);
			return source;
		}
	}

	/**
		Transformer implementation that store the given set of
		extracts in the parent class. The original source of extracts
		is returned unmodifed.
	*/
	class CollectFromExtracts
		extends AbstractSourceItemTransformer
	{
		
		CollectFromExtracts() 
		{
			super(Item.EXTRACT, Item.EXTRACT);
		}
	
		@Override
		public Set<Integer> transform(TransformContext context, final Set<Integer> source) 
		{
			collected.addAll(source);
			//System.out.println("Collected extracts: " + collected);
			return source;
		}
	}

	/**
		Transformer implementation that collect the parent extracts from 
		the source or collected extracts. The original source of extracts is 
		returned unmodifed.
	*/
	class CollectParents
		extends AbstractSourceItemTransformer
	{
		
		private final boolean useCollectedAsSource;
		private final boolean includeSource;
		
		
		CollectParents(boolean useCollectedAsSource, boolean includeSource) 
		{
			super(Item.EXTRACT, Item.EXTRACT);
			this.useCollectedAsSource = useCollectedAsSource;
			this.includeSource = includeSource;
		}
	
		@Override
		public Set<Integer> transform(TransformContext context, final Set<Integer> source) 
		{
			Set<Integer> sourceToUse = useCollectedAsSource ? collected : source;
			parents.addAll(new ExtractToParentExtractTransformer(includeSource).transform(context, sourceToUse));
			return source;
		}
	}

	/**
		Transformer implementation that collect the child extracts from 
		the source or collected extracts. The original source of extracts is 
		returned unmodifed.
	*/
	class CollectChildren
		extends AbstractSourceItemTransformer
	{
		
		private final boolean useCollectedAsSource;
		private final boolean includeSource;
		
		
		CollectChildren(boolean useCollectedAsSource, boolean includeSource) 
		{
			super(Item.EXTRACT, Item.EXTRACT);
			this.useCollectedAsSource = useCollectedAsSource;
			this.includeSource = includeSource;
		}
	
		@Override
		public Set<Integer> transform(TransformContext context, final Set<Integer> source) 
		{
			Set<Integer> sourceToUse = useCollectedAsSource ? collected : source;
			children.addAll(new ExtractToChildExtractTransformer(includeSource).transform(context, sourceToUse));
			return source;
		}
	}

	/**
		Transformer implementation that reset the collected extracts before forwarding the call
		to the parent transformer. This should wrap the first created transformer that
		collect extracts to the same CollectedExtract instance.
		@since 3.17
	*/
	class ResetCollected
		extends AbstractSourceItemTransformer
	{
		private final SourceItemTransformer parent;
		
		ResetCollected(SourceItemTransformer parent) 
		{
			super(parent.getSourceItemType(), parent.getTargetItemType());
			this.parent = parent;
		}
	
		@Override
		public Set<Integer> transform(TransformContext context, final Set<Integer> source) 
		{
			resetCollected();
			return parent.transform(context, source);
		}
	}
	
}
