/**
	$Id$

	Copyright (C) 2012 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.validator;


import java.util.List;

import org.apache.commons.collections4.IteratorUtils;

import net.sf.basedb.core.Application;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.util.extensions.ClientContext;
import net.sf.basedb.util.extensions.ExtensionsInvoker;
import net.sf.basedb.util.extensions.Registry;
import net.sf.basedb.util.extensions.manager.Settings;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.extensions.NodeValidatorAction;

/**
	Wrapper for node validators that hooks into the extension system when 
	validating nodes. The parent (or core) validator is always invoked first,
	then the extension system is queried for additional validators. Methods
	that return a value return the result from the last call (unless nothing
	else is specified).

	@author Nicklas
	@since 3.2
*/
public class ExtensionNodeValidator<I>
	implements NodeValidator<I>
{

	private final NodeValidator<I> parent;
	private final Object key;
	private List<NodeValidatorAction<I>> xtValidators;
	
	/**
		Wrap the given parent node validator.
	*/
	public ExtensionNodeValidator(NodeValidator<I> parent, Object key)
	{
		this.parent = parent;
		this.key = key;
	}
	
	/*
		From the NodeValidator interface
		-----------------------------
	*/
	/**
		Call the same method on the parent validator and then on
		all extension validator actions. The result from the last call
		is returned.
	 */
	@Override
	public boolean preMissingItem(DbControl dc, OverviewContext context, Node parentNode) 
	{
		boolean result = parent.preMissingItem(dc, context, parentNode);
		for (NodeValidatorAction<I> nvAction : getValidators(dc))
		{
			result = nvAction.preMissingItem(dc, context, parentNode);
		}
		return result;
	}

	/**
		Call the same method on the parent validator and then on
		all extension validator actions. The result from the last call
		is returned.
	 */
	@Override
	public boolean preDeniedItem(DbControl dc, OverviewContext context, Node parentNode)
	{
		boolean result = parent.preDeniedItem(dc, context, parentNode);
		for (NodeValidatorAction<I> nvAction : getValidators(dc))
		{
			result = nvAction.preDeniedItem(dc, context, parentNode);
		}
		return result;
	}

	/**
		Call the same method on the parent validator and then on
		all extension validator actions. The result from the last call
		is returned.
	 */
	@Override
	public boolean preValidate(DbControl dc, OverviewContext context, I item, Node parentNode) 
	{
		boolean result = parent.preValidate(dc, context, item, parentNode);
		for (NodeValidatorAction<I> nvAction : getValidators(dc))
		{
			result = nvAction.preValidate(dc, context, item, parentNode);
		}
		return result;
	}

	/**
		Call the same method on the parent validator and then on
		all extension validator actions.
	 */
	@Override
	public void postMissingItem(DbControl dc, OverviewContext context, Node node, Node parentNode) 
	{
		parent.postMissingItem(dc, context, node, parentNode);
		for (NodeValidatorAction<I> nvAction : getValidators(dc))
		{
			nvAction.postMissingItem(dc, context, node, parentNode);
		}
	}

	/**
		Call the same method on the parent validator and then on
		all extension validator actions.
	 */
	@Override
	public void postDeniedItem(DbControl dc, OverviewContext context, Node node, Node parentNode) 
	{
		parent.postDeniedItem(dc, context, node, parentNode);
		for (NodeValidatorAction<I> nvAction : getValidators(dc))
		{
			nvAction.postDeniedItem(dc, context, node, parentNode);
		}
	}

	/**
		Call the same method on the parent validator and then on
		all extension validator actions.
	*/
	@Override
	public void postValidate(DbControl dc, OverviewContext context, Node node, Node parentNode) 
	{
		parent.postValidate(dc, context, node, parentNode);
		for (NodeValidatorAction<?> nvAction : getValidators(dc))
		{
			nvAction.postValidate(dc, context, node, parentNode);
		}
	}

	/**
		Call the same method on the parent validator and then on
		all extension validator actions.
	*/
	@Override
	public void postValidateFolder(DbControl dc, OverviewContext context, Node folderNode, Node parentNode) 
	{
		parent.postValidateFolder(dc, context, folderNode, parentNode);
		for (NodeValidatorAction<?> nvAction : getValidators(dc))
		{
			nvAction.postValidateFolder(dc, context, folderNode, parentNode);
		}
	}
	// --------------------------------

	@Override
	public String toString()
	{
		return super.toString() + "-->" + parent.toString();
	}

	private Iterable<NodeValidatorAction<I>> getValidators(DbControl dc)
	{
		if (xtValidators == null)
		{
			// Create a context with the parent node as the current item
			ClientContext clientContext = new ClientContext(dc, key);
			
			// Find and use any registered extensions
			Registry registry = Application.getExtensionsManager().getRegistry();
			Settings settings = Application.getExtensionsManager().getSettings();
			ExtensionsInvoker<NodeValidatorAction<I>> invoker = registry.useExtensions(clientContext, settings, "net.sf.basedb.util.overview.validator");
			
			xtValidators = IteratorUtils.toList(invoker.iterator());
		}
		return xtValidators;
	}
	
}
