/**
	$Id: ResourceBundleWrapper.java 6444 2014-04-09 12:21:20Z nicklas $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.resources;


import java.util.Locale;
import java.util.MissingResourceException;
import java.util.ResourceBundle;

/**
	A wrapper around {@link ResourceBundle} objects. We use the wrapper because we don't
	want to throw exceptions if a key is missing. Instead, we return the key as the value
	and log a message to system log. There are also some other useful variants for retreiving
	values.
	
	@author Nicklas
	@since 2.16
	@base.modified $Date: 2014-04-09 14:21:20 +0200 (Wed, 09 Apr 2014) $
*/
public class ResourceBundleWrapper
{

	private static final org.slf4j.Logger log = 
		org.slf4j.LoggerFactory.getLogger(ResourceBundleWrapper.class);

	private final String name;
	private final ResourceBundle parent;
	
	private boolean returnKeys;

	/**
		The wrapper is created by {@link ResourceBundleFactory}.
		@param name The base name of the bundle
		@param parent The underlying {@link ResourceBundle}
	*/
	public ResourceBundleWrapper(String name, ResourceBundle parent)
	{
		this.name = name;
		this.parent = parent;
		this.returnKeys = false;
	}

	/**
		Get the base name of the resource bundle.
	*/
	public String getName()
	{
		return name;
	}
	
	/**
		Get the actual locale used by the bundle, which may be a different
		locale from the one that was used to load it.
	*/
	public Locale getLocale()
	{
		return parent.getLocale();
	}
	
	/**
		Set a flag indicating that the various <code>getString()</code> methods
		should always return the key instead of trying to find a value for it.
		
		@param returnKeys TRUE to always return the key, FALSE to try to find values
	*/
	public void setReturnKeys(boolean returnKeys)
	{
		this.returnKeys = returnKeys;
	}

	/**
		Get access to the underlying resource bundle.
	*/
	public ResourceBundle getResourceBundle()
	{
		return parent;
	}
	
	/**
		Utility method for returning the value for the key given
		by <code>key.0</code> (variant=false) or <code>key.1</code> 
		(variant=true) depending on the value of the variant parameter.
		@see #getString(String, String...)
	*/
	public String getString(String key, boolean variant, String... replacements)
	{
		return getString(key + (variant ? ".1" : ".0"), replacements);
	}
	
	/**
	 	Utility method for returning the value for the key given by:
	 	<code>key.index</code>.
	 	@see #getString(String, String...)
	*/
	public String getString(String key, int index, String... replacements)
	{
		return getString(key + "." + index, replacements);
	}
	
	/**
		Get a string value from the bundle. The value may contain placeholders
		of the form <code>{index}</code> which will be replaced by the 
		corresponding value from the replacements array. Index values
		that are outside the range of the array are not replaced.
		<p>
		If no value is found for the key or if the {@link #setReturnKeys(boolean)}
		has been enabled the key is returned.
		
		@param key The key for the value
		@param replacements An optional array with string replacments
		@return The value found for the key
	*/
	public String getString(String key, String... replacements)
	{
		if (returnKeys) return key;
		String value = key;
		try
		{
			value = parent.getString(key);
		}
		catch (MissingResourceException ex)
		{
			String msg = "Missing value for key '" + key + "' in bundle '" + getName() + "'";
			if (log.isDebugEnabled())
			{
				log.debug(msg, ex);
			}
			else
			{
				log.warn(msg);
			}
		}
		
		if (replacements != null && replacements.length > 0)
		{
			for (int i = 0; i < replacements.length; ++i)
			{
				value = value.replace("{" + i + "}", replacements[i]);
			}
		}
		return value;
	}
	
	@Override
	public String toString()
	{
		return "ResourceBundleWrapper[name=" + getName() + "; locale=" + getLocale() + "; parent=" + parent.toString();
	}
	
	
}
