/**
	$Id: InternalIdMethod.java 7619 2019-03-04 08:29:50Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins.batchimport;

import java.util.List;

import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.Values;

/**
	Identification method implementation that finds an item
	by using the <code>getById()</code> method.

	@author nicklas
	@version 2.8
	@base.modified $Date: 2019-03-04 09:29:50 +0100 (Mon, 04 Mar 2019) $
*/
public class InternalIdMethod
	implements IdMethod
{
	
	/**
		Use the internal ID for finding items.
	*/
	public static final InternalIdMethod INTERNAL_ID = 
		new InternalIdMethod("internalId", "Internal ID");
	
	private final String method;
	private final String title;

	/**
		Create a new id method.
	*/
	public InternalIdMethod(String method, String title)
	{
		this.method = method;
		this.title = title;
	}
	
	/*
		From the IdMethod interface
		---------------------------
	*/
	@Override
	public String getMethod()
	{
		return method;
	}
	
	@Override
	public String getTitle()
	{
		return title;
	}

	@Override
	public boolean isTrulyUnique()
	{
		return true;
	}

	@Override
	public String getColumnMappingParameterName()
	{
		return AbstractItemImporter.internalIdColumnMapping.getName();
	}
	
	/**
		Adds a restriction to the query:
		<code>id = :identifier</code>
	*/
	@Override
	public <I extends BasicItem> ItemQuery<I> prepareQuery(DbControl dc, ItemQuery<I> query)
	{
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("id"),
				Expressions.parameter("identifier")
			)
		);
		return query;
	}

	/**
		Find the item with <code>id = identifier</code>
	*/
	@Override
	public <I extends BasicItem> List<I> find(DbControl dc, ItemQuery<I> query, String identifier)
	{
		int id = Values.getInt(identifier);
		query.setParameter("identifier", id, Type.INT);
		return query.list(dc);
	}
	// ---------------------------
	
	/*
		From the Object class
		---------------------
	*/
	@Override
	public String toString()
	{
		return title;
	}
	// --------------------------

}
