import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.time.LocalDate;
import java.time.ZoneId;
import java.util.Date;
import java.util.regex.Pattern;

import org.apache.poi.ss.usermodel.CellStyle;
import org.apache.poi.ss.usermodel.Font;
import org.apache.poi.ss.usermodel.IndexedColors;
import org.apache.poi.ss.usermodel.Workbook;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;

import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.excel.CellStyleCreator;
import net.sf.basedb.util.excel.XlsxTableWriter;
import net.sf.basedb.util.excel.XlsxToCsvUtil;
import net.sf.basedb.util.excel.XlsxToCsvUtil.SheetInfo;
import net.sf.basedb.util.formatter.DateFormatter;
import net.sf.basedb.util.formatter.NumberFormatter;
import net.sf.basedb.util.parser.FlatFileParser;
import net.sf.basedb.util.parser.FlatFileParser.Data;
import net.sf.basedb.util.parser.FlatFileParser.Line;

/*
 $Id$

 Copyright (C) 2019 Nicklas Nordborg

 This file is part of BASE - BioArray Software Environment.
 Available at http://base.thep.lu.se/

 BASE is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 3
 of the License, or (at your option) any later version.

 BASE is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with BASE. If not, see <http://www.gnu.org/licenses/>.
 */

public class TestExcelFunctions
{
	static boolean ok = true;

	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}

	static boolean test_all()
	{
		write("++Testing Excel functions");

		test_create_xlsx(null, "data/test.xlsx", "Test"); // Creates a new workbook
		test_read_xlsx("data/test.xlsx", "Test");
		test_read_xlsx_via_ffp("data/test.xlsx", "Test");

		test_create_xlsx("data/test.xlsx", "data/test2.xlsx", "Test2"); // Adds a sheet to an existing workbook
		test_read_xlsx("data/test2.xlsx", null);
		test_read_xlsx_via_ffp("data/test2.xlsx", null);

		write("++Testing Excel functions " + (ok ? "OK" : "Failed") + "\n");
		return ok;
	}


	public static void test_create_xlsx(String inFile, String outFile, String name)
	{
		OutputStream out = null;
		Workbook workbook = null;
		try
		{
			if (inFile != null)
			{
				workbook = new XSSFWorkbook(new java.io.File(inFile));
			}
			
			XlsxTableWriter tw = new XlsxTableWriter(workbook, name);
			CellStyleCreator styleCreator = tw.getCellStyleCreator();

			CellStyle headerStyle = styleCreator.createStyle();
			Font bold = styleCreator.createFont();
			bold.setBold(true);
			headerStyle.setFont(bold);

			CellStyle commentStyle = styleCreator.createStyle();
			Font greyItalic = styleCreator.createFont();
			greyItalic.setItalic(true);
			greyItalic.setColor(IndexedColors.GREY_50_PERCENT.getIndex());
			commentStyle.setFont(greyItalic);
			
			Object[] data = new Object[7];
			// Headers
			data[0] = "Name";
			data[1] = "Line";
			data[2] = "Age";
			data[3] = "Pi";
			data[4] = "Birthdate";
			data[5] = "Now";
			data[6] = "Comment";
			
			styleCreator.setBaseStyle(headerStyle);
			tw.tablePrintHeaders(data);
			
			Date today = Date.from(LocalDate.now().atStartOfDay(ZoneId.systemDefault()).toInstant());
			Date now = new Date();
			
			data[0] = "Foo source";
			data[1] = 1;
			data[2] = 25;
			data[3] = Math.PI;
			data[4] = today;
			data[5] = now;
			data[6] = "Auto-formatted data line";
			styleCreator.setBaseStyle(null);
			tw.tablePrintData(data);
			
			tw.setColumnFormatter(3, new NumberFormatter(2, true));
			tw.setColumnFormatter(4, new DateFormatter("yyyy-MMM-dd"));
			tw.setColumnFormatter(5, new DateFormatter("yyyy-MMM-dd HH:mm"));
			
			data[0] = "Bar code";
			data[1] = 2;
			data[2] = 50;
			data[3] = Math.E;
			data[4] = today;
			data[5] = now;
			data[6] = "Column formatter line";

			tw.tablePrintData(data);
			
			tw.println();
			tw.print("Printed");
			tw.print(3);
			tw.print(555);
			tw.print(532.678f);
			tw.print(today);
			tw.print(now);
			tw.println("Printed line");
			styleCreator.setBaseStyle(commentStyle);
			tw.print("#End-of-file");
			tw.print((Object)null);
			tw.print("<-There should be an empty cell here");
			
			out = FileUtil.getOutputStream(new java.io.File(outFile));
			tw.saveTo(out);
			out.close();
			tw.close();

			write("--Create Excel file OK (" + outFile + ")");
		}
		catch (Throwable ex)
		{
			write("--Create Excel file FAILED (" + outFile + ")");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			FileUtil.close(workbook);
			FileUtil.close(out);
		}
	}
	
	public static void test_read_xlsx(String inFile, String name)
	{
		InputStream in = null;
		try
		{
			in = FileUtil.getInputStream(new java.io.File(inFile));
			
			XlsxToCsvUtil xlsx = new XlsxToCsvUtil();			
			xlsx.readWorkbook(in);
			in.close();
			
			if (name != null)
			{
				SheetInfo sheet = xlsx.getSheetAsCsv(name);
				in = sheet.parseToCsv();
			}
			else
			{
				in = xlsx.getWorkbookAsCsv();
			}
			
			BufferedReader reader = new BufferedReader(new InputStreamReader(in, xlsx.getCharset()));
			String line = null;
			while ((line = reader.readLine()) != null)
			{
				if (!TestUtil.getSilent()) write("  " + line);
			}
			
			write("--Read Excel file via XlsxToCsvUtil OK (" + inFile + ")");
		}
		catch (Throwable ex)
		{
			write("--Read Excel file via XlsxToCsvUtil FAILED (" + inFile + ")");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			FileUtil.close(in);
		}
	}

	public static void test_read_xlsx_via_ffp(String inFile, String name)
	{
		InputStream in = null;
		try
		{
			in = FileUtil.getInputStream(new java.io.File(inFile));
			
			boolean parseSections = name == null;
			
			FlatFileParser ffp = new FlatFileParser();
			ffp.setParseAllExcelSheets(parseSections);
			ffp.setDataHeaderRegexp(Pattern.compile("Name\\t.*"));
			ffp.setInputStream(in, name);
			
			while (true)
			{
				if (parseSections)
				{
					if (!ffp.hasMoreSections()) break;
					Line s = ffp.nextSection();
					if (!TestUtil.getSilent()) write("  SECTION: " + s.line());
				}
				ffp.parseHeaders();
				if (!TestUtil.getSilent()) 
				{
					write("  WORKSHEET: " + ffp.getExcelSheet());
					write("  HEADER:" + ffp.getLine(ffp.getLineCount()-1).line());
				}
				
				while (ffp.hasMoreData())
				{
					Data d = ffp.nextData();
					write(d);
				}
				if (!parseSections) break;
			}
			
			write("--Read Excel file via FlatFileParser OK (" + inFile + ")");
		}
		catch (Throwable ex)
		{
			write("--Read Excel file via FlatFileParser FAILED (" + inFile + ")");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			FileUtil.close(in);
		}
	}

	
	static void write(String message)
	{
		System.out.println(message);
	}

	static void write(Data d)
	{
		StringBuilder sb = new StringBuilder();
		for (int i = 0; i < d.columns(); i++)
		{
			sb.append("\t").append(d.getString(i));
		}
		if (!TestUtil.getSilent()) write("  DATA: " + sb);
	}
	
}
